/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxNoise: a noise generator
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <stdlib.h>

#define __AXEL_PLUGIN__
#include "axel.h"

/* Optimisation dependencies: none */
static AxBase ax_noise_optimise (AXEL axel, AxBase noise);

static AxBase
ax_noise_init (AXEL axel, AxBase base, CDSet parameters)
{
  ax_noise_optimise (axel, base);
  return base;
}

static AxBase
ax_noise_clone (AXEL axel, AxBase base)
{
  AxBase new_noise = ax_base_new (axel);
  ax_noise_optimise (axel, new_noise);
  return new_noise;
}

static int
ax_noise_destroy (AXEL axel, AxBase base)
{
  ss_free (base);
  return 0;
}

/* An AxChunkFunc for creating noise */
static AxCount
ax_noise_write_chunk (AXEL axel, AxChunk chunk, AxCount offset,
		      AxCount count, int channelname, void * data)
{
  AxPCM * d;
  AxPCM r, rmax = (AxPCM)RAND_MAX;
  AxCount i;

  ax_dprintf ("[ax_noise_write_chunk] (%p, +%ld) @ %ld\n", data, count,
	      offset);

  d = &chunk->data[offset];
  for (i = 0; i < count; i++) {
    r = (AxPCM) random ();
    r *= 2.0;
    r /= rmax;
    *d++ = r - 1.0;
  }

  return count;
}

static AxCount
ax_noise_process (AXEL axel, AxBase base, AxCount count,
		       AxStream input, AxStream output)
{
  return ax_stream_chunkfuncify (axel, output, count,
				 ax_noise_write_chunk, base);
}

static AxCount
ax_noise_length (AXEL axel, AxBase base)
{
  return AX_COUNT_INFINITE;
}

static struct _AxMethods _ax_noise_methods = {
  ax_noise_clone,
  ax_noise_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_noise_process,
  ax_noise_length,
  NULL, /* seek */
  NULL, /* flush */
};

static AxBase
ax_noise_optimise (AXEL axel, AxBase noise)
{
  ax_base_set_methods (axel, noise, &_ax_noise_methods);
  return noise;
}

static struct _AxMetaText noise_metatext = {
  "axelstd::noise",
  "Generators::Noise",
  "White noise generator",
  "Copyright (C) 2001 CSIRO Australia",
  "http://www.metadecks.org/axel/plugins/noise.html",
  AX_ONE_AUTHOR ("Conrad Parker", "Conrad.Parker@CSIRO.AU"),
};

static struct _AxPlugin noise_plugin = {
  &noise_metatext,
  AX_FLAGS_NONE,
  CD_EMPTY_SET, /* new scheme */
  ax_noise_init,
  CD_EMPTY_SET, /* process scheme */
  NULL, /* suggests */
  NULL, /* plugin data */
};

CDList
axel_load (AXEL axel)
{
  CDList plugins = cd_list_new (axel);

  plugins = cd_list_prepend (axel, plugins, CD_POINTER(&noise_plugin));

  return plugins;
}
