/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxBase: A generic interface for sound generation and processing.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#define __AXEL__
#include "axel.h"


AxBase
ax_new (AXEL axel, AxPlugin plugin, CDSet parameters)
{
  AxBase base;

  if (plugin == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  base = ax_base_new (axel);

  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  base->plugin = plugin;

  if (plugin->init != NULL)
    plugin->init (axel, base, parameters);

  return base;
}

CDSet
ax_suggest (AXEL axel, AxPlugin plugin, CDSet parameters)
{
  if (plugin == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  if (!plugin->suggest) {
    /* XXX: should ax_set_error: is NOOP appropriate ?? */
    return AxNone;
  }

  return plugin->suggest (axel, plugin, parameters, plugin->plugin_data);
}

static int
ax_parameter_scheme_get_key (AXEL axel, CDSet scheme_set, char * name)
{
  CDSet s;
  AxParameterScheme scheme;

  for (s = scheme_set; s; s = s->next) {
    scheme = (AxParameterScheme)s->data.s_pointer;
    if (!strcmp (scheme->name, name)) return s->key;
  }

  ax_set_error (axel, AX_ERROR_NOENTITY);

  return -1;
}

int
ax_get_init_parameter_key (AXEL axel, AxPlugin plugin, char * name)
{
  if (plugin == AxNone) {
    ax_dprintf ("[ax_get_init_parameter_key] plugin == AxNone\n");
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return ax_parameter_scheme_get_key (axel, plugin->init_scheme, name);
}

int
ax_get_parameter_key (AXEL axel, AxBase base, char * name)
{
  if (base == AxNone) {
    ax_dprintf ("[ax_get_parameter_key] base == AxNone\n");
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  if (base->plugin == AxNone) {
    ax_dprintf ("[ax_get_parameter_key] base->plugin == AxNone\n");
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return ax_parameter_scheme_get_key (axel, base->plugin->process_scheme, name);
}

AxParameter
ax_set_parameter (AXEL axel, AxBase base, int key, AxParameter parameter)
{
  ax_dprintf ("[ax_set_parameter] base %p, [%d] ==> %p\n", base, key,
	      parameter.s_pointer);
  base->parameters = cd_set_replace (axel, base->parameters, key, parameter);
  return parameter;
}

AxParameter
ax_get_parameter (AXEL axel, AxBase base, int key)
{
  AxParameter p = cd_set_find (axel, base->parameters, key);
  ax_dprintf ("[ax_get_parameter] base %p, [%d] == %p\n", base, key,
	      p.s_pointer);
  return p;
}

AxParameterType
ax_get_parameter_type (AXEL axel, AxBase base, int key)
{
  AxPlugin plugin = base->plugin;
  CDScalar k;
  AxParameterScheme scheme;

  k = cd_set_find (axel, plugin->process_scheme, key);
  scheme = (AxParameterScheme)k.s_pointer;
  return scheme->type;
}

AxBase
ax_base_new_subclass (AXEL axel, size_t size)
{
  AxBase base = ax_malloc (size);
  _ax_context_copy (axel, &base->context_limit);
  _ax_register_base (axel, base);
  return base;
}

AxBase
ax_base_new (AXEL axel)
{
  return ax_base_new_subclass (axel, sizeof (struct _AxBase));
}

AxCount
ax_base_get_mixlength (AXEL axel, AxBase base)
{
  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return _ax_base_get_mixlength (axel, base);
}

AxSamplerate
ax_base_get_samplerate (AXEL axel, AxBase base)
{
  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return _ax_base_get_samplerate (axel, base);
}

AxTempo
ax_base_get_tempo (AXEL axel, AxBase base)
{
  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return _ax_base_get_tempo (axel, base);
}

CDSet
ax_base_get_channels (AXEL axel, AxBase base)
{
  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  return _ax_base_get_channels (axel, base);
}

void *
ax_base_set_instance_data (AXEL axel, AxBase base, void * data)
{
  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  return _ax_set_instance_data (axel, base, data);
}

void *
ax_base_get_instance_data (AXEL axel, AxBase base)
{
  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  return _ax_get_instance_data (axel, base);
}

int
ax_base_has_samplerate (AXEL axel, AxBase base)
{
  AxSamplerate asr, bsr;

  asr = ax_get_samplerate (axel);
  bsr = _ax_base_get_samplerate (axel, base);

  return (asr == bsr);
}

int
ax_base_has_tempo (AXEL axel, AxBase base)
{
  AxTempo at, bt;

  at = ax_get_tempo (axel);
  bt = _ax_base_get_tempo (axel, base);

  return (at == bt);
}

int
ax_base_encompasses_mixlength (AXEL axel, AxBase base)
{
  AxCount aml, bml;

  aml = ax_get_mixlength (axel);
  bml = _ax_base_get_mixlength (axel, base);

  return (aml < bml);
}

int
ax_base_encompasses_channels (AXEL axel, AxBase base)
{
  CDSet s, as, bs;
  CDScalar k;

  as = ax_get_channels (axel);
  bs = _ax_base_get_channels (axel, base);

  for (s = as; s; s = s->next) {
    k = cd_set_find (axel, bs, s->key);
    if (k.s_pointer == NULL) return 0;
  }

  return 1;
}

AxMethods
ax_base_set_methods (AXEL axel, AxBase base, AxMethods methods)
{
  AxMethods old = base->methods;
  _ax_set_methods (axel, base, methods);
  return old;
}

AxMethods
ax_base_get_methods (AXEL axel, AxBase base)
{
  return _ax_get_methods (axel, base);
}

AxPlugin
ax_base_set_plugin (AXEL axel, AxBase base, AxPlugin plugin)
{
  AxPlugin old = base->plugin;
  _ax_set_plugin (axel, base, plugin);
  return old;
}

AxPlugin
ax_base_get_plugin (AXEL axel, AxBase base)
{
  return _ax_get_plugin (axel, base);
}

AxBase
ax_clone_subclass (AXEL axel, AxBase base)
{
  if (!base) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return NULL;
  }
  if (!base->methods || !base->methods->clone) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return NULL;
  }
  return _ax_clone (axel, base);
}

int
ax_destroy (AXEL axel, AxBase base)
{
  if (!base) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  _ax_unregister_base (axel, base);

  if (!base->methods || !base->methods->destroy) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return -1;
  }
  return _ax_destroy (axel, base);
}

int
ax_destroy_list (AXEL axel, CDList list)
{
  cd_list_destroy_with (axel, list, (CDDestroyFunc)ax_destroy);
  return 0;
}

/*
 * Prepare the methods for process, seek and length calls.
 *
 * "Prepare" means to make sure the base has enough internal buffers
 * to deal with the current context (sample rate, mixlength).
 *
 * If the methods has a ready() function, that is checked first. If it
 * does not have a ready() function, it is assumed to always be ready.
 */
AxBase
ax_prepare (AXEL axel, AxBase base)
{
  int is_ready = 0;

  if (!base) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return NULL;
  }
  if (base->methods && base->methods->prepare) {
    if (base->methods->ready) {
      is_ready = base->methods->ready (axel, base);
    }

    _ax_context_merge (axel, &base->context_limit);

    if (!is_ready) base =_ax_prepare (axel, base);
  }

  return base;
}

AxCount
ax_process_fast (AXEL axel, AxBase base, AxCount count, AxStream input,
		 AxStream output)
{
  AxCount n;

  if (!base) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }
  if (!base->methods || !base->methods->process) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return -1;
  }
  n = _ax_process (axel, base, count, input, output);
  if (n > 0) base->offset += n;
  return n;
}

AxCount
ax_process (AXEL axel, AxBase base, AxCount count, AxStream input,
	    AxStream output)
{
  AxCount processed;
  AxCount n;
  AxError error;
  char * str;

  ax_debug_down ();

  processed = ax_process_fast (axel, base, count, input, output);

  if (processed == -1) {
    error = ax_last_error (axel);
    str = ax_error_string (axel, error);
    ax_dprintf ("*** ERROR in ax_process: %s\n", str);
    switch (error) {
    case AX_ERROR_NOOP:
      n = ax_stream_write (axel, output, count, input);
      if (n > 0) {
	base->offset += n;
	processed = n;
      }
      break;
    case AX_ERROR_SILENCE:
      n = ax_stream_write0 (axel, output, count);
      if (n > 0) {
	base->offset += n;
	processed = n;
      }
      break;
    default: break;
    }
  }

  ax_debug_up ();

  return processed;
}

AxCount
ax_length (AXEL axel, AxBase base)
{
  if (!base) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }
  if (!base->methods || !base->methods->length) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return -1;
  }
  return _ax_length (axel, base);
}  

AxCount
ax_seek (AXEL axel, AxBase base, AxCount offset, int whence)
{
  AxCount new_offset, len;

  if (!base) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  new_offset = base->offset;

  switch (whence) {
  case SEEK_SET: new_offset = offset; break;
  case SEEK_CUR: new_offset += offset; break;
  case SEEK_END:
    len = ax_length (axel, base);
    if (len == -1) return -1;
    new_offset = len + offset;
    break;
  default:
    ax_set_error (axel, AX_ERROR_INVALID);
    return -1;
    break;
  }

  if (new_offset == base->offset) return new_offset;

  ax_dprintf ("SEEK %p @ %ld\n", base, new_offset);

  if (base->methods && base->methods->seek)
    base->offset = base->methods->seek (axel, base, new_offset);
  else
    base->offset = new_offset;

  return base->offset;
}

AxCount
ax_tell (AXEL axel, AxBase base)
{
  if (!base) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }
  return base->offset;
}

int
ax_flush (AXEL axel, AxBase base)
{
  if (!base) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }
  if (!base->methods || !base->methods->flush) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return -1;
  }
  return _ax_flush (axel, base);
}

AxMetaText
ax_get_meta_text (AXEL axel, AxBase base)
{
  AxPlugin plugin = base->plugin;
  return (plugin ? plugin->metatext : NULL);
}

AxMetaText
ax_set_meta_text (AXEL axel, AxBase base, AxMetaText mt)
{
  AxPlugin plugin;
  AxMetaText old;

  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  plugin = base->plugin;
  if (plugin == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  old = plugin->metatext;
  plugin->metatext = mt;
  return old;
}

int
ax_is_writeable (AXEL axel, AxBase base)
{
  AxPlugin plugin;

  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;
  if (plugin == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return (plugin->flags & AX_PLUGIN_WRITEABLE);
}

int
ax_is_seekable (AXEL axel, AxBase base)
{
  AxPlugin plugin;

  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;
  if (plugin == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return (plugin->flags & AX_PLUGIN_SEEKABLE);
}

int
ax_is_cacheable (AXEL axel, AxBase base)
{
  AxPlugin plugin;

  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;
  if (plugin == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return (plugin->flags & AX_PLUGIN_CACHEABLE);
}

int
ax_is_causal (AXEL axel, AxBase base)
{
  AxPlugin plugin;

  if (base == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;
  if (plugin == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return (plugin->flags & AX_PLUGIN_CAUSAL);
}
