/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxChannel: An indexed, sparse, monophonic PCM data container.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A channel contains a sequence of chunks which contain raw PCM data.
 *
 * Channels are named, such that when streams are mixed together their
 * correspondingly named channels are mixed together.
 *
 * A channel is indexed by sample count. When reading a channel, all
 * index points for which no chunk is defined are defined as zero. Writing
 * to a channel stops early if no chunk is available at a required point.
 *
 * If two or more chunks overlap, the chunk with the latest start index
 * is always used for both reading and writing data values for the region
 * of overlap.
 *
 * Invariants
 * ----------
 *
 * A channel must be contained within a stream.
 */

#define __AXEL__
#include "axel.h"

/* AxChannel */

AxChannel
ax_channel_new (AXEL axel)
{
  AxChannel c;

  c = (AxChannel) ax_malloc (sizeof (struct _AxChannel));

  c->chunks = cd_list_new (axel);
  c->_current_offset = 0;
  c->_current_chunk = AxNone;

  return c;
}

AxChannel
ax_channel_clone (AXEL axel, AxChannel channel)
{
  AxChannel new_channel = ax_channel_new (axel);
  new_channel->chunks = cd_list_clone (axel, channel->chunks,
				       (CDCloneFunc)ax_chunk_clone);
  return new_channel;
}

int
ax_channel_destroy (AXEL axel, AxBase base)
{
  AxChannel channel = (AxChannel)base;
  CDList l;

  for (l = channel->chunks; l; l = l->next) {
    ax_chunk_free (axel, (AxChunk)l->data.s_pointer);
  }

  cd_list_free (axel, channel->chunks);
  ax_free (channel);
  return 0;
}

AxChunk
ax_channel_add_chunk (AXEL axel, AxChannel channel, AxChunk chunk)
{
  channel->chunks = cd_list_insert (axel, channel->chunks, CD_TYPE_POINTER,
                                    CD_POINTER(chunk),
				    (CDCmpFunc)ax_chunk_later);
  return chunk;
}

AxChunk
ax_channel_add_new_chunk (AXEL axel, AxChannel channel, AxCount offset,
			  AxCount length)
{
  AxChunk chunk = ax_chunk_new (axel, offset, length);
  return ax_channel_add_chunk (axel, channel, chunk);
}

void
ax_channel_remove_chunk (AXEL axel, AxChannel channel, AxChunk chunk)
{
  channel->chunks = cd_list_remove (axel, channel->chunks, CD_TYPE_POINTER,
				    CD_POINTER(chunk));
}

AxChunk
ax_channel_find_chunk_before (AXEL axel, AxChannel channel, AxCount index)
{
  CDList l;
  AxChunk u, up = AxNone;

  for (l = channel->chunks; l; l = l->next) {
    u = (AxChunk)l->data.s_pointer;
    if (u->start_index > index) return up;
    up = u;
  }

  return up;
}

/*
 * ax_chunk_item_valid_length (l)
 *
 * Returns the length for which the chunk item 'l' is valid. The valid
 * length of a chunk item in a channel is defined as the minimum of
 * (the chunk's actual length) and (the difference between the offset
 * of the chunk and the offset of the following chunk). ie. if the
 * chunk is followed by another chunk that cuts it off early, the valid
 * length is the difference between the two chunk offsets. Crikey, email
 * me if you're confused.
 */
static AxCount
ax_chunk_item_valid_length (CDList l)
{
  CDList ln = l->next;
  AxChunk u = (AxChunk)l->data.s_pointer, un;

  if (ln == AxNone) {
    return u->length;
  } else {
    un = (AxChunk)ln->data.s_pointer;
    return MIN (u->length, un->start_index - u->start_index);
  }
}

CDList
ax_channel_get_chunk_item_at (AxChannel channel, AxCount offset)
{
  CDList l;
  AxChunk u;
  AxCount vl; /* valid length */

  for (l = channel->chunks; l; l = l->next) {
    u = (AxChunk)l->data.s_pointer;
    vl = ax_chunk_item_valid_length (l);
    if (u->start_index <= offset && u->start_index + vl > offset) {
      /* chunk validly spans offset: return its item */
      return l;
    }
  }

  /* No chunks found spanning offset */
  return AxNone;
}

CDList
ax_channel_get_chunk_item_after (AxChannel channel, AxCount offset)
{
  CDList l;
  AxChunk u;

  for (l = channel->chunks; l; l = l->next) {
    u = (AxChunk)l->data.s_pointer;
    if (u->start_index >= offset) return l;
  }

  return AxNone;
}

AxCount
ax_channel_write0 (AXEL axel, AxChannel channel, AxCount length)
{
  CDList l = channel->_current_chunk;
  AxChunk u;
  AxCount remaining = length, n, vl;
  AxCount offset = channel->_current_offset;

  while (remaining > 0) {
    if (l == AxNone) break; /* No more chunks */

    u = (AxChunk)l->data.s_pointer;

    if (u->start_index > offset) { /* skip ahead to start of next chunk */
      n = MIN (remaining, u->start_index - offset);
      offset += n;
      remaining -= n;
    }

    if (remaining > 0) {
      vl = ax_chunk_item_valid_length (l);
      n = _ax_chunk_clear_region (axel, u, offset, MIN(remaining, vl),
				  0, NULL);
      offset += n;
      remaining -= n;
    }

    l = l->next;
  }

  channel->_current_chunk = l;
  channel->_current_offset += length;

  return length;
}

/*
 * ax_channel_chunkfuncify (axel, channel, count, func, data)
 *
 * Apply the AxChunkFunc func to 'count' samples from consecutive chunks
 * of 'channel'.
 * Stops early if the channel runs out of chunks.
 * Returns the number of samples func'ed.
 */
AxCount
ax_channel_chunkfuncify (AXEL axel, AxChannel channel, AxCount count,
			 AxChunkFunc func, int channelname, void * data)
{
  AxChunk u;
  AxCount remaining = count, funced = 0, n, vl;
  AxError error;

  ax_dprintf ("[ax_channel_chunkfuncify] (%p, +%ld) @ %ld\n",
	  channel, count, channel->_current_offset);

  while (remaining > 0) {
    channel->_current_chunk =
      ax_channel_get_chunk_item_at (channel, channel->_current_offset);
    if (channel->_current_chunk == AxNone) {
      ax_dprintf ("[ax_channel_chunkfuncify] channel incomplete, funced %ld\n",
		  funced);
      return funced; /* Channel incomplete */
    }

    u = (AxChunk)channel->_current_chunk->data.s_pointer;
    vl = ax_chunk_item_valid_length (channel->_current_chunk);

    n = func (axel, u, channel->_current_offset, MIN(remaining, vl),
	      channelname, data);

    if (n == -1) {
      error = ax_last_error (axel);
      switch (error) {
      case AX_ERROR_SILENCE:
	n = _ax_chunk_clear_region (axel, u, channel->_current_offset,
				    MIN(remaining, vl), 0, NULL);
	break;
      default:
	n = 0;
      }
    }

    funced += n;
    remaining -= n;
    channel->_current_offset += n;
  }

  return funced;
}

/*
 * ax_channel_chunkchunkfuncify (axel, src, dest, count, func, data)
 *
 * Apply the AxChunkChunkFunc func to corresponding chunks of 'src' and
 * 'dest' to 'count' samples.
 * Stops early if 'dest' cannot contain part of the region for which
 * 'src' is defined. Copies zeroes to 'dest' wherever 'src' is empty.
 * Returns the number of samples func'ed.
 */
AxCount
ax_channel_chunkchunkfuncify (AXEL axel, AxChannel src, AxChannel dest,
			      AxCount count, AxChunkChunkFunc func,
			      int channelname, void * data)
{
  AxChunk su, du;
  AxCount remaining = count, funced = 0, n, vl;
  AxError error;

  ax_dprintf ("[ax_channel_ccf...] (%p -> %p, +%ld), src @ %ld, dest @ %ld\n",
	  src, dest, count, src->_current_offset, dest->_current_offset);

  while (remaining > 0) {
    n = 0; /* watch for early changes to n */

    dest->_current_chunk =
	ax_channel_get_chunk_item_at (dest, dest->_current_offset);
    if (dest->_current_chunk == AxNone) {
      ax_dprintf ("[ax_channel_ccf...] channel incomplete after %ld\n", funced);
      return funced; /* Destination channel incomplete */
    }

    src->_current_chunk =
	ax_channel_get_chunk_item_at (src, src->_current_offset);
    if (src->_current_chunk == AxNone) { /* No source data at offset */
      src->_current_chunk =
	ax_channel_get_chunk_item_after (src, src->_current_offset);

      if (src->_current_chunk == AxNone) {
        /* No following source data at all */
	ax_dprintf ("[ax_channel_ccf...] no source data after %ld\n",
		    src->_current_offset);
	n = ax_channel_write0 (axel, dest, remaining);
	funced += n;
	remaining -= n;
	return funced;
      } else {
	ax_dprintf ("[ax_channel_ccf...] no source data at %ld\n",
		    src->_current_offset);
      }
    }

    /* *** Now, sl is the current or following source chunk item *** */

    su = (AxChunk)src->_current_chunk->data.s_pointer;

    if (su->start_index > src->_current_offset) { /* No source data at offset */
      ax_dprintf ("[ax_channel_ccf...] no source data at %ld (warn 2)\n",
		  src->_current_offset);
      n = ax_channel_write0 (axel, dest, su->start_index - src->_current_offset);
      funced += n;
      remaining -= n;
      src->_current_offset += n;
    }

    if (remaining > 0) {
      if (n > 0) {
	dest->_current_chunk =
	  ax_channel_get_chunk_item_at (dest, dest->_current_offset);
	if (dest->_current_chunk == AxNone) {
	  ax_dprintf ("[ax_channel_ccf...] dest incomplete after %ld\n",
		      dest->_current_chunk);
	  return funced; /* Destination channel incomplete */
	}
      }
 
      du = (AxChunk)dest->_current_chunk->data.s_pointer;

      vl = ax_chunk_item_valid_length (dest->_current_chunk);
      n = func (axel, su, src->_current_offset, du, dest->_current_offset,
                MIN(remaining, vl), channelname, data);

      if (n == -1) {
	error = ax_last_error (axel);
	switch (error) {
	case AX_ERROR_SILENCE:
	  n = _ax_chunk_clear_region (axel, du, dest->_current_offset,
				      MIN(remaining, vl), 0, NULL);
	  break;
	default:
	  n = 0;
	}
      }

      funced += n;
      remaining -= n;
      src->_current_offset += n;
      dest->_current_offset += n;
    }
  }

  ax_dprintf ("[ax_channel_ccf...] funced %ld\n", funced);

  return funced;
}

/*
 * ax_channel_chunkchunkchunkfuncify (axel, src1, src2, dest, count, func,
 *                                    data)
 *
 * Apply the AxChunkChunkChunkFunc func to corresponding chunks of 'src1',
 * 'src2' and 'dest' to 'count' samples.
 * Stops early if 'dest' cannot contain part of the region for which
 * both 'src1' and 'src2' is defined. Copies zeroes to 'dest' wherever
 * either 'src1' or 'src2' are empty.
 * Returns the number of samples func'ed.
 */
AxCount
ax_channel_chunkchunkchunkfuncify (AXEL axel, AxChannel src1, AxChannel src2,
				   AxChannel dest, AxCount count,
				   AxChunkChunkChunkFunc func,
				   int channelname, void * data)
{
  AxChunk s1u, s2u, du;
  AxCount remaining = count, funced = 0, n, vl, undef_length;
  AxError error;

  while (remaining > 0) {
    n = 0; /* watch for early changes to n */

    dest->_current_chunk =
      ax_channel_get_chunk_item_at (dest, dest->_current_offset);
    if (dest->_current_chunk == AxNone)
      return funced; /* Destination channel incomplete */

    src1->_current_chunk =
      ax_channel_get_chunk_item_at (src1, src1->_current_offset);
    if (src1->_current_chunk == AxNone) {
      src1->_current_chunk =
	ax_channel_get_chunk_item_after (src1, src1->_current_offset);
    }

    src2->_current_chunk =
      ax_channel_get_chunk_item_at (src2, src2->_current_offset);
    if (src2->_current_chunk == AxNone) {
      src2->_current_chunk =
	ax_channel_get_chunk_item_after (src2, src2->_current_offset);
    }

    if (src1->_current_chunk == AxNone || src2->_current_chunk == AxNone) {
      n = ax_channel_write0 (axel, dest, remaining);
      funced += n;
      remaining -= n;
      return funced;
    }

    s1u = (AxChunk)src1->_current_chunk->data.s_pointer;
    s2u = (AxChunk)src2->_current_chunk->data.s_pointer;

    if (s1u->start_index > src1->_current_offset ||
	s2u->start_index > src2->_current_offset) {
      undef_length = MAX (s1u->start_index - src1->_current_offset,
			  s2u->start_index - src2->_current_offset);
      n = ax_channel_write0 (axel, dest, undef_length);
      funced += n;
      remaining -= n;
      src1->_current_offset += n;
      src2->_current_offset += n;
    }
    
    if (remaining > 0) {
      if (n > 0) {
	dest->_current_chunk =
          ax_channel_get_chunk_item_at (dest, dest->_current_offset);
	if (dest->_current_chunk == AxNone)
          return funced; /* Destination channel incomplete */
      }
      
      du = (AxChunk)dest->_current_chunk->data.s_pointer;
      
      vl = ax_chunk_item_valid_length (dest->_current_chunk);
      n = func (axel, s1u, src1->_current_offset, s2u, src2->_current_offset,
		du, dest->_current_offset, MIN(remaining, vl), channelname,
		data);

      if (n == -1) {
	error = ax_last_error (axel);
	switch (error) {
	case AX_ERROR_SILENCE:
	  n = _ax_chunk_clear_region (axel, du, dest->_current_offset,
				      MIN(remaining, vl), 0, NULL);
	  break;
	default:
	  n = 0;
	}
      }

      funced += n;
      remaining -= n;
      src1->_current_offset += n;
      src2->_current_offset += n;
      dest->_current_offset += n;
    }
  }

  return funced;
}

/*
 * ax_channel_copy (src, dest, count)
 *
 * Copies 'count' samples from 'src' to 'dest'.
 * Stops early if 'dest' cannot contain part of 'src'. Copies zeroes to
 * 'dest' wherever 'src' is correspondingly empty.
 * Returns the count of samples actually copied.
 */
AxCount
ax_channel_copy (AXEL axel, AxChannel src, AxChannel dest, AxCount count)
{
  return ax_channel_chunkchunkfuncify (axel, src, dest, count,
				       (AxChunkChunkFunc)_ax_chunk_copy,
				       0, NULL);
}

AxCount
ax_channel_mix (AXEL axel, AxChannel src, AxChannel dest, AxCount count)
{
  return ax_channel_chunkchunkfuncify (axel, src, dest, count,
                                       (AxChunkChunkFunc)_ax_chunk_add_inplace,
				       0, NULL);
}

AxCount
ax_channel_interleave_2 (AXEL axel, AxChannel src1, AxChannel src2,
			 AxPCM * dest, AxCount count)
{
  return
    ax_channel_chunkchunkfuncify (axel, src1, src2, count,
				  (AxChunkChunkFunc)_ax_chunk_interleave_2,
				  0, dest);
}

AxCount
ax_channel_deinterleave_2 (AXEL axel, AxChannel dest1, AxChannel dest2,
			   AxPCM * src, AxCount count)
{
  return
    ax_channel_chunkchunkfuncify (axel, dest1, dest2, count,
				  (AxChunkChunkFunc)_ax_chunk_deinterleave_2,
				  0, src);
}

AxCount
_ax_channel_write (AXEL axel, AxChannel channel, AxCount count,
		   AxChannel data)
{
  AxCount n = ax_channel_copy (axel, data, channel, count);
  data->_current_offset += n;
  data->_current_chunk =
    ax_channel_get_chunk_item_after (data, data->_current_offset);
  channel->_current_offset += n;
  channel->_current_chunk =
    ax_channel_get_chunk_item_after (channel, channel->_current_offset);
  return n;
}

AxCount
_ax_channel_length (AXEL axel, AxChannel channel)
{
  AxChunk last =
    (AxChunk)(cd_list_last (axel, channel->chunks, CD_TYPE_POINTER)).s_pointer;
  if (last == AxNone) return 0;
  return (last->start_index + last->length);
}

AxCount
_ax_channel_seek (AXEL axel, AxChannel channel, AxCount offset)
{
  if (offset == channel->_current_offset) return offset;
  channel->_current_offset = offset;
  /* Cache the current chunk */
  channel->_current_chunk =
    ax_channel_get_chunk_item_after (channel, offset);
  return offset;
}
