/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxChunk: A contiguous chunk of monophonic PCM data.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A chunk contains raw PCM data.
 *
 * Invariants
 * ----------
 *
 * A chunk must always be contained within a channel. The data within
 * a chunk is only valid where it is not overlapped by a later chunk
 * in the same channel; elsewhere, the chunk's data is not used.
 *
 */

#define __AXEL__
#include "axel.h"

typedef AxCount (*AxPFunc) (AxPCM * src, AxCount count, void * data);
typedef AxCount (*AxPPFunc) (AxPCM * src, AxPCM * dest, AxCount count,
			     void * data);
typedef AxCount (*AxPPPFunc) (AxPCM * src1, AxPCM * src2, AxPCM * dest,
			      AxCount count, void * data);

AxChunk
ax_chunk_new (AXEL axel, AxCount start_index, AxCount length)
{
  AxChunk u;

  u = (AxChunk) ax_malloc (sizeof (struct _AxChunk));
  u->start_index = start_index;
  u->length = length;

  u->data = (AxPCM *) ax_malloc (length * sizeof (AxPCM));

  return u;
}

AxChunk
ax_chunk_clone (AXEL axel, AxChunk chunk)
{
  AxChunk u = ax_chunk_new (axel, chunk->start_index, chunk->length);
  memcpy (u->data, chunk->data, (chunk->length * sizeof (AxPCM)));
  return u;
}

void
ax_chunk_free (AXEL axel, AxChunk chunk)
{
  ax_free (chunk->data);
  ax_free (chunk);
}

int
ax_chunk_later (AXEL axel, AxChunk u1, AxChunk u2)
{
  return (u1->start_index > u2->start_index);
}

AxCount
ax_chunk_clear (AXEL axel, AxChunk chunk)
{
  AxCount len = chunk->length;
  memset (chunk->data, (AxPCM)0, len * sizeof (AxPCM));
  return len;
}


/*
 * FUNCTION APPLIERS
 */

static AxCount
_ax_pfunc_apply (AXEL axel, AxPFunc func, AxChunk chunk, AxCount start,
		 AxCount count, void * data)
{
  AxCount chunk_start = start - chunk->start_index;

  /* chunk_start: 'start' relative to chunk */
  if (chunk_start < 0) {
    count += chunk_start;
    chunk_start = 0;
  }

  if (chunk_start + count > chunk->length) {
    count = chunk->length - chunk_start;
  }

  func (&chunk->data[chunk_start], count, data);

  return (count);
}

/*
 * _ax_ppfunc_apply (src, src_offset, dest, dest_offset, count, func)
 *
 * Apply AxPCMPCMFunc 'func' to 'count' samples of chunks 'src' and 'dest'.
 */
static AxCount
_ax_ppfunc_apply (AXEL axel, AxPPFunc func, AxChunk src, AxCount src_offset,
		  AxChunk dest, AxCount dest_offset, AxCount count,
		  void * data)
{
  AxCount dest_start = dest_offset - dest->start_index;
  AxCount src_start;
  AxPCM *s, *d;

  if (dest_start < 0) {
    count += dest_start;
    dest_offset -= dest_start;
    src_offset -= dest_start;
    dest_start = 0;
  }

  src_start = src_offset - src->start_index;

  /* Shorten count if either chunk is shorter than required */
  if (src_start + count > src->length)
    count = src->length - src_start;

  if (dest_start + count > dest->length)
    count = dest->length - dest_start;

  s = &src->data[src_start];
  d = &dest->data[dest_start];

  func (s, d, count, data);

  return count;
}

/*
 * _ax_pppfunc_apply (src, src_offset, dest, dest_offset, count, func)
 *
 * Apply AxPPPFunc 'func' to 'count' samples of chunks 'src1', 'src2'
 * and 'dest'.
 */
static AxCount
_ax_pppfunc_apply (AXEL axel, AxPPPFunc func,
		   AxChunk src1, AxCount src1_offset,
		   AxChunk src2, AxCount src2_offset,
		   AxChunk dest, AxCount dest_offset,
		   AxCount count, void * data)
{
  AxCount dest_start = dest_offset - dest->start_index;
  AxCount src1_start, src2_start;
  AxPCM *s1, *s2, *d;

  if (dest_start < 0) {
    count += dest_start;
    dest_offset -= dest_start;
    src1_offset -= dest_start;
    src2_offset -= dest_start;
    dest_start = 0;
  }

  src1_start = src1_offset - src1->start_index;
  src2_start = src2_offset - src2->start_index;

  /* Shorten count if any chunk is shorter than required */
  if (src1_start + count > src1->length)
    count = src1->length - src1_start;

  if (src2_start + count > src2->length)
    count = src2->length - src2_start;

  if (dest_start + count > dest->length)
    count = dest->length - dest_start;

  s1 = &src1->data[src1_start];
  s2 = &src2->data[src2_start];
  d = &dest->data[dest_start];

  func (s1, s2, d, count, data);

  return count;
}

/* PFunc appliers */

/*
 * _ax_chunk_clear_region (chunk, start, length, unused)
 *
 * Clear 'chunk' from stream index 'start' for 'length' samples.
 * Returns the count of samples actually cleared.
 */
AxCount
_ax_chunk_clear_region (AXEL axel, AxChunk chunk, AxCount start,
			AxCount length, int channelname, void * unused)
{
  return _ax_pfunc_apply (axel,_ax_pcm_clear_region,
			  chunk, start, length, NULL);
			  
}

AxCount
_ax_chunk_gain (AXEL axel, AxChunk chunk, AxCount start, AxCount count,
		int channelname, /* (AxPCM *) */ void * gain)
{
  return _ax_pfunc_apply (axel, _ax_pcm_gain, chunk, start, count, gain);
}


/* PPFunc appliers */

/*
 * _ax_chunk_copy (src, dest, offset, count, unused)
 *
 * Copy data from 'src' to 'dest' from stream index 'offset' for
 * 'count'. Returns the count of samples actually copied.
 */
AxCount
_ax_chunk_copy (AXEL axel, AxChunk src, AxCount src_offset, AxChunk dest,
		AxCount dest_offset, AxCount count, int channelname,
		void * unused)
{
  return _ax_ppfunc_apply (axel, _ax_pcm_copy, src, src_offset,
			   dest, dest_offset, count, NULL);
}

/*
 * _ax_chunk_add_inplace (src, dest, start, count, unused)
 *
 * Add data from 'src' to data in 'dest' from stream index 'start' for
 * 'count' samples. Returns the count of samples actually added.
 */
AxCount
_ax_chunk_add_inplace (AXEL axel, AxChunk src, AxCount src_offset,
		       AxChunk dest, AxCount dest_offset, AxCount count,
		       int channelname, void * unused)
{
  return _ax_ppfunc_apply (axel, _ax_pcm_add, src, src_offset,
			   dest, dest_offset, count, NULL);
			   
}

/*
 * _ax_chunk_mult_inplace (src, dest, start, count, unused)
 *
 * Multiply data of 'dest' by that of 'src' from stream index 'start' for
 * 'count' samples. Returns the count of samples actually multiplied.
 */
AxCount
_ax_chunk_mult_inplace (AXEL axel, AxChunk src, AxCount src_offset,
			AxChunk dest, AxCount dest_offset, AxCount count,
			int channelname, void * unused)
{
  return _ax_ppfunc_apply (axel, _ax_pcm_mult, src, src_offset,
			   dest, dest_offset, count, NULL);
}

/*
 * _ax_chunk_fade_inplace (src, src_offset, dest, dest_offset, count, unused)
 *
 * Fade data of 'dest' by that of 'src' from stream index 'start' for
 * 'count' samples. Returns the count of samples actually faded.
 */
AxCount
_ax_chunk_fade_inplace (AXEL axel, AxChunk src, AxCount src_offset,
			AxChunk dest, AxCount dest_offset, AxCount count,
			int channelname, void * unused)
{
  return _ax_ppfunc_apply (axel, _ax_pcm_fade, src, src_offset,
			   dest, dest_offset, count, NULL);
}

/*
 * _ax_chunk_interleave_2 (axel, src1, src1_offset, src2, src2_offset,
 *                         count, dest)
 *
 * Interleave data of 'src1' from stream index 'src1_offset' with that
 * of 'src2' from stream index 'src2_offset' for 'count' samples, storing
 * the resulting interleaved PCM data in 'dest'.
 *
 * Returns the count of sample frames interleaved.
 */
AxCount
_ax_chunk_interleave_2 (AXEL axel, AxChunk src1, AxCount src1_offset,
			AxChunk src2, AxCount src2_offset, AxCount count,
			int unused, void * dest)
{
  return _ax_ppfunc_apply (axel, _ax_pcm_interleave_2, src1, src1_offset,
			   src2, src2_offset, count, dest);
}

/*
 * _ax_chunk_deinterleave_2 (axel, dest1, dest1_offset, dest2, dest2_offset,
 *                           count, src)
 *
 * Deinterleave data of 'src' and store the result in streams 'dest1' (from
 * stream index 'dest1_offset') and 'dest2' (from stream index
 * 'dest2_offset').
 *
 * Returns the number of sample frames deinterleaved.
 */
AxCount
_ax_chunk_deinterleave_2 (AXEL axel, AxChunk dest1, AxCount dest1_offset,
			  AxChunk dest2, AxCount dest2_offset, AxCount count,
			  int unused, void * src)
{
  return _ax_ppfunc_apply (axel, _ax_pcm_deinterleave_2, dest1, dest1_offset,
			   dest2, dest2_offset, count, src);
}

/* PPPFunc appliers */

/*
 * _ax_chunk_blend_inplace (src, src_offset, blend, blend_offset,
 *                          dest, dest_offset, count, unused)
 *
 * Blend data of 'src' into that of 'dest' by blend values given in
 * 'blend'.
 */
AxCount
_ax_chunk_blend_inplace (AXEL axel, AxChunk src, AxCount src_offset,
			 AxChunk blend, AxCount blend_offset,
			 AxChunk dest, AxCount dest_offset, AxCount count,
			 int channelname, void * unused)
{
  return _ax_pppfunc_apply (axel, _ax_pcm_blend, src, src_offset,
			    blend, blend_offset, dest, dest_offset,
			    count, NULL);
}
