/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxContext: AXEL core context.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <string.h>

#define __AXEL__
#include "axel.h"

static void
ax_context_destroy (AXEL axel)
{
  AxContext ctx = axel->context;
  AxEnvironment env = axel->environment;

  env->purging = 1;
  /* XXX: ax_destroy_list (axel, env->plugins); */
  /* XXX:  ax_destroy_list (axel, env->bases); */
  ax_free (ctx);
  ax_free (env);
}

static AXEL
ax_add_thread_context (AxContext ctx, AxEnvironment env)
{
  AXEL axel = ax_malloc (sizeof (struct _AxThreadContext));
  axel->context = ctx;
  axel->environment = env;
  env->refcount++;
  return axel;
}

AxContext
_ax_context_copy (AXEL axel, AxContext dest)
{
  AxContext ctx = axel->context;

  if (dest == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  dest->samplerate = ctx->samplerate;
  dest->tempo = ctx->tempo;
  dest->mixlength = ctx->mixlength;
  dest->channels = cd_set_clone_keys (axel, ctx->channels);

  return dest;
}

/*
 * _ax_context_merge (axel, dest)
 *
 * Merges the context of axel into context 'dest'. Copies over the samplerate
 * and tempo, and expands the mixlength and channels if they are greater in
 * 'axel's context than in 'dest'.
 */
AxContext
_ax_context_merge (AXEL axel, AxContext dest)
{
  AxContext ctx = axel->context;
  CDSet s;
  CDScalar k;

  if (dest == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  dest->samplerate = ctx->samplerate;
  dest->tempo = ctx->tempo;

  if (ctx->mixlength > dest->mixlength)
    dest->mixlength = ctx->mixlength;

  for (s = ctx->channels; s; s = s->next) {
    k = cd_set_find (axel, dest->channels, s->key);
    if (k.s_pointer == NULL)
      dest->channels = cd_set_insert (axel, dest->channels, s->key,
				      CD_POINTER(NULL));
  }

  return dest;
}

/*
 * axel_init ()
 */
AXEL
axel_init (void)
{
  AXEL axel;
  AxEnvironment env =
    (AxEnvironment) ax_malloc (sizeof (struct _AxEnvironment));
  AxContext ctx = (AxContext) ax_malloc (sizeof (struct _AxContext));

  env->refcount = 0;
  env->plugins = cd_list_new (ctx);
  env->bases = cd_list_new (ctx);
  env->purging = FALSE;

  ctx->mixlength = AXEL_DEFAULT_MIXLENGTH;
  ctx->samplerate = AXEL_DEFAULT_SAMPLERATE;
  ctx->tempo = AXEL_DEFAULT_TEMPO;

  axel = ax_add_thread_context (ctx, env);
  ax_channelset_initialise_defaults (axel);
  ctx->channels = AX_MONO;

  ax_plugin_initialise_defaults (axel);

  return axel;
}

/*
 * axel_init_clone ()
 */
AXEL
axel_init_clone (AXEL axel)
{
  AXEL new_axel = ax_add_thread_context (axel->context, axel->environment);
  return new_axel;
}

void
axel_purge (AXEL axel)
{
  AxEnvironment env = axel->environment;
  env->refcount--;
  if (env->refcount <= 0) {
    ax_context_destroy (axel);
  }
  ax_free (axel);
}

AxError
ax_set_error (AXEL axel, AxError error)
{
  AxError old = axel->last_error;
  axel->last_error = error;
  return old;
}

AxError
ax_last_error (AXEL axel)
{
  return axel->last_error;
}

AxCount
ax_set_mixlength (AXEL axel, AxCount mixlength)
{
  AxContext ctx = axel->context;
  AxCount old = ctx->mixlength;
  ctx->mixlength = mixlength;
  return old;
}

AxCount
ax_get_mixlength (AXEL axel)
{
  AxContext ctx = axel->context;
  return ctx->mixlength;
}

AxSamplerate
ax_set_samplerate (AXEL axel, AxSamplerate samplerate)
{
  AxContext ctx = axel->context;
  AxSamplerate old = ctx->samplerate;
  ctx->samplerate = samplerate;
  return old;
}

AxSamplerate
ax_get_samplerate (AXEL axel)
{
  AxContext ctx = axel->context;
  return ctx->samplerate;
}

AxTempo
ax_set_tempo (AXEL axel, AxTempo tempo)
{
  AxContext ctx = axel->context;
  AxTempo old = ctx->tempo;
  ctx->tempo = tempo;
  return old;
}

AxTempo
ax_get_tempo (AXEL axel)
{
  AxContext ctx = axel->context;
  return ctx->tempo;
}

CDSet
ax_set_channels (AXEL axel,  CDSet channels)
{
  AxContext ctx = axel->context;
  CDSet old = ctx->channels;
  ctx->channels = cd_set_clone_keys (axel, channels);
  return old;
}

CDSet
ax_get_channels (AXEL axel)
{
  AxContext ctx = axel->context;
  return ctx->channels;
}

AXEL
_ax_register_plugin (AXEL axel, AxPlugin plugin)
{
  AxEnvironment env = axel->environment;
  env->plugins = cd_list_append (axel, env->plugins, CD_POINTER(plugin));
  return axel;
}

AXEL
_ax_unregister_plugin (AXEL axel, AxPlugin plugin)
{
  AxEnvironment env = axel->environment;
  if (env->purging) return axel;

  env->plugins = cd_list_remove (axel, env->plugins, CD_TYPE_POINTER,
				 CD_POINTER(plugin));
  return axel;
}

AXEL
_ax_register_base (AXEL axel, AxBase base)
{
  AxEnvironment env = axel->environment;
  env->bases = cd_list_append (axel, env->bases, CD_POINTER(base));
  return axel;
}

AXEL
_ax_unregister_base (AXEL axel, AxBase base)
{
  AxEnvironment env = axel->environment;
  if (env->purging) return axel;

  env->bases = cd_list_remove (axel, env->bases, CD_TYPE_POINTER,
			       CD_POINTER(base));
  return axel;
}

static int
plugin_id_eq (AXEL axel, AxPlugin plugin, char * identifier)
{
  if (plugin == AxNone) return FALSE;
  if (plugin->metatext == AxNone) return FALSE;
  return !(strcmp(plugin->metatext->identifier, identifier));
}

AxPlugin
ax_find_plugin (AXEL axel, char * identifier)
{
  AxEnvironment env = axel->environment;
  CDList l = cd_list_find_first (axel, env->plugins, CD_TYPE_POINTER,
				 (CDCmpFunc)plugin_id_eq,
				 CD_POINTER(identifier));

  if (l == AxNone) return AxNone;
  return (AxPlugin)l->data.s_pointer;
}
