/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxDeck: A high level audio mixing abstraction.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A deck contains a number of tracks which are mixed in parallel.
 *
 * Invariants
 * ----------
 *
 * Decks are independent entities. The creation of a new deck does not
 * depend on the existence of any other bases. Any tracks (and, by
 * extension, layers and sounds) added to a deck become part of that
 * deck and are destroyed when the deck is destroyed.
 *
 */

#define __AXEL__
#include "axel.h"

/* Optimisation dependencies: optimise on change of nr. tracks */
static AxDeck ax_deck_optimise (AXEL axel, AxDeck deck);

static AxDeck
ax_deck_replace_mixstream (AXEL axel, AxDeck deck)
{
  AxCount mixlength = _ax_base_get_mixlength (axel, deck);

  if (deck->_mixstream != AxNone)
    ax_destroy (axel, (AxBase)deck->_mixstream);

  deck->_mixstream = ax_stream_new_contiguous (axel, mixlength);

  return deck;
}

AxBase
ax_deck_init (AXEL axel, AxBase base)
{
  AxDeck deck = (AxDeck)base;
  deck->tracks = cd_list_new (axel);
  deck->_mixstream = AxNone;
  ax_deck_replace_mixstream (axel, deck);
  ax_deck_optimise (axel, deck);
  return (AxBase)deck;
}

AxDeck
ax_deck_new (AXEL axel)
{
  AxDeck deck = (AxDeck) ax_base_new_subclass (axel, sizeof (struct _AxDeck));
  return (AxDeck)ax_deck_init (axel, (AxBase)deck);
}

AxBase
ax_deck_clone (AXEL axel, AxBase base)
{
  AxDeck deck = (AxDeck)base;
  AxDeck new_deck = ax_deck_new (axel);
  new_deck->tracks = cd_list_clone (axel, deck->tracks,
				    (CDCloneFunc)ax_track_clone);
  return (AxBase)new_deck;
}

static int
ax_deck_destroy (AXEL axel, AxBase base)
{
  AxDeck deck = (AxDeck)base;
  ax_destroy_list (axel, deck->tracks);
  ax_destroy (axel, (AxBase)deck->_mixstream);
  ax_free (deck);
  return 0;
}

static int
ax_deck_ready (AXEL axel, AxBase base)
{
  return (ax_base_encompasses_mixlength (axel, base) &&
	  ax_base_encompasses_channels (axel, base));
}

static AxBase
ax_deck_prepare (AXEL axel, AxBase base)
{
  AxDeck deck = (AxDeck)base;
  ax_deck_replace_mixstream (axel, deck);
  return base;
}

AxTrack
_ax_deck_add_track (AXEL axel, AxDeck deck, AxTrack track)
{
  deck->tracks = cd_list_prepend (axel, deck->tracks, CD_POINTER(track));
  ax_deck_optimise (axel, deck);
  return track;
}

AxTrack
_ax_deck_remove_track (AXEL axel, AxDeck deck, AxTrack track)
{
  deck->tracks = cd_list_remove (axel, deck->tracks, CD_TYPE_POINTER,
				 CD_POINTER(track));
  ax_deck_optimise (axel, deck);
  return track;
}

int
ax_deck_nr_tracks (AXEL axel, AxDeck deck)
{
  return cd_list_length (axel, deck->tracks);
}

static AxCount
ax_deck_length (AXEL axel, AxBase base)
{
  AxDeck deck = (AxDeck)base;
  AxCount length, maxlength = 0;
  CDList l;
  AxBase track;
  
  for (l = deck->tracks; l; l = l->next) {
    track = (AxBase)l->data.s_pointer;
    length = ax_length (axel, track);
    ax_dprintf ("[ax_deck_length] found track %p length %ld\n", track, length);
    maxlength = MAX (maxlength, length);
  }

  return maxlength;
}

CDList
ax_deck_get_tracks (AXEL axel, AxDeck deck)
{
  return cd_list_copy (axel, deck->tracks);
}

#if 0
static AxCount
ax_deck_process (AXEL axel, AxBase base, AxCount count, AxStream input,
		 AxStream output)
{
  AxDeck deck = (AxDeck)base;
  CDList tl, ml;
  AxTrack track;
  AxStream mixstream;
  AxCount processed = 0, remaining = count, n, minn;
  AxCount input_offset;

  ax_dprintf ("PROCESS DECK (%p, +%ld, %p -> %p) @ %ld\n",
	      deck, count, input, output, ax_tell (axel, base));

  while (remaining > 0) {
    /* Get mixlength from each track */
    minn = deck->mixlength;
    input_offset = ax_tell (axel, (AxBase)input);
    for (tl = deck->tracks, ml = deck->mixstreams; tl && ml;
         tl = tl->next, ml = ml->next) {
      track = (AxTrack)tl->data;
      mixstream = (AxStream)ml->data;
      ax_seek (axel, (AxBase)input, input_offset);
      n = ax_process (axel, (AxBase)track, MIN (remaining, deck->mixlength),
		      input, mixstream);
      minn = MIN (minn, n);
    }


    if (output != AxNone) {
      /* mix tracks together to the output */
      n = ax_streams_mix (axel, deck->mixstreams, output, minn);
    } else {
      /* don't need to create output */
      n = minn;
    }

    processed += n;
    remaining -= n;
  }

  ax_dprintf ("[ax_deck_process] processed %ld\n", processed);

  return processed;
}
#endif

static AxCount
ax_deck_process (AXEL axel, AxBase base, AxCount count,
		 AxStream input, AxStream output)
{
  AxDeck deck = (AxDeck)base;
  CDList l;
  AxTrack track;
  AxCount remaining = count, processed = 0, n;
  AxCount current_offset = ax_tell (axel, base);
  AxCount input_offset = ax_tell (axel, (AxBase)input);
  AxCount output_offset = ax_tell (axel, (AxBase)output);
  AxCount mixlength = _ax_base_get_mixlength (axel, deck);
  AxStream mixstream = deck->_mixstream;

  ax_dprintf ("PROCESS DECK (%p, +%ld, %p -> %p) @ %ld\n",
	      deck, count, input, output, current_offset);

  while (remaining > 0) {

    l = deck->tracks;
    track = (AxTrack)l->data.s_pointer;

    n = MIN (remaining, mixlength);

    n = ax_process (axel, (AxBase)track, n, input, output);
    ax_seek (axel, (AxBase)output, output_offset, SEEK_SET);
    n = ax_stream_gain (axel, output, n, track->gain);

    for (l = l->next; l; l = l->next) {
      track = (AxTrack)l->data.s_pointer;
      
      ax_seek (axel, (AxBase)input, input_offset, SEEK_SET);
      ax_seek (axel, (AxBase)mixstream, 0, SEEK_SET);
      n = ax_process (axel, (AxBase)track, n, input, mixstream);
      
      ax_seek (axel, (AxBase)mixstream, 0, SEEK_SET);
      n = ax_stream_gain (axel, mixstream, n, track->gain);
      
      ax_seek (axel, (AxBase)mixstream, 0, SEEK_SET);
      ax_seek (axel, (AxBase)output, output_offset, SEEK_SET);
      n = ax_stream_mix (axel, mixstream, output, n);
    }

    processed += n;
    remaining -= n;
  }

  ax_dprintf ("[ax_deck_process] processed %ld\n", processed);

  return processed;
}

static AxCount
ax_deck_twotrack_process (AXEL axel, AxBase base, AxCount count,
			  AxStream input, AxStream output)
{
  AxDeck deck = (AxDeck)base;
  CDList l;
  AxTrack track1, track2;
  AxCount n;
  AxCount current_offset = ax_tell (axel, base);
  AxCount input_offset = ax_tell (axel, (AxBase)input);
  AxCount output_offset = ax_tell (axel, (AxBase)output);
  AxStream mixstream = deck->_mixstream;

  ax_dprintf ("PROCESS DECK [twotrack] (%p, +%ld, %p -> %p) @ %ld\n",
	      deck, count, input, output, current_offset);

  l = deck->tracks;
  track1 = (AxTrack)l->data.s_pointer;

  l = l->next;
  track2 = (AxTrack)l->data.s_pointer;

  n = ax_process (axel, (AxBase)track1, count, input, output);

  ax_seek (axel, (AxBase)output, output_offset, SEEK_SET);
  n = ax_stream_gain (axel, output, n, track1->gain);

  ax_seek (axel, (AxBase)input, input_offset, SEEK_SET);
  ax_seek (axel, (AxBase)mixstream, 0, SEEK_SET);
  n = ax_process (axel, (AxBase)track2, n, input, mixstream);

  ax_seek (axel, (AxBase)mixstream, 0, SEEK_SET);
  n = ax_stream_gain (axel, mixstream, n, track2->gain);

  ax_seek (axel, (AxBase)mixstream, 0, SEEK_SET);
  ax_seek (axel, (AxBase)output, output_offset, SEEK_SET);
  n = ax_stream_mix (axel, mixstream, output, n);

  ax_dprintf ("[ax_deck_twotrack_process] processed %ld\n", n);

  return n;
}

static AxCount
ax_deck_onetrack_process (AXEL axel, AxBase base, AxCount count,
			  AxStream input, AxStream output)
{
  AxDeck deck = (AxDeck)base;
  AxTrack track = (AxTrack)deck->tracks->data.s_pointer;
  AxCount n;

  ax_dprintf ("PROCESS DECK [onetrack] (%p, +%ld, %p -> %p) @ %ld\n",
	      deck, count, input, output, ax_tell (axel, base));

  n = ax_process (axel, (AxBase)track, count, input, output);

  ax_dprintf ("*** deck @ %ld\ttrack @ %ld\n", ax_tell (axel, base),
              ax_tell (axel, (AxBase)track));

  ax_dprintf ("[ax_deck_onetrack_process] processed %ld\n", n);

  return n;
}

static AxCount
ax_deck_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxDeck deck = (AxDeck)base;
  CDList l;
  AxTrack track;

  for (l = deck->tracks; l; l = l->next) {
    track = (AxTrack)l->data.s_pointer;
    ax_seek (axel, (AxBase)track, offset, SEEK_SET);
  }

  return offset;
}

static int
ax_deck_flush (AXEL axel, AxBase base)
{
  AxDeck deck = (AxDeck)base;
  CDList l;
  AxTrack track;

  for (l = deck->tracks; l; l = l->next) {
    track = (AxTrack)l->data.s_pointer;
    ax_flush (axel, (AxBase)track);
  }

  return 0;
}

static struct _AxMethods _ax_deck_empty_methods = {
  ax_deck_clone,   /* clone */
  ax_deck_destroy, /* destroy */
  ax_deck_ready,   /* ready */
  ax_deck_prepare, /* preapre */
  ax_null_process, /* process */
  ax_null_length,  /* length */
  NULL,            /* seek */
  NULL,            /* flush */
};

static struct _AxMethods _ax_deck_methods = {
  ax_deck_clone,   /* clone */
  ax_deck_destroy, /* destroy */
  ax_deck_ready,   /* ready */
  ax_deck_prepare, /* preapre */
  ax_deck_process, /* process */
  ax_deck_length,  /* length */
  ax_deck_seek,    /* seek */
  ax_deck_flush,   /* flush */
};

static struct _AxMethods _ax_deck_onetrack_methods = {
  ax_deck_clone,            /* clone */
  ax_deck_destroy,          /* destroy */
  ax_deck_ready,            /* ready */
  ax_deck_prepare,          /* preapre */
  ax_deck_onetrack_process, /* process */
  ax_deck_length,           /* length */
  ax_deck_seek,             /* seek */
  ax_deck_flush,   /* flush */
};

static struct _AxMethods _ax_deck_twotrack_methods = {
  ax_deck_clone,            /* clone */
  ax_deck_destroy,          /* destroy */
  ax_deck_ready,            /* ready */
  ax_deck_prepare,          /* preapre */
  ax_deck_twotrack_process, /* process */
  ax_deck_length,           /* length */
  ax_deck_seek,             /* seek */
  ax_deck_flush,   /* flush */
};

static AxDeck
ax_deck_optimise (AXEL axel, AxDeck deck)
{
  int nr_tracks = cd_list_length (axel, deck->tracks);

  switch (nr_tracks) {
  case 0: _ax_set_methods (axel, deck, &_ax_deck_empty_methods); break;
  case 1: _ax_set_methods (axel, deck, &_ax_deck_onetrack_methods); break;
  case 2: _ax_set_methods (axel, deck, &_ax_deck_twotrack_methods); break;
  default: _ax_set_methods (axel, deck, &_ax_deck_methods); break;
  }

  return deck;
}
