/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxEnvelope: Generic mathy envelopes for axel.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#define __AXEL__
#include "axel.h"

/* Optimisation dependencies: optimise when adding and removing points
 * or changing envelope type */
static AxEnvelope ax_envelope_optimise (AXEL axel, AxEnvelope envelope);

static void
ax_envelope_debug (AXEL axel, AxEnvelope envelope)
{
#ifdef DEBUG
  CDList l;
  AxPoint point;

  printf ("envelope %p\n", envelope);
  printf ("envelope->points: %p\n", envelope->points);

  /*  printf (" has %d points\n", cd_list_length (axel, envelope->points));*/
  for (l = envelope->points; l; l = l->next) {
    point = (AxPoint)l->data.s_pointer;
    switch (envelope->timetype) {
    case AX_TIME_SAMPLES:
      printf ("%ld samples, %f\n", point->time.samples, point->value);
      break;
    case AX_TIME_SECONDS:
      printf ("%f seconds, %f\n", point->time.seconds, point->value);
      break;
    case AX_TIME_BEAT24S:
      printf ("%d beat24s, %f\n", point->time.beat24s, point->value);
      break;
    default:
      printf ("*** unknown envelope->timetype ***\n");
      break;
    }
  } 
#endif
}

static AxPoint
ax_point_new (AxTime time, AxPCM value)
{
  AxPoint point = ax_malloc (sizeof (struct _AxPoint));
  point->time = time;
  point->value = value;
  return point;
}

static AxPoint
ax_point_clone (AXEL axel, AxPoint point)
{
  return ax_point_new (point->time, point->value);
}

static int
ax_point_later_X (AXEL axel, AxPoint p1, AxPoint p2)
{
  return _ax_time_gt(AX_TIME_SAMPLES, p1->time, p2->time);
}

static int
ax_point_later_S (AXEL axel, AxPoint p1, AxPoint p2)
{
  return _ax_time_gt(AX_TIME_SECONDS, p1->time, p2->time);
}

static int
ax_point_later_B (AXEL axel, AxPoint p1, AxPoint p2)
{
  return _ax_time_gt(AX_TIME_BEAT24S, p1->time, p2->time);
}

AxBase
ax_envelope_init (AXEL axel, AxBase base)
{
  AxEnvelope envelope = (AxEnvelope)base;
  envelope->type = AX_ENVELOPE_LINEAR;
  envelope->points = cd_list_new (axel);
  ax_envelope_optimise (axel, envelope);
  return (AxBase)envelope;
}

AxEnvelope
ax_envelope_new (AXEL axel, AxEnvelopeType type)
{
  AxEnvelope envelope =
    (AxEnvelope)ax_base_new_subclass (axel, sizeof (struct _AxEnvelope));
  ax_envelope_init (axel, (AxBase)envelope);
  envelope->type = type;
  ax_envelope_debug (axel, envelope);
  return envelope;
}

AxBase
ax_envelope_clone (AXEL axel, AxBase base)
{
  AxEnvelope envelope = (AxEnvelope)base;
  AxEnvelope new_envelope = ax_envelope_new (axel, envelope->type);
  new_envelope->points = cd_list_clone (axel, envelope->points,
					(CDCloneFunc)ax_point_clone);
  ax_envelope_optimise (axel, new_envelope);
  return (AxBase)new_envelope;
}

static int
ax_envelope_destroy (AXEL axel, AxBase base)
{
  AxEnvelope envelope = (AxEnvelope)base;
  cd_list_free_all (axel, envelope->points);
  ax_free (envelope);
  return 0;
}

AxEnvelopeType
ax_envelope_set_type (AXEL axel, AxEnvelope envelope, AxEnvelopeType type)
{
  AxEnvelopeType old = envelope->type;
  envelope->type = type;
  ax_envelope_optimise (axel, envelope);
  return old;
}

AxEnvelopeType
ax_envelope_get_type (AXEL axel, AxEnvelope envelope)
{
  return envelope->type;
}

AxTimeType
ax_envelope_set_timetype (AXEL axel, AxEnvelope envelope,
			  AxTimeType timetype)
{
  AxTimeType old = envelope->timetype;
  envelope->timetype = timetype;
  return old;
}

AxTimeType
ax_envelope_get_timetype (AXEL axel, AxEnvelope envelope)
{
  return envelope->timetype;
}

AxTime
ax_envelope_get_duration (AXEL axel, AxEnvelope envelope)
{
  CDList l = cd_list_last_item (axel, envelope->points);
  AxPoint p;

  if (l == AxNone) return _ax_time_zero (envelope->timetype);

  p = (AxPoint)l->data.s_pointer;
  if (p == AxNone) return _ax_time_zero (envelope->timetype);

  return p->time;
}

AxPoint
ax_envelope_add_point (AXEL axel, AxEnvelope envelope, AxTime time,
		       AxPCM value)
{
  AxPoint point = ax_point_new (time, value);
  switch (envelope->timetype) {
  case AX_TIME_SAMPLES:
    envelope->points = cd_list_insert (axel, envelope->points,
				       CD_TYPE_POINTER, CD_POINTER(point),
				       (CDCmpFunc)ax_point_later_X);
    break;
  case AX_TIME_SECONDS:
    envelope->points = cd_list_insert (axel, envelope->points,
				       CD_TYPE_POINTER, CD_POINTER(point),
				       (CDCmpFunc)ax_point_later_S);
    break;
  case AX_TIME_BEAT24S:
    envelope->points = cd_list_insert (axel, envelope->points,
				       CD_TYPE_POINTER, CD_POINTER(point),
				       (CDCmpFunc)ax_point_later_B);
    break;
  default: /* uncommon, we should hope */
    ax_free (point); point = AxNone;
    break;
  }
  ax_envelope_debug (axel, envelope);
  ax_envelope_optimise (axel, envelope);
  return point;
}

AxEnvelope
ax_envelope_remove_point (AXEL axel, AxEnvelope envelope, AxPoint point)
{
  envelope->points = cd_list_remove (axel, envelope->points,
				     CD_TYPE_POINTER, CD_POINTER(point));
  ax_free (point);
  ax_envelope_debug (axel, envelope);
  ax_envelope_optimise (axel, envelope);
  return envelope;
}

AxEnvelope
ax_envelope_scale (AXEL axel, AxEnvelope envelope, AxPCM gain)
{
  CDList l;
  AxPoint p;

  for (l = envelope->points; l; l = l->next) {
    p = (AxPoint)l->data.s_pointer;
    p->value *= gain;
  }

  return envelope;
}

AxEnvelope
ax_envelope_shift (AXEL axel, AxEnvelope envelope, AxTime delta)
{
  CDList l;
  AxPoint p;

  for (l = envelope->points; l; l = l->next) {
    p = (AxPoint)l->data.s_pointer;
    p->time = _ax_time_add (envelope->timetype, p->time, delta);
  }

  return envelope;
}

static CDList
ax_envelope_point_item_before (AXEL axel, AxEnvelope envelope, AxCount offset)
{
  CDList l, lp = AxNone;
  AxPoint point;
  AxTime ptime;

  for (l = envelope->points; l; l = l->next) {
    point = (AxPoint)l->data.s_pointer;
    ptime = ax_time_convert (axel, point->time, envelope->timetype,
			     AX_TIME_SAMPLES);
    if (ptime.samples > offset) break;
    lp = l;
  }
  return lp;
}

static AxCount
ax_envelope_constant_write_chunk (AXEL axel, AxChunk chunk, AxCount offset,
				  AxCount count, int channelname, void * data)
{
  AxEnvelope envelope = (AxEnvelope)data;
  AxPoint point;
  AxPCM value;
  AxPCM * d;
  AxCount n;

  point = (AxPoint)envelope->points->data.s_pointer;
  value = point->value;
  d = &chunk->data[offset];

  n = _ax_pcm_set (d, value, count);
  envelope->_current_offset += n;
  return n;
}

static AxCount
ax_envelope_spline_write_chunk (AXEL axel, AxChunk chunk, AxCount offset,
				AxCount count, int channelname, void * data)
{
  AxEnvelope envelope = (AxEnvelope)data;
  /* XXX: Implement ;) */
  return -1;
}

/* An AxChunkFunc for creating envelope data */
static AxCount
ax_envelope_linear_write_chunk (AXEL axel, AxChunk chunk, AxCount offset,
				AxCount count, int channelname, void * data)
{
  AxEnvelope envelope = (AxEnvelope)data;
  AxCount remaining = count, written = 0;
  AxCount pos = envelope->_current_offset;
  CDList l, nl;
  AxPoint point, next_point;
  AxCount px, npx, n;
  AxTime t;
  AxPCM py, npy, gradient;
  AxPCM * d;

  ax_dprintf ("[ax_envelope_linear_write_chunk] (%ld, +%ld) @ %ld\n",
	  offset, count, pos);


  l = envelope->_current_point_item;

  if (l == AxNone) {/* No points before start */
    l = envelope->points;
    if (l == AxNone) {/* No points at all */
      n = _ax_chunk_clear_region (axel, chunk, offset, count, 0, NULL);
      envelope->_current_offset += n;
      return n;
    }
  }

  nl = l->next;
  if (nl == AxNone) {
    /* if the last point was before offset, and there were
     * more points, set l to the second last and nl to the last */
    nl = l; l = l->prev;
    if (l == AxNone) {/* Constant envelope (one point) */
      return ax_envelope_constant_write_chunk (axel, chunk, offset, count,
					       channelname, envelope);
    }
  }

  point = (AxPoint)l->data.s_pointer;
  t = ax_time_convert (axel, point->time, envelope->timetype, AX_TIME_SAMPLES);
  px = t.samples;
  py = point->value;
  
  next_point = (AxPoint)nl->data.s_pointer;
  t = ax_time_convert (axel, next_point->time, envelope->timetype,
		       AX_TIME_SAMPLES);
  npx = t.samples;
  npy = next_point->value;

  while (remaining > 0) {
    if (nl->next == AxNone) {
      /* These are the last two points, so fill out with this gradient */
      n = remaining;
    } else {
      n = MIN (remaining, npx - pos);
    }
    gradient = (npy - py) / (AxPCM)(npx - px);
    
    d = &chunk->data[offset];
    /*    _ax_pcm_write_linear (d, px - chunk->start_index, py, gradient, n);*/
    n = _ax_pcm_write_linear (d, px, py, npx, npy, pos, n);
    
    remaining -= n;
    written += n;
    pos += n;
    offset += n;
    
    if (remaining > 0) {
      l = nl; point = next_point; px = npx; py = npy;
      
      nl = nl->next;
      next_point = (AxPoint)nl->data.s_pointer;
      t = ax_time_convert (axel, next_point->time, envelope->timetype,
			   AX_TIME_SAMPLES);
      npx = t.samples;
      npy = next_point->value;
    }
  }

  envelope->_current_point_item = l;
  envelope->_current_offset = pos;

  return written;
}

static AxCount
ax_envelope_constant_process (AXEL axel, AxBase base, AxCount count,
			      AxStream input, AxStream output)
{
  AxEnvelope envelope = (AxEnvelope)base;
  return ax_stream_chunkfuncify (axel, output, count,
				 ax_envelope_constant_write_chunk, envelope);
}

static AxCount
ax_envelope_spline_process (AXEL axel, AxBase base, AxCount count,
			    AxStream input, AxStream output)
{
  return -1;
}

static AxCount
ax_envelope_linear_process (AXEL axel, AxBase base, AxCount count,
			    AxStream input, AxStream output)
{
  AxEnvelope envelope = (AxEnvelope)base;
  return ax_stream_chunkfuncify (axel, output, count,
				 ax_envelope_linear_write_chunk, envelope);
}

static AxCount
ax_envelope_process (AXEL axel, AxBase base, AxCount count,
		     AxStream input, AxStream output)
{
  AxEnvelope envelope = (AxEnvelope)base;

  switch (envelope->type) {
  case AX_ENVELOPE_LINEAR:
    return ax_envelope_linear_process (axel, base, count, input, output);
    break;
  case AX_ENVELOPE_SPLINE:
    return ax_envelope_spline_process (axel, base, count, input, output);
    break;
  default:
    break;
  }
  ax_set_error (axel, AX_ERROR_NOOP);
  return 0;
}

static AxCount
ax_envelope_length (AXEL axel, AxBase base)
{
  AxEnvelope envelope = (AxEnvelope)base;
  AxTime duration = ax_envelope_get_duration (axel, envelope);
  AxTime t = ax_time_convert (axel, duration, envelope->timetype,
			      AX_TIME_SAMPLES);
  return t.samples;
}

static AxCount
ax_envelope_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxEnvelope envelope = (AxEnvelope)base;
  envelope->_current_point_item =
    ax_envelope_point_item_before (axel, envelope, offset);
  envelope->_current_offset = offset;
  return offset;
}

static struct _AxMethods _ax_envelope_empty_methods = {
  ax_envelope_clone,
  ax_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_null_process,
  ax_null_length,
  ax_envelope_seek,
  NULL, /* flush */
};

static struct _AxMethods _ax_envelope_constant_methods = {
  ax_envelope_clone,
  ax_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_envelope_constant_process,
  ax_envelope_length,
  ax_envelope_seek,
  NULL, /* flush */
};

static struct _AxMethods _ax_envelope_linear_methods = {
  ax_envelope_clone,
  ax_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_envelope_linear_process,
  ax_envelope_length,
  ax_envelope_seek,
  NULL, /* flush */
};

static struct _AxMethods _ax_envelope_spline_methods = {
  ax_envelope_clone,
  ax_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_envelope_spline_process,
  ax_envelope_length,
  ax_envelope_seek,
  NULL, /* flush */
};

static struct _AxMethods _ax_envelope_methods = {
  ax_envelope_clone,
  ax_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_envelope_process,
  ax_envelope_length,
  ax_envelope_seek,
  NULL, /* flush */
};

static AxEnvelope
ax_envelope_optimise (AXEL axel, AxEnvelope envelope)
{
  if (cd_list_is_empty (axel, envelope->points)) {
    _ax_set_methods (axel, envelope, &_ax_envelope_empty_methods);
  } else if (cd_list_is_singleton (axel, envelope->points)) {
    _ax_set_methods (axel, envelope, &_ax_envelope_constant_methods);
  } else {
    switch (envelope->type) {
    case AX_ENVELOPE_LINEAR:
      _ax_set_methods (axel, envelope, &_ax_envelope_linear_methods); break;
    case AX_ENVELOPE_SPLINE:
      _ax_set_methods (axel, envelope, &_ax_envelope_spline_methods); break;
    default:
      _ax_set_methods (axel, envelope, &_ax_envelope_methods); break;
    }
  }

  return envelope;
}
