/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxGain: a gain filter
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#define __AXEL_PLUGIN__
#include "axel.h"

#define GAIN_ENVELOPE_KEY 1

typedef struct _AxGain * AxGain;

struct _AxGain {
  AxStream _gain_envstream;
};

/* Optimisation dependencies: none */
static AxBase ax_gain_optimise (AXEL axel, AxBase gain);

static void
ax_gain_replace_mixstream (AXEL axel, AxBase gain)
{
  AxCount mixlength = ax_base_get_mixlength (axel, gain);
  AxGain gi = (AxGain)ax_base_get_instance_data (axel, gain);

  if (gi->_gain_envstream != AxNone)
    ax_destroy (axel, (AxBase)gi->_gain_envstream);

  gi->_gain_envstream =
    ax_stream_new_contiguous (axel, mixlength);
}

static AxBase
ax_gain_init (AXEL axel, AxBase base, CDSet parameters)
{
  ax_base_set_instance_data (axel, base, ss_malloc (sizeof (struct _AxGain)));
  ax_gain_replace_mixstream (axel, base);
  ax_gain_optimise (axel, base);
  return base;
}

static AxBase
ax_gain_clone (AXEL axel, AxBase base)
{
  AxBase new_gain = ax_base_new (axel);
  ax_gain_init (axel, new_gain, CD_EMPTY_SET);
  ax_gain_optimise (axel, new_gain);
  return (AxBase)new_gain;
}

static int
ax_gain_destroy (AXEL axel, AxBase base)
{
  ss_free (ax_base_get_instance_data (axel, base));
  ss_free (base);
  return 0;
}

static int
ax_gain_ready (AXEL axel, AxBase base)
{
  return (ax_base_encompasses_mixlength (axel, base) &&
	  ax_base_encompasses_channels (axel, base));
}

static AxBase
ax_gain_prepare (AXEL axel, AxBase base)
{
  ax_gain_replace_mixstream (axel, base);
  return base;
}

static AxCount
ax_gain_process (AXEL axel, AxBase base, AxCount count,
		 AxStream input, AxStream output)
{
  AxCount remaining = count, processed = 0, n;
  AxCount output_offset;
  AxCount mixlength = ax_base_get_mixlength (axel, base);
  AxBase gain_envelope;
  AxGain gi = ax_base_get_instance_data (axel, base);

  ax_dprintf ("PROCESS GAIN (%p, +%ld) @ %ld\n", base, count,
	      ax_tell (axel, base));

  gain_envelope =
    (AxBase) (ax_get_parameter (axel, base, GAIN_ENVELOPE_KEY)).s_pointer;

  if (gain_envelope == AxNone) {
    ax_set_error (axel, AX_ERROR_NOOP);
    return -1;
  }

  while (remaining > 0) {
    n = MIN (remaining, mixlength);

    output_offset = ax_tell (axel, (AxBase)output);
    n = ax_stream_copy (axel, input, output, n);

    ax_seek (axel, (AxBase)gi->_gain_envstream, 0, SEEK_SET);
    n = ax_process (axel, gain_envelope, n, AxNone,
		    gi->_gain_envstream);

    ax_seek (axel, (AxBase)gi->_gain_envstream, 0, SEEK_SET);
    ax_seek (axel, (AxBase)output, output_offset, SEEK_SET);
    n = ax_stream_mult (axel, gi->_gain_envstream, output, n);

    remaining -= n;
    processed += n;
  }

  ax_dprintf ("[ax_gain_process] processed %ld\n", processed);

  return processed;
}

static AxCount
ax_gain_length (AXEL axel, AxBase base)
{
  AxBase gain_envelope =
    (AxBase) (ax_get_parameter (axel, base, GAIN_ENVELOPE_KEY)).s_pointer;

  if (gain_envelope == AxNone) {
    return AX_COUNT_INFINITE;
  }

  return ax_length (axel, gain_envelope);
}

static AxCount
ax_gain_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxBase gain_envelope =
    (AxBase) (ax_get_parameter (axel, base, GAIN_ENVELOPE_KEY)).s_pointer;

  return ax_seek (axel, (AxBase)gain_envelope, offset, SEEK_SET);
}

static struct _AxMethods _ax_gain_methods = {
  ax_gain_clone,
  ax_gain_destroy,
  ax_gain_ready,
  ax_gain_prepare,
  ax_gain_process,
  ax_gain_length,
  ax_gain_seek,
};

static AxBase
ax_gain_optimise (AXEL axel, AxBase gain)
{
  ax_base_set_methods (axel, gain, &_ax_gain_methods);
  return gain;
}

static struct _AxParameterScheme gain_envelope_scheme = {
  "Gain envelope",
  "An envelope to control the amplitude",
  AX_TYPE_BASE,
  AX_CONSTRAINT_TYPE_NONE,
  AX_CONSTRAINT_EMPTY,
  AX_HINT_DEFAULT,
};

static struct _AxMetaText gain_metatext = {
  "builtin::gain",
  "Processors::Gain Adjustment",
  "Adjusts the gain of its input",
  "Copyright (C) 2001 CSIRO Australia",
  "http://www.metadecks.org/axel/plugins/gain.html",
  AX_ONE_AUTHOR ("Conrad Parker", "Conrad.Parker@CSIRO.AU"),
};

static struct _AxPlugin gain_plugin = {
  &gain_metatext,
  AX_FLAGS_NONE,
  CD_EMPTY_SET, /* new scheme */
  ax_gain_init,
  CD_EMPTY_SET, /* process_scheme */
  NULL, /* suggests */
  NULL, /* plugin_data */
};

/* module init function */
CDList
__gain_init (AXEL axel)
{
  CDList plugins = cd_list_new (axel);

  gain_plugin.process_scheme =
    cd_set_insert (axel, gain_plugin.process_scheme, GAIN_ENVELOPE_KEY,
		   CD_POINTER(&gain_envelope_scheme));

  plugins = cd_list_prepend (axel, plugins, CD_POINTER(&gain_plugin));

  return plugins;
}
