/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxLayer: A sound sequence abstraction.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A layer is contained within a track. It consists of a sequence of
 * sounds. A layer has a time type (AxTimeType) with which its sounds
 * are indexed.
 * XXX: Overlapping sounds in a layer should be mixed together.
 *
 * Invariants
 * ----------
 *
 * A layer must be contained in a track.
 *
 */

#define __AXEL__
#include "axel.h"

/* Optimisation dependencies: none */
static AxLayer ax_layer_optimise (AXEL axel, AxLayer layer);

/* Coherency dependencies: ensure coherency on addition+removal of sounds */
static AxLayer ax_layer_ensure_coherency (AXEL axel, AxLayer layer);

void
ax_layer_debug (AXEL axel, AxLayer layer)
{
#ifdef DEBUG
  CDList l;
  AxSound s;

  ax_dprintf ("Layer (0x%p): ", layer);
  if (layer == AxNone) return;

  for (l = layer->sounds; l; l = l->next) {
    s = (AxSound)l->data.s_pointer;
    /* XXX: assumes samples */
    ax_dprintf ("[0x%p: %ld, +%ld] ", s, s->start_time.samples,
		s->duration.samples);
  }
  ax_dprintf ("\n");
#endif
}

static AxBase
ax_layer_init (AXEL axel, AxBase base)
{
  AxLayer layer = (AxLayer)base;
  layer->timetype = AX_TIME_SAMPLES;
  layer->sounds = cd_list_new (axel);
  /*  layer->_current_time = _ax_time_zero (layer->timetype);*/
  layer->_current_sound_item = AxNone;
  ax_layer_optimise (axel, layer);
  return (AxBase)layer;
}

static AxBase
ax_new_layer (AXEL axel)
{
  return (AxBase) ax_base_new_subclass (axel, sizeof (struct _AxLayer));
}

AxBase
ax_layer_clone (AXEL axel, AxBase base)
{
  AxLayer layer = (AxLayer)base;
  AxLayer new_layer = (AxLayer)ax_new_layer (axel);
  AxCount offset = ax_tell (axel, base);

  new_layer->timetype = layer->timetype;
  new_layer->sounds =
    cd_list_clone_with (axel, layer->sounds,
			(CDCloneWithFunc)ax_sound_clone_with_layer, new_layer);
  ax_seek (axel, (AxBase)new_layer, offset, SEEK_SET);

  new_layer->track = layer->track;
  _ax_track_add_layer_above (axel, layer->track, new_layer, layer);

  return (AxBase)new_layer;
}

static int
ax_layer_destroy (AXEL axel, AxBase base)
{
  AxLayer layer = (AxLayer)base;
  if (layer->track)
    _ax_track_remove_layer (axel, layer->track, layer);
  ax_destroy_list (axel, layer->sounds);
  ax_free (layer);
  return 0;
}

AxLayer
ax_add_layer_ontop (AXEL axel, AxTrack track, AxTimeType timetype)
{
  AxLayer layer = (AxLayer)ax_new_layer (axel);

  layer->track = track;
  _ax_track_add_layer_above (axel, track, layer, AxNone);
  ax_layer_init (axel, (AxBase)layer);
  layer->timetype = timetype;

  return layer;
}

AxLayer
ax_add_layer_above (AXEL axel, AxLayer above, AxTimeType timetype)
{
  AxLayer layer;
  if (above == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  layer = (AxLayer)ax_new_layer (axel);
  layer->track = above->track;
  _ax_track_add_layer_above (axel, above->track, layer, above);
  ax_layer_init (axel, (AxBase)layer);
  layer->timetype = timetype;

  return layer;
}

AxLayer
ax_move_layer_ontop (AXEL axel, AxLayer layer, AxTrack track)
{
  if (layer->track)
    _ax_track_remove_layer (axel, layer->track, layer);

  if (track)
    _ax_track_add_layer_above (axel, track, layer, AxNone);

  return layer;
}

AxLayer
ax_move_layer_above (AXEL axel, AxLayer layer, AxLayer above)
{
  if (layer->track)
    _ax_track_remove_layer (axel, layer->track, layer);

  if (above->track)
    _ax_track_add_layer_above (axel, above->track, layer, above);

  return layer;
}

AxLayer
ax_raise_layer (AXEL axel, AxLayer layer)
{
  AxLayer above;

  if (layer->track) {
    above = _ax_track_get_layer_above (axel, layer->track, layer);
    ax_move_layer_above (axel, layer, above);
  }

  return layer;
}

AxLayer
ax_lower_layer (AXEL axel, AxLayer layer)
{
  AxLayer below;

  if (layer->track) {
    below = _ax_track_get_layer_below (axel, layer->track, layer);
    ax_move_layer_above (axel, below, layer);
  }

  return layer;
}

AxLayer
_ax_remove_layer (AXEL axel, AxLayer layer)
{
  if (layer->track)
    _ax_track_remove_layer (axel, layer->track, layer);

  layer->track = AxNone;
  return layer;
}

AxTrack
ax_layer_get_track (AXEL axel, AxLayer layer)
{
  return layer->track;
}

AxDeck
ax_layer_get_deck (AXEL axel, AxLayer layer)
{
  AxTrack track = layer->track;
  if (track == AxNone) return AxNone;
  return ax_track_get_deck (axel, track);
}

AxTimeType
ax_layer_set_timetype (AXEL axel, AxLayer layer, AxTimeType new_type)
{
  AxTimeType old_type = layer->timetype;
  CDList l;
  AxDeck deck;
  AxSound sound;

  if (old_type == new_type) return old_type;

  deck = ax_layer_get_deck (axel, layer);

  for (l = layer->sounds; l; l = l->next) {
    sound = (AxSound)l->data.s_pointer;
    sound->start_time = ax_time_convert (axel, sound->start_time, old_type,
                                         new_type);
    sound->duration = ax_time_convert (axel, sound->duration, old_type,
                                       new_type);
  }

  layer->timetype = new_type;

  return old_type;
}

AxTimeType
ax_layer_get_timetype (AXEL axel, AxLayer layer)
{
  return layer->timetype;
}

AxSound
_ax_layer_add_sound (AXEL axel, AxLayer layer, AxSound sound,
		     AxTime start_time)
{
  sound->start_time = start_time;
  layer->sounds = cd_list_insert (axel, layer->sounds, CD_TYPE_POINTER,
				  CD_POINTER(sound),
				  (CDCmpFunc)ax_sound_later);
  ax_layer_ensure_coherency (axel, layer);
  return sound;
}

AxSound
_ax_layer_remove_sound (AXEL axel, AxLayer layer, AxSound sound)
{
  layer->sounds = cd_list_remove (axel, layer->sounds, CD_TYPE_POINTER,
				  CD_POINTER(sound));
  ax_layer_ensure_coherency (axel, layer);
  return sound;
}

/*
 * ax_layer_get_sound_item_before (layer, time)
 *
 * Finds the last sound with a start_time before 'time'.
 */
static CDList
ax_layer_get_sound_item_before (AXEL axel, AxLayer layer, AxTime time)
{
  AxSound s;
  CDList l, lp = AxNone;

  for (l = layer->sounds; l; l = l->next) {
    s = (AxSound)l->data.s_pointer;
    if (_ax_time_gt (layer->timetype, s->start_time, time)) return lp;
    lp = l;
  }
  
  return lp;
}

/*
 * ax_layer_get_sound_item_at (layer, time)
 *
 * Finds the sound occurring at 'time'. If no sound is playing at 'time',
 * returns AxNone.
 */
CDList
ax_layer_get_sound_item_at (AXEL axel, AxLayer layer, AxTime time)
{
  AxTime t;
  CDList l = ax_layer_get_sound_item_before (axel, layer, time);
  AxSound s;

  if (l == AxNone) return AxNone;

  s = (AxSound)l->data.s_pointer;

  t = _ax_time_add (layer->timetype, s->start_time, s->duration);
  if (_ax_time_le (layer->timetype, t, time)) l = AxNone;

  return l;
}

/*
 * ax_layer_get_sound_item_after (layer, time)
 *
 * Finds the first sound at or after 'time' and returns its list item.
 */
static CDList
ax_layer_get_sound_item_after (AXEL axel, AxLayer layer, AxTime time)
{
  AxSound s;
  CDList l;

  for (l = layer->sounds; l; l = l->next) {
    s = (AxSound)l->data.s_pointer;
    if (_ax_time_ge (layer->timetype, s->start_time, time)) return l;
  }

  return AxNone;
}

/*
 * ax_layer_get_sound_after (layer, time)
 *
 * Finds the first sound with a start_time at or after 'time'.
 */
AxSound
ax_layer_get_sound_after (AXEL axel, AxLayer layer, AxTime time)
{
  CDList l = ax_layer_get_sound_item_after (axel, layer, time);

  if (l == AxNone) return AxNone;
  else return (AxSound)l->data.s_pointer;
}

AxSound
_ax_layer_get_sound_prev (AXEL axel, AxLayer layer, AxSound sound)
{
  CDList sound_item;
  AxSound sn;

  if (layer->sounds == NULL) return AxNone;

  if (sound == AxNone) {
    sn = (AxSound) layer->sounds->data.s_pointer;
  } else {
    sound_item = cd_list_find (axel, layer->sounds, CD_TYPE_POINTER,
			       CD_POINTER(sound));
    if (sound_item == NULL) return AxNone;
    sound_item = sound_item->prev;
    if (sound_item == NULL) return AxNone;
    sn = (AxSound) sound_item->data.s_pointer;
  }

  return sn;
}

AxSound
_ax_layer_get_sound_next (AXEL axel, AxLayer layer, AxSound sound)
{
  CDList sound_item;
  AxSound sn;

  if (layer->sounds == NULL) return AxNone;

  if (sound == NULL) {
    sn =
      (AxSound)(cd_list_last (axel, layer->sounds, CD_TYPE_POINTER)).s_pointer;
  } else {
    sound_item = cd_list_find (axel, layer->sounds, CD_TYPE_POINTER,
			       CD_POINTER(sound));
    if (sound_item == NULL) return AxNone;
    sound_item = sound_item->next;
    if (sound_item == NULL) return AxNone;
    sn = (AxSound) sound_item->data.s_pointer;
  }

  return sn;
}

AxLayer
ax_get_layer_below (AXEL axel, AxLayer layer)
{
  if (layer->track)
    return _ax_track_get_layer_below (axel, layer->track, layer);
  else
    return AxNone;
}

AxLayer
ax_get_layer_above (AXEL axel, AxLayer layer)
{
  if (layer->track)
    return _ax_track_get_layer_above (axel, layer->track, layer);
  else
    return AxNone;
}

static AxCount
ax_layer_length (AXEL axel, AxBase base)
{
  AxLayer layer = (AxLayer)base;
  AxSound sound =
    (AxSound)(cd_list_last (axel, layer->sounds, CD_TYPE_POINTER)).s_pointer;
  AxTime end, t;

  if (sound == AxNone) {
    ax_dprintf ("[ax_layer_length] layer %p has no sounds\n", layer);
    return 0;
  }

  /* Convert sound's end time to offset and return that */
  end = _ax_time_add (layer->timetype, sound->start_time, sound->duration);
  t = ax_time_convert (axel, end, layer->timetype, AX_TIME_SAMPLES);

  ax_dprintf ("[ax_layer_length] (%p) last sound ends at %d ticks == %ld samples\n",
	      layer, end.beat24s, t.samples);

  return t.samples;
}

static AxCount
ax_layer_process_sound (AXEL axel, AxLayer layer, AxCount current_offset,
			AxSound sound,
			AxCount sound_offset, AxCount sound_length,
			AxCount count, AxStream input, AxStream output)
{
  AxCount remaining = count, processed = 0, n;

  /* If the next sound starts after the current offset, fill up to it */
  if (sound_offset > current_offset) {
    n = MIN (remaining, sound_offset - current_offset);

    ax_dprintf ("[ax_layer_process_sound] %p is after offset, filling %ld\n",
		sound, n);

    if (output != AxNone)
      n = ax_stream_write (axel, output, n, input);
    current_offset += n;
    processed += n;
    remaining -= n;
  }

  /* If that didn't fill the output, process part of the next sound */
  if (processed < count) {
    n = MIN (remaining, sound_offset + sound_length - current_offset);
    /* XXX: fix following line to use SEEK_CUR ??? */
    ax_seek (axel, (AxBase)sound, current_offset - sound_offset, SEEK_SET);
    n = ax_process (axel, (AxBase)sound, n, input, output);
    processed += n;
    remaining -= n;
  }

  return processed;
}

static AxCount
ax_layer_process (AXEL axel, AxBase base, AxCount count, AxStream input,
                  AxStream output)
{
  AxLayer layer = (AxLayer)base;
  AxCount processed = 0, remaining = count, n;
  AxCount sound_offset, sound_length, next_offset;
  AxCount current_offset = ax_tell (axel, (AxBase)layer);
  AxSound sound, sn;
  AxTime t;

  ax_dprintf ("PROCESS LAYER (%p, +%ld, %p -> %p) @ %ld\n",
	      layer, count, input, output, current_offset);

  while (remaining > 0) {

    if (layer->_current_sound_item == AxNone) {
      /* No more sounds */
      ax_dprintf ("[ax_layer_process] ## no more sounds!\n");

      n = (output == AxNone) ? remaining :
	ax_stream_write (axel, output, remaining, input);
      current_offset += n;
      processed += n;
      remaining -= n;
      break;
    }

    sound = (AxSound)layer->_current_sound_item->data.s_pointer;
    t = ax_time_convert (axel, sound->start_time, layer->timetype,
			 AX_TIME_SAMPLES);
    sound_offset = t.samples;

    t = ax_time_convert (axel, sound->duration, layer->timetype,
			 AX_TIME_SAMPLES);
    sound_length = t.samples;

    if (layer->_current_sound_item->next) {
      sn = (AxSound)(layer->_current_sound_item->next->data.s_pointer);
      t = ax_time_convert (axel, sn->start_time, layer->timetype,
			   AX_TIME_SAMPLES);
      next_offset = t.samples;
      if (next_offset < sound_offset + sound_length)
        sound_length = next_offset - sound_offset;
    }

    /* *** We now have the next sound and its valid length *** */
    ax_dprintf ("[ax_layer_process] to process sound %p, [%ld, +%ld]\n",
		sound, sound_offset, sound_length);

    n = ax_layer_process_sound (axel, layer, current_offset,
				sound, sound_offset, sound_length,
				remaining, input, output);
    current_offset += n;
    processed += n;
    remaining -= n;

    if (current_offset >= sound_offset + sound_length)
      layer->_current_sound_item = layer->_current_sound_item->next;
  }

  ax_dprintf ("[ax_layer_process] processed %ld\n", processed);

  if (processed == 0) {
    ax_set_error (axel, AX_ERROR_NOOP);
    return -1;
  }

  return processed;
}

static AxCount
ax_layer_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxLayer layer = (AxLayer)base;
  AxTime current_time;

  /* Evaluate this offset as a time value using the current samplerate and
   * tempo */
  current_time = ax_time_convert (axel, (AxTime)offset, AX_TIME_SAMPLES,
				  layer->timetype);

  /* Cache the current sound item */
  layer->_current_sound_item =
    ax_layer_get_sound_item_at (axel, layer, current_time);

  if (layer->_current_sound_item == AxNone)
    layer->_current_sound_item =
      ax_layer_get_sound_item_after (axel, layer, current_time);

  return offset;
}

static int
ax_layer_flush (AXEL axel, AxBase base)
{
  AxLayer layer = (AxLayer)base;
  AxBase sound;

  if (layer->_current_sound_item == AxNone) return 0;

  sound = (AxBase)layer->_current_sound_item->data.s_pointer;

  return ax_flush (axel, sound);
}

static AxLayer
ax_layer_ensure_coherency (AXEL axel, AxLayer layer)
{
  AxCount offset = ax_tell (axel, (AxBase)layer);

  ax_layer_seek (axel, (AxBase)layer, offset);

  return layer;
}

static struct _AxMethods _ax_layer_methods = {
  ax_layer_clone,   /* clone */
  ax_layer_destroy, /* destroy */
  NULL,             /* ready */
  NULL,             /* prepare */
  ax_layer_process, /* process */
  ax_layer_length,  /* length */
  ax_layer_seek,    /* seek */
  ax_layer_flush,   /* flush */
};

static AxLayer
ax_layer_optimise (AXEL axel, AxLayer layer)
{
  _ax_set_methods (axel, layer, &_ax_layer_methods);
  return layer;
}
		  
