/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxMonitor: device output
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <sys/types.h>

#define __AXEL__
#include "axel.h"

#define DEFAULT_FORMAT AFMT_U8
#define DEFAULT_STEREO 1
#define DEFAULT_FREQUENCY 44100
#define DEFAULT_NUMFRAGS 64
#define DEFAULT_FRAGSIZE 17

/*#define DEBUG_FILE*/

#ifdef DEBUG_FILE
/* make sure this file exists; we're not using O_CREAT! */
#define FILENAME "/tmp/axel.out"
#else
#define FILENAME "/dev/dsp"
#endif

/* Optimisation dependencies: none */
static AxMonitor ax_monitor_optimise (AXEL axel, AxMonitor monitor);

static AxBase
ax_monitor_reset_device (AXEL axel, AxBase base)
{
  AxMonitor monitor = (AxMonitor)base;
  CDSet channels = ax_get_channels (axel);
  AxCount nr_channels;
  int fragmentsize;

  nr_channels = cd_set_size (axel, channels);

  if (nr_channels == 1)
    monitor->stereo = 0;
  else if (nr_channels > 1)
    monitor->stereo = 1;

  monitor->mask = 0;
  monitor->format = DEFAULT_FORMAT;

  monitor->frequency = ax_get_samplerate (axel);
  monitor->numfrags = DEFAULT_NUMFRAGS;
  monitor->fragsize = DEFAULT_FRAGSIZE;


#ifndef DEBUG_FILE
  if (ioctl (monitor->file, SNDCTL_DSP_GETFMTS, &monitor->mask) == -1) {
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return AxNone;
  }

  if (monitor->mask & AFMT_U8) {
    monitor->format = AFMT_U8;
  }
  if (monitor->mask & AFMT_U16_LE) {
    monitor->format = AFMT_U16_LE;
  }
  if (monitor->mask & AFMT_S16_LE) {
    monitor->format = AFMT_S16_LE;
  }
  if (monitor->mask & AFMT_U16_BE) {
    monitor->format = AFMT_U16_BE;
  }
  if (monitor->mask & AFMT_S16_BE) {
    monitor->format = AFMT_S16_BE;
  }
  if (monitor->mask & AFMT_S8) {
    monitor->format = AFMT_S8;
  }
  if (monitor->mask & AFMT_S16_LE) {
    monitor->format = AFMT_S16_LE;
  }

  if (ioctl(monitor->file, SNDCTL_DSP_SETFMT, &monitor->format) == -1) {
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return AxNone;
  }

  if (ioctl(monitor->file, SNDCTL_DSP_STEREO, &(monitor->stereo)) == -1) {
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return AxNone;
  }

  if (ioctl(monitor->file, SNDCTL_DSP_SPEED, &(monitor->frequency)) == -1) {
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return AxNone;
  }

  fragmentsize = (monitor->numfrags << 16) | monitor->fragsize;
  if (ioctl(monitor->file, SNDCTL_DSP_SETFRAGMENT, &fragmentsize) == -1) {
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return AxNone;
  }
#else
  monitor->format = AFMT_S16_LE;
#endif /* DEBUG_FILE */

  return base;
}

static AxBase
ax_monitor_init (AXEL axel, AxBase base)
{
  AxMonitor monitor = (AxMonitor)base;

  if ((monitor->file = open (FILENAME, O_WRONLY | O_NDELAY, 0)) == -1) {
    printf ("couldn't open " FILENAME "\n");
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return AxNone;
  }

  ax_monitor_reset_device (axel, base);

  ax_monitor_optimise (axel, monitor);
  return (AxBase)monitor;
}

AxMonitor
ax_monitor_new (AXEL axel)
{
  AxBase monitor = ax_base_new_subclass (axel, sizeof (struct _AxMonitor));
  ax_monitor_init (axel, monitor);
  return (AxMonitor)monitor;
}

static AxBase
ax_monitor_clone (AXEL axel, AxBase base)
{
  AxMonitor new_monitor = ax_monitor_new (axel);
  ax_monitor_optimise (axel, new_monitor);
  return (AxBase)new_monitor;
}

static int
ax_monitor_destroy (AXEL axel, AxBase base)
{
  AxMonitor monitor = (AxMonitor)base;
  if (monitor->file != -1) {
    close (monitor->file);
  }
  ax_free (monitor);
  return 0;
}

static int
ax_monitor_ready (AXEL axel, AxBase base)
{
  AxMonitor monitor = (AxMonitor) base;
  CDSet channels = ax_get_channels (axel);
  int samplerate = (int) ax_get_samplerate (axel);
  AxCount nr_channels;

  nr_channels = cd_set_size (axel, channels);

  return (samplerate == monitor->frequency &&
	  ((nr_channels == 1 && monitor->stereo == 0) ||
	   (nr_channels > 1 && monitor->stereo == 1)));
}

static AxBase
ax_monitor_prepare (AXEL axel, AxBase base)
{
  ax_monitor_reset_device (axel, base);
  return base;
}

static AxCount
ax_monitor_write_short (AXEL axel, AxMonitor monitor, AxCount count)
{
  AxCount n = 0;
  fd_set fds;

  if (!(monitor->format & AFMT_S16_LE)) {
    printf ("###### device cannot play AFMT_S16_LE nicely\n");
    return count;
  }

  FD_ZERO (&fds);
  FD_SET (monitor->file, &fds);

  if ((select (monitor->file + 1, NULL, &fds, NULL,
	       NULL /* XXX: no timeout */)) == 0) {
    printf ("select error\n");
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return -1;
  }
  
  n = write (monitor->file, monitor->playbuffer,
	     count * sizeof(short));
  if (n == -1) {
    printf ("####### system error writing to fd %d #######\n",
	    monitor->file);
    
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return -1;
  }

  return n;
}

static AxCount
ax_monitor_fill_playbuffer (AXEL axel, AxMonitor monitor, AxPCM * data,
			    AxCount count)
{
  AxCount i;
  AxPCM value;
  const AxPCM max_value = (AxPCM)SHRT_MAX / 2;

  for (i = 0; i < count; i++) {
    value = *data++ * max_value;
    monitor->playbuffer[i] = (short)value;
  }
  return count;
}

/* An AxChunkFunc for making noise */
static AxCount
ax_monitor_chunk (AXEL axel, AxChunk chunk, AxCount offset,
		  AxCount count, int channelname, void * data)
{
  AxMonitor monitor = (AxMonitor)data;
  AxCount remaining = count, written = 0, n, playcount;
  AxPCM * d;
  
  if (monitor->file == -1) {
    ax_dprintf ("[ax_monitor_chunk] no file\n");
    ax_set_error (axel, AX_ERROR_NOENTITY); /* XXX: different error ? */
    return -1;
  }

  ax_dprintf ("[ax_monitor_chunk] (%p [chunk %p], +%ld), @ %ld\n",
	      monitor, chunk, count, offset);

  while (remaining > 0) {
    playcount = MIN (remaining, AX_MONITOR_BUFFERLEN);
    d = &chunk->data[offset];
    n = ax_monitor_fill_playbuffer (axel, monitor, d, playcount);

    n = ax_monitor_write_short (axel, monitor, n);
    if (n == -1) {
      return -1;
    } else {
      n /= sizeof (short);
    }

    offset += n;
    written += n;
    remaining -= n;      
}

  return written;
}

static AxCount
ax_monitor_process (AXEL axel, AxBase base, AxCount count,
		    AxStream input, AxStream output)
{
  AxMonitor monitor = (AxMonitor)base;
  AxCount nr_channels = ax_stream_nr_channels (axel, input);
  AxCount remaining = count, processed = 0, n, nn;

  if (nr_channels == 1 && monitor->stereo == 0) { /* MONO */
    return ax_stream_chunkfuncify (axel, input, count,
				   ax_monitor_chunk, monitor);
  } else if (nr_channels == 2 && monitor->stereo == 1) { /* STEREO */

    while (remaining > 0) {
      n = MIN (remaining, AX_MONITOR_BUFFERLEN/2);
      n = ax_stream_interleave_2 (axel, input,
				  AX_CHANNEL_LEFT, AX_CHANNEL_RIGHT,
				  monitor->databuffer, n);
      nn = 2 * n;
      nn = ax_monitor_fill_playbuffer (axel, monitor, monitor->databuffer, nn);
      nn = ax_monitor_write_short (axel, monitor, nn);

      processed += n;
      remaining -= n;
    }
    return processed;
  } else {
    printf ("[ax_monitor_process] unsupported stream/output channel\n");
    printf ("combination %ld / %d\n", nr_channels, monitor->stereo ? 2 : 1);
    return -1;
  }
}

static AxCount
ax_monitor_length (AXEL axel, AxBase base)
{
  return AX_COUNT_INFINITE;
}

static AxCount
ax_monitor_seek (AXEL axel, AxBase base, AxCount count)
{
  return count;
}

static int
ax_monitor_flush (AXEL axel, AxBase base)
{
  AxMonitor monitor = (AxMonitor)base;

#ifndef DEBUG_FILE
  if (ioctl(monitor->file, SNDCTL_DSP_POST, NULL) == -1) {
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return -1;
  }
#endif

  return 0;
}

static struct _AxMethods _ax_monitor_methods = {
  ax_monitor_clone,
  ax_monitor_destroy,
  ax_monitor_ready,
  ax_monitor_prepare,
  ax_monitor_process,
  ax_monitor_length,
  ax_monitor_seek,
  ax_monitor_flush,
};

static AxMonitor
ax_monitor_optimise (AXEL axel, AxMonitor monitor)
{
  _ax_set_methods (axel, (AxBase)monitor, &_ax_monitor_methods);
  return monitor;
}
