/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxPCM: generic functions for anonymous blocks of PCM data.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * This file contains only generic functions to manipulate blocks of
 * AxPCM data. The AxPCM type is defined in <axel_types.h>,  usually as
 * a floating point value (float).
 *
 * The functions in this file are an excellent target for short vector
 * machine optimisations.
 *
 * Invariants
 * ----------
 *
 * N/A
 *
 */

#define __AXEL__
#include "axel.h"


/* PFunc */

/*
 * _ax_pcm_clear_region (data, count)
 */
AxCount
_ax_pcm_clear_region (AxPCM * data, AxCount count, void * unused)
{
  memset (data, (AxPCM)0, count * sizeof (AxPCM));
  return count;
}


/* PVFunc */

AxCount
_ax_pcm_set (AxPCM * data, AxPCM value, AxCount count)
{
  AxCount i;

  for (i = 0; i < count; i++) {
    *data++ = value;
  }

  return count;
}

AxCount
_ax_pcm_gain (AxPCM * data, AxCount count, void * gain)
{
  AxPCM _gain = *(AxPCM *)gain;
  AxCount i;

  for (i = 0; i < count; i++) {
    *data++ *= _gain;
  }

  return count;
}


/* PPFunc */

/*
 * _ax_pcm_copy (src, dest, count)
 *
 * Copy PCM data from src to dest.
 */
AxCount
_ax_pcm_copy (AxPCM * src, AxPCM * dest, AxCount count, void * unused)
{
  memcpy (dest, src, count * sizeof (AxPCM));
  return count;
}

/*
 * _ax_pcm_add (src, dest, count)
 *
 * Add PCM data from src to dest.
 */
AxCount
_ax_pcm_add (AxPCM * src, AxPCM * dest, AxCount count, void * unused)
{
  AxCount i;

  for (i = 0; i < count; i++) {
    *dest++ += *src++;
  }

  return count;
}

/*
 * _ax_pcm_mult (src, dest, count)
 *
 * Multiply PCM data of dest by that in src.
 */
AxCount
_ax_pcm_mult (AxPCM * src, AxPCM * dest, AxCount count, void * unused)
{
  AxCount i;

  for (i = 0; i < count; i++) {
    *dest++ *= *src++;
  }

  return count;
}

/*
 * _ax_pcm_fade (src, dest, count)
 *
 * Fade PCM data of dest by that in src.
 */
AxCount
_ax_pcm_fade (AxPCM * src, AxPCM * dest, AxCount count, void * unused)
{
  AxCount i;

  for (i = 0; i < count; i++) {
    *dest++ *= (1.0 - *src++);
  }

  return count;
}

/*
 * _ax_pcm_interleave_2 (src1, src2, count, dest)
 *
 * Interleave data of src1 and src2, storing result in dest
 */
AxCount
_ax_pcm_interleave_2 (AxPCM * src1, AxPCM * src2, AxCount count, void * data)
{
  AxPCM * dest = (AxPCM *)data;
  AxCount i;

  for (i = 0; i < count; i++) {
    *dest++ = *src1++;
    *dest++ = *src2++;
  }

  return count;
}

/*
 * _ax_pcm_deinterleave_2 (dest1, dest2, count, src)
 *
 * Deinterleave data of src, storing result in dest1 and dest2
 */
AxCount
_ax_pcm_deinterleave_2 (AxPCM * dest1, AxPCM * dest2, AxCount count, void * data)
{
  AxPCM * src = (AxPCM *)data;
  AxCount i;

  for (i = 0; i < count; i++) {
    *dest1++ = *src++;
    *dest2++ = *src++;
  }

  return count;
}

/* PPPFunc */

/*
 * _ax_pcm_blend (src, blend, dest, count)
 *
 * Blend PCM data of 'src' into 'dest' by blend values in 'blend'.
 */
AxCount
_ax_pcm_blend (AxPCM * src, AxPCM * blend, AxPCM * dest, AxCount count,
	       void * unused)
{
  AxCount i;
  AxPCM b;

  for (i = 0; i < count; i++) {
    b = *blend++;
    *dest++ = (*dest * b) + (*src++ * (1.0 - b));
  }

  return count;
}


/* Miscellaneous */

/*
 * ax_pcm_write_linear (data, x1, y1, x2, y2, offset, count)
 *
 * Write 'count' samples at 'data' following line passing through (x1, y1)
 * and (x2, y2), with writing starting at x = 'offset'.
 */
AxCount
_ax_pcm_write_linear (AxPCM * data, AxCount x1, AxPCM y1, AxCount x2, AxPCM y2,
		      AxCount offset, AxCount count)
{
  AxCount i;

  ax_dprintf ("[ax_pcm_write_linear] ((%ld, %f) -> (%ld, %f), %ld +%ld)\n",
	  x1, y1, x2, y2, offset, count);

  for (i = 0; i < count; i++) {
    *data++ = y1 + (AxPCM)(i + offset - x1) * (y2 - y1) / (x2 - x1);
  }

  return count;
}
