/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxPlugin: A container for AxBase types.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <dlfcn.h>

#define __AXEL__
#include "axel.h"

static CDList
ax_plugin_initialise_static (AXEL axel)
{
  CDList plugins = cd_list_new (axel);

  plugins = cd_list_join (axel, plugins, __gain_init (axel));
  plugins = cd_list_join (axel, plugins, __sndfile_init (axel));

  return plugins;
}

static CDList
ax_plugin_init (AXEL axel, const char * path)
{
  void * module;
  AxPluginInitFunc init;

  module = dlopen (path, RTLD_NOW);

  if (!module) {
    fprintf (stderr, "Error opening %s: %s\n", path, dlerror ());
    ax_set_error (axel, AX_ERROR_SYSTEM);
    return CD_EMPTY_LIST;
  }

  if ((init = dlsym (module, "axel_load")) != NULL) {
    return init (axel);
  }

  return CD_EMPTY_LIST;
}


/*
 * Implementation of stripname for matching plugins of form
 * lib(.*).so
 *
 * If you want to implement support for other dynamic library
 * naming schemes, the correct thing to do is implement another
 * one of these functions. The g_module stuff deals with the
 * actual filenames portably.
 */
static char *
stripname (char * d_name)
{
  int len, blen;
#define BUF_LEN 8
  char buf[BUF_LEN];

  if (strncmp (d_name, "lib", 3)) return 0;

  len = strlen (d_name);
  snprintf (buf, BUF_LEN, ".so.%d", AXEL_PLUGIN_API_MAJOR);
  blen = strlen (buf);
  if (strncmp (&d_name[len-blen], buf, blen)) return 0;

  d_name[len-blen] = '\0';

  return &d_name[3];
}

#define BUFLEN 256

static CDList
init_dynamic_plugins_dir (AXEL axel, char * dirname)
{
  CDList plugins = cd_list_new (axel);
  DIR * dir;
  struct dirent * dirent;
  char * name;
  static char buf[BUFLEN];

  dir = opendir (dirname);
  if (!dir) {
    /* fail silently */
    return CD_EMPTY_LIST;
  }

  while ((dirent = readdir (dir)) != NULL) {
    if ((name = stripname (dirent->d_name)) != NULL) {
      snprintf (buf, BUFLEN, "%s/%s.so", dirname, dirent->d_name);
      plugins = cd_list_join (axel, plugins, ax_plugin_init (axel, buf));
    }
  }

  return plugins;
}


static CDList
ax_plugin_initialise_dynamic (AXEL axel)
{
  return init_dynamic_plugins_dir (axel, PACKAGE_PLUGIN_DIR);
}

void
ax_plugin_initialise_defaults (AXEL axel)
{
  CDList plugins = cd_list_new (axel);

  plugins = cd_list_join (axel, plugins, ax_plugin_initialise_static (axel));
  plugins = cd_list_join (axel, plugins, ax_plugin_initialise_dynamic (axel));

  cd_list_apply (axel, plugins, (CDFunc)_ax_register_plugin);
}

#if 0

AxPlugin
ax_plugin_new (void)
{
  AxPlugin plugin = ax_malloc (sizeof (struct _AxPlugin));
#if 0
  plugin->text = ax_meta_text_new ();
#endif
  return plugin;
}

AxMetaText
ax_plugin_get_meta_text (AxPlugin plugin)
{
  return plugin->text;
}

AxMetaText
ax_plugin_set_meta_text (AxPlugin plugin, AxMetaText mt)
{
  AxMetaText old = plugin->text;
  plugin->text = mt;
  return old;
}

int
ax_plugin_writeable (AxPlugin plugin)
{
  return (plugin->flags & AX_PLUGIN_WRITEABLE);
}

int
ax_plugin_seekable (AxPlugin plugin)
{
  return (plugin->flags & AX_PLUGIN_SEEKABLE);
}

int
ax_plugin_cacheable (AxPlugin plugin)
{
  return (plugin->flags & AX_PLUGIN_CACHEABLE);
}

int
ax_plugin_causal (AxPlugin plugin)
{
  return (plugin->flags & AX_PLUGIN_CAUSAL);
}

#endif
