/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxSndfile: a libsndfile handler
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <stdio.h>
#include <sndfile.h>
#include <string.h>

#define __AXEL__
#include "axel.h"

#define PATH_KEY 1

typedef struct _AxSndfileInstance * AxSndfileInstance;

struct _AxSndfileInstance {
  char * path;
  int writing;
  SNDFILE * file;
  SF_INFO info;
};


/* Optimisation dependencies: none */
static AxBase ax_sndfile_optimise (AXEL axel, AxBase sndfile);


static AxBase
ax_sndfile_create (AXEL axel, AxBase sndfile, const char * path, int writing)
{
  AxSndfileInstance si = ax_malloc (sizeof (struct _AxSndfileInstance));

  si->path = strdup (path);
  si->writing = writing;

  if (writing) {
    si->info.samplerate = ax_get_samplerate (axel);
    si->info.channels = 1; /* XXX: how many channels, or specify? */
    si->info.pcmbitwidth = 16; /* XXX: assumes 16bit */
    si->info.format = SF_FORMAT_WAV; /* XXX: assumes WAV */

    si->file = sf_open_write (path, &si->info);
  } else {
    si->file = sf_open_read (path, &si->info);
  }

  if (si->file == NULL) {
    ax_set_error (axel, AX_ERROR_SYSTEM);
    ax_destroy (axel, (AxBase)sndfile);
    return AxNone;
  }

  sndfile->instance_data = si;

  return sndfile;
}

static AxBase
ax_sndfile_reader_init (AXEL axel, AxBase base, CDSet parameters)
{
  char * path;

  path = (cd_set_find (axel, parameters, PATH_KEY)).s_string;

  ax_sndfile_create (axel, base, path, 0);
  ax_sndfile_optimise (axel, base);
  return base;
}

static AxBase
ax_sndfile_writer_init (AXEL axel, AxBase base, CDSet parameters)
{
  char * path;

  path = (cd_set_find (axel, parameters, PATH_KEY)).s_string;

  ax_sndfile_create (axel, base, path, 1);
  ax_sndfile_optimise (axel, base);
  return base;
}

static AxBase
ax_sndfile_clone (AXEL axel, AxBase base)
{
  AxBase new_sndfile = ax_base_new (axel);
  AxSndfileInstance si = (AxSndfileInstance)base->instance_data;
  ax_sndfile_create (axel, new_sndfile, si->path, si->writing);
  ax_sndfile_optimise (axel, new_sndfile);
  return new_sndfile;
}

static int
ax_sndfile_destroy (AXEL axel, AxBase base)
{
  AxSndfileInstance si = (AxSndfileInstance)base->instance_data;
  if (si->file != NULL) sf_close (si->file);
  ax_free (si);
  ax_free (base);
  return 0;
}

/* An AxChunkFunc for creating sndfile */
static AxCount
ax_sndfile_read_into_chunk (AXEL axel, AxChunk chunk, AxCount offset,
			    AxCount count, int channelname, void * data)
{
  AxBase sndfile = (AxBase)data;
  static double sfbuffer[4096];
  AxPCM * d;
  AxCount remaining = count, written = 0, i, n;
  AxSndfileInstance si = (AxSndfileInstance)sndfile->instance_data;

  ax_dprintf ("[ax_sndfile_read_into_chunk] (%p, +%ld) @ %ld\n",
	      sndfile, count, ax_tell (axel, sndfile));

  d = &chunk->data[offset];

  while (remaining > 0) {
    n = MIN (remaining, 4096);
    n = sf_read_double (si->file, sfbuffer, n, 1 /* normalize */);
    if (n == 0) { /* EOF */
      n = _ax_pcm_set (d, 0.0, remaining);
    } else {
      for (i = 0; i < n; i++) {
	*d++ = (AxPCM)sfbuffer[i];
      }
    }
    remaining -= n;
    written += n;
  }

  return written;
}

static AxCount
ax_sndfile_write_from_chunk (AXEL axel, AxChunk chunk, AxCount offset,
			     AxCount count, int channelname, void * data)
{
  AxBase sndfile = (AxBase)data;
  static double sfbuffer[4096];
  AxPCM * d;
  AxCount remaining = count, read = 0, i, n;
  AxSndfileInstance si = (AxSndfileInstance) sndfile->instance_data;

  ax_dprintf ("[ax_sndfile_write_from_chunk] (%p, +%ld) @ %ld\n",
	      sndfile, count, ax_tell (axel, sndfile));

  d = &chunk->data[offset];

  while (remaining > 0) {
    n = MIN (remaining, 4096);
    for (i = 0; i < n; i++) {
      sfbuffer[i] = (double)*d++;
    }
    n = sf_write_double (si->file, sfbuffer, n, 1 /* normalise */);
    if (n == 0) { /* EOF */
      n = remaining;
    }
    remaining -= n;
    read += n;
  }

  return read;
}

static AxCount
ax_sndfile_reader_process (AXEL axel, AxBase base, AxCount count,
			   AxStream input, AxStream output)
{
  return ax_stream_chunkfuncify (axel, output, count,
				 ax_sndfile_read_into_chunk, base);

}

static AxCount
ax_sndfile_writer_process (AXEL axel, AxBase base, AxCount count,
			   AxStream input, AxStream output)
{
  return ax_stream_chunkfuncify (axel, output, count,
				 ax_sndfile_write_from_chunk, base);
}

static AxCount
ax_sndfile_length (AXEL axel, AxBase base)
{
  AxSndfileInstance si = (AxSndfileInstance)base->instance_data;
  return si->info.samples;
}

static AxCount
ax_sndfile_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxSndfileInstance si = (AxSndfileInstance)base->instance_data;
  return sf_seek (si->file, offset, SEEK_SET);
}

static struct _AxMethods _ax_sndfile_reader_methods = {
  ax_sndfile_clone,
  ax_sndfile_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_sndfile_reader_process,
  ax_sndfile_length,
  ax_sndfile_seek,
  NULL, /* flush */
};

static struct _AxMethods _ax_sndfile_writer_methods = {
  ax_sndfile_clone,
  ax_sndfile_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_sndfile_writer_process,
  ax_sndfile_length,
  ax_sndfile_seek,
  NULL, /* flush */
};  

static AxBase
ax_sndfile_optimise (AXEL axel, AxBase sndfile)
{
  AxSndfileInstance si = (AxSndfileInstance)sndfile->instance_data;

  if (si->writing)
    ax_base_set_methods (axel, sndfile, &_ax_sndfile_writer_methods);
  else
    ax_base_set_methods (axel, sndfile, &_ax_sndfile_reader_methods);

  return sndfile;
}


static struct _AxParameterScheme path_scheme = {
  "path",
  "Path to sound file",
  AX_TYPE_STRING,
  AX_CONSTRAINT_TYPE_NONE,
  AX_CONSTRAINT_EMPTY,
  AX_HINT_FILENAME,
};

static AxNamedParameter types[] = {
  AX_NAMED_PARAMETER ("WAV", CD_INT(SF_FORMAT_WAV)),
  AX_NAMED_PARAMETER ("AIFF", CD_INT(SF_FORMAT_AIFF)),
  AX_NAMED_PARAMETER ("Sun/NeXT AU format (big endian)", CD_INT(SF_FORMAT_AU)),
  AX_NAMED_PARAMETER ("DEC AU format (little endian)", CD_INT(SF_FORMAT_AULE)),
  AX_NAMED_PARAMETER ("RAW PCM data", CD_INT(SF_FORMAT_RAW)),
  AX_NAMED_PARAMETER ("Ensoniq PARIS", CD_INT(SF_FORMAT_PAF)),
  AX_NAMED_PARAMETER ("Amiga IFF / SVX8 / SV16", CD_INT(SF_FORMAT_SVX)),
  AX_NAMED_PARAMETER ("Sphere NIST", CD_INT(SF_FORMAT_NIST)),
  AX_NAMED_PARAMETER ("Windows Media Audio", CD_INT(SF_FORMAT_WMA)),
  AX_NAMED_PARAMETER ("Sekd Samplitude", CD_INT(SF_FORMAT_SMPLTD)),
  AX_NAMED_PARAMETER ("VOC", CD_INT(SF_FORMAT_VOC)),
  AX_NAMED_PARAMETER ("Sound Designer 2", CD_INT(SF_FORMAT_SD2)),
  AX_NAMED_PARAMETER ("Rex2", CD_INT(SF_FORMAT_REX2)),
  AX_NAMED_PARAMETER ("PCM", CD_INT(SF_FORMAT_PCM)),
  AX_NAMED_PARAMETER ("32 bit floats", CD_INT(SF_FORMAT_FLOAT)),
  AX_NAMED_PARAMETER ("U-Law encoded", CD_INT(SF_FORMAT_ULAW)),
  AX_NAMED_PARAMETER ("A-Law encoded", CD_INT(SF_FORMAT_ALAW)),
  AX_NAMED_PARAMETER ("IMA ADPCM", CD_INT(SF_FORMAT_IMA_ADPCM)),
  AX_NAMED_PARAMETER ("Microsoft ADPCM", CD_INT(SF_FORMAT_MS_ADPCM)),
  AX_NAMED_PARAMETER ("Big endian PCM", CD_INT(SF_FORMAT_PCM_BE)),
  AX_NAMED_PARAMETER ("Little endian PCM", CD_INT(SF_FORMAT_PCM_LE)),
  AX_NAMED_PARAMETER ("Signed 8 bit PCM", CD_INT(SF_FORMAT_PCM_S8)),
  AX_NAMED_PARAMETER ("Unsigned 8 bit PCM", CD_INT(SF_FORMAT_PCM_U8)),
  AX_NAMED_PARAMETER ("SVX Fibonacci Delta", CD_INT(SF_FORMAT_SVX_FIB)),
  AX_NAMED_PARAMETER ("SVX Exponential Delta", CD_INT(SF_FORMAT_SVX_EXP)),
  AX_NAMED_PARAMETER ("GSM 6.10 Encoding", CD_INT(SF_FORMAT_GSM610)),
  AX_NAMED_PARAMETER ("32kbs G721 ADPCM", CD_INT(SF_FORMAT_G721_32)),
  AX_NAMED_PARAMETER ("24kbs G723 ADPCM", CD_INT(SF_FORMAT_G723_24)),
};

static struct _AxParameterScheme type_scheme = {
  "type",
  "Type of sound file",
  AX_TYPE_INT,
  AX_CONSTRAINT_TYPE_LIST,
  AX_CONSTRAINT_EMPTY,
  AX_HINT_DEFAULT,
};

static struct _AxMetaText sndfile_reader_metatext = {
  "builtin::sndfile_reader",
  "File::Sndfile Reader",
  "Reads PCM audio files using libsndfile",
  "Copyright (C) 2001 CSIRO Australia",
  "http://www.metadecks.org/axel/plugins/sndfile.html",
  AX_ONE_AUTHOR ("Conrad Parker", "Conrad.Parker@CSIRO.AU"),
};

static struct _AxMetaText sndfile_writer_metatext = {
  "builtin::sndfile_writer",
  "File::Sndfile Writer",
  "Writes PCM audio files using libsndfile",
  "Copyright (C) 2001 CSIRO Australia",
  "http://www.metadecks.org/axel/plugins/sndfile.html",
  AX_ONE_AUTHOR ("Conrad Parker", "Conrad.Parker@CSIRO.AU"),
};

static struct _AxPlugin sndfile_reader_plugin = {
  &sndfile_reader_metatext,
  AX_FLAGS_NONE,
  CD_EMPTY_SET, /* init scheme */
  ax_sndfile_reader_init,
  CD_EMPTY_SET, /* process scheme */
  NULL, /* suggests */
  NULL, /* plugin data */
};

static struct _AxPlugin sndfile_writer_plugin = {
  &sndfile_writer_metatext,
  AX_FLAGS_NONE,
  CD_EMPTY_SET, /* init scheme */
  ax_sndfile_writer_init,
  CD_EMPTY_SET, /* process scheme */
  NULL, /* suggests */
  NULL, /* plugin data */
};

/* module init function */
CDList
__sndfile_init (AXEL axel)
{
  CDList plugins = cd_list_new (axel);
  int i;

  sndfile_reader_plugin.init_scheme =
    cd_set_insert (axel, sndfile_reader_plugin.init_scheme, PATH_KEY,
		   CD_POINTER(&path_scheme));

  plugins = cd_list_prepend (axel, plugins,
			     CD_POINTER(&sndfile_reader_plugin));

  type_scheme.constraint.list = cd_list_new (axel);
  for (i = 0; i < CD_ARRAY_SIZE(types); i++) {
    type_scheme.constraint.list =
      cd_list_append (axel, type_scheme.constraint.list, CD_POINTER(types[i]));
  }

  sndfile_writer_plugin.init_scheme =
    cd_set_insert (axel, sndfile_writer_plugin.init_scheme, PATH_KEY,
		   CD_POINTER(&path_scheme));

  plugins = cd_list_prepend (axel, plugins,
			     CD_POINTER(&sndfile_writer_plugin));

  return plugins;
}

