/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxSound: An instance of an base within an AxLayer sequence.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A sound is contained within a layer. Each sound is a unique entity, but
 * many sounds may have the same source base.
 *
 * Invariants
 * ----------
 *
 * A sound must be contained in a layer.
 */

#define __AXEL__
#include "axel.h"

/* Optimisation dependencies: none */
static AxSound ax_sound_optimise (AXEL axel, AxSound sound);

/*
 * ax_sound_replace_mixstreams (axel, sound)
 *
 * Replaces mix- (and envelope-) streams of 'sound' with new ones of
 * the axel's mixlength.
 */
static void
ax_sound_replace_mixstreams (AXEL axel, AxSound sound)
{
  AxCount mixlength = _ax_base_get_mixlength (axel, sound);

  if (sound->_rate_envstream != AxNone)
    ax_destroy (axel, (AxBase)sound->_rate_envstream);
  if (sound->_gain_envstream != AxNone)
    ax_destroy (axel, (AxBase)sound->_gain_envstream);
  if (sound->_blend_envstream != AxNone)
    ax_destroy (axel, (AxBase)sound->_blend_envstream);

  sound->_rate_envstream =
    ax_stream_new_contiguous (axel, mixlength);
  sound->_gain_envstream =
    ax_stream_new_contiguous (axel, mixlength);
  sound->_blend_envstream =
    ax_stream_new_contiguous (axel, mixlength);
}

static AxBase
ax_sound_init (AXEL axel, AxBase base)
{
  AxSound sound = (AxSound)base;
  sound->rate_envelope = sound->gain_envelope = sound->blend_envelope = AxNone;
  sound->cutin = sound->cutlength = 0;
  ax_sound_replace_mixstreams (axel, sound);
  ax_sound_optimise (axel, sound);
  return (AxBase)sound;
}

static AxSound
ax_sound_new (AXEL axel)
{
  return (AxSound) ax_base_new_subclass (axel, sizeof (struct _AxSound));
}

AxBase
ax_sound_clone_invalid (AXEL axel, AxBase base)
{
  AxSound sound = (AxSound)base;
  AxSound new_sound = ax_sound_new (axel);

  memcpy (new_sound, sound, sizeof (struct _AxSound));
  new_sound->layer = AxNone;

  return (AxBase)new_sound;
}

AxBase
ax_sound_clone_with_layer (AXEL axel, AxBase base, AxLayer new_layer)
{
  AxSound sound = (AxSound)base;
  AxSound new_sound = ax_sound_new (axel);

  memcpy (new_sound, sound, sizeof (struct _AxSound));
  new_sound->layer = new_layer;
  _ax_layer_add_sound (axel, new_layer, new_sound, new_sound->start_time);

  return (AxBase)new_sound;
}

static int
ax_sound_destroy (AXEL axel, AxBase base)
{
  AxSound sound = (AxSound)base;

  ax_remove_sound (axel, sound);

  if (sound->rate_envelope)
    ax_destroy (axel, sound->rate_envelope);
  if (sound->gain_envelope)
    ax_destroy (axel, sound->gain_envelope);
  if (sound->blend_envelope)
    ax_destroy (axel, sound->blend_envelope);
  if (sound->_rate_envstream)
    ax_destroy (axel, (AxBase)sound->_rate_envstream);
  if (sound->_gain_envstream)
    ax_destroy (axel, (AxBase)sound->_gain_envstream);
  if (sound->_blend_envstream)
    ax_destroy (axel, (AxBase)sound->_blend_envstream);
  ax_free (sound);

  return 0;
}

static int
ax_sound_ready (AXEL axel, AxBase base)
{
  return 0;
}

static AxBase
ax_sound_prepare (AXEL axel, AxBase base)
{
  AxSound sound = (AxSound)base;
  ax_sound_replace_mixstreams (axel, sound);
  return base;
}

AxBase
ax_sound_set_source (AXEL axel, AxSound sound, AxBase source)
{
  AxBase old = sound->source;
  sound->source = source;
  return old;
}

AxBase
ax_sound_get_source (AXEL axel, AxSound sound)
{
  return sound->source;
}

AxSound
ax_add_sound (AXEL axel, AxBase source, AxLayer layer, AxTime start_time,
	      AxTime duration)
{
  AxSound sound = ax_sound_new (axel);

  sound->layer = layer;
  sound->start_time = start_time;
  sound->duration = duration;
  _ax_layer_add_sound (axel, layer, sound, start_time);
  sound->source = source;
  ax_sound_init (axel, (AxBase)sound);
  return sound;
}

/* XXX: This breaks sound invariant */
AxSound
ax_remove_sound (AXEL axel, AxSound sound)
{
  if (sound->layer == AxNone) return AxNone;
  return _ax_layer_remove_sound (axel, sound->layer, sound);
}

AxLayer
ax_sound_get_layer (AXEL axel, AxSound sound)
{
  return sound->layer;
}

AxTrack
ax_sound_get_track (AXEL axel, AxSound sound)
{
  if (sound->layer == AxNone) return AxNone;
  return ax_layer_get_track (axel, sound->layer);
}

AxDeck
ax_sound_get_deck (AXEL axel, AxSound sound)
{
  if (sound->layer == AxNone) return AxNone;
  return ax_layer_get_deck (axel, sound->layer);
}

AxTime
ax_sound_move (AXEL axel, AxSound sound, AxTime start_time)
{
  AxTime old = sound->start_time;
  AxLayer layer = sound->layer;
  if (ax_remove_sound (axel, sound) != AxNone) {
    _ax_layer_add_sound (axel, layer, sound, start_time);
    return old;
  } else {
    return _ax_time_invalid (sound->layer->timetype);
  } 
}

AxSound
ax_sound_get_prev (AXEL axel, AxSound sound)
{
  if (sound->layer)
    return _ax_layer_get_sound_prev (axel, sound->layer, sound);
  else
    return AxNone;
}

AxSound
ax_sound_get_next (AXEL axel, AxSound sound)
{
  if (sound->layer)
    return _ax_layer_get_sound_next (axel, sound->layer, sound);
  else
    return AxNone;
}

int
ax_sound_later (AXEL axel, AxSound s1, AxSound s2)
{
  AxLayer layer = ax_sound_get_layer (axel, s1);
  return _ax_time_gt (layer->timetype, s1->start_time, s2->start_time);
}

AxTime
ax_sound_set_start_time (AXEL axel, AxSound sound, AxTime start_time)
{
  AxTime old = sound->start_time;
  AxLayer layer = sound->layer;
  _ax_layer_remove_sound (axel, layer, sound);
  _ax_layer_add_sound (axel, layer, sound, start_time);
  return old;
}

AxTime
ax_sound_get_start_time (AXEL axel, AxSound sound)
{
  return sound->start_time;
}

AxTime
ax_sound_set_duration (AXEL axel, AxSound sound, AxTime duration)
{
  AxTime old = sound->duration;
  sound->duration = duration;
  return old;
}

AxTime
ax_sound_get_duration (AXEL axel, AxSound sound)
{
  return sound->duration;
}

AxBase
ax_sound_set_rate_envelope (AXEL axel, AxSound sound, AxBase rate_envelope)
{
  AxBase old = sound->rate_envelope;
  sound->rate_envelope = rate_envelope;
  return old;
}

AxBase
ax_sound_get_rate_envelope (AXEL axel, AxSound sound)
{
  if (sound == AxNone) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return AxNone;
  }
  return sound->rate_envelope;
}

AxBase
ax_sound_set_gain_envelope (AXEL axel, AxSound sound, AxBase gain_envelope)
{
  AxBase old;

  if (sound == AxNone) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return AxNone;
  }
  old = sound->gain_envelope;
  sound->gain_envelope = gain_envelope;

  return old;
}

AxBase
ax_sound_get_gain_envelope (AXEL axel, AxSound sound)
{
  if (sound == AxNone) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return AxNone;
  }
  return sound->gain_envelope;
}

AxBase
ax_sound_set_blend_envelope (AXEL axel, AxSound sound,
			       AxBase blend_envelope)
{
  AxBase old = sound->blend_envelope;
  sound->blend_envelope = blend_envelope;
  return old;
}

AxBase
ax_sound_get_blend_envelope (AXEL axel, AxSound sound)
{
  if (sound == AxNone) {
    ax_set_error (axel, AX_ERROR_INVALID);
    return AxNone;
  }
  return sound->blend_envelope;
}

static AxCount
_ax_sound_fade (AXEL axel, AxSound sound, AxCount count, AxStream input,
		AxStream output)
{
  AxCount output_offset, n;

  output_offset = ax_tell (axel, (AxBase)output);

  /* XXX: this wanted to use 'block' not 'count': have we bounds
   * checked the sound ?? */

  ax_seek (axel, (AxBase)sound->_blend_envstream, 0, SEEK_SET);
  n = ax_process (axel, sound->blend_envelope, count, AxNone,
		  sound->_blend_envstream);
  ax_stream_write (axel, output, count, input);
  ax_seek (axel, (AxBase)sound->_blend_envstream, 0, SEEK_SET);
  ax_seek (axel, (AxBase)output, output_offset, SEEK_SET);
  n = ax_stream_fade (axel, sound->_blend_envstream, output, count);

  return n;
}

/* Do rate conversion, handle offset etc.: get raw sound data */
static AxCount
_ax_sound_get_raw (AXEL axel, AxSound sound, AxCount offset, AxCount count,
		   AxStream input, AxStream output)
{
  AxCount block, n;

  ax_dprintf ("[_ax_sound_get_raw] (%p, +%ld, %p -> %p) @ %ld\n",
	      sound, count, input, output, offset);

  /* XXX: deal with rate conversion!! */

  if (sound->cutlength > 0) {
    if (offset > sound->cutlength) {
      ax_set_error (axel, AX_ERROR_SILENCE);
      return -1;
    }
    block = MIN (count, sound->cutlength - offset);
  } else {
    block = count;
  }

  ax_dprintf ("[_ax_sound_get_raw] block +%ld (cutin: %ld, cutlength: %ld)\n",
	      block, sound->cutin, sound->cutlength);

  ax_seek (axel, sound->source, sound->cutin + offset, SEEK_SET);
  n = ax_process (axel, sound->source, block, input, output);

  if (n == -1) {
    ax_dprintf ("error getting source data: %s\n",
		ax_error_string(axel, ax_last_error(axel)));
  } else {

    if (block < count)
      n += ax_stream_write0 (axel, output, count - block);
  }

  ax_dprintf ("[_ax_sound_get_raw] got %ld raw samples\n", n);

  return n;
}

static AxCount
_ax_sound_apply_gain (AXEL axel, AxSound sound, AxCount offset, AxCount count,
		      AxStream data, AxCount data_offset)
{
  AxCount n;

  ax_dprintf ("in _ax_sound_apply_gain (%p, +%ld)\n", sound, count);

  ax_seek (axel, sound->gain_envelope, offset, SEEK_SET);
  ax_seek (axel, (AxBase)sound->_gain_envstream, 0, SEEK_SET);
  n = ax_process (axel, sound->gain_envelope, count,
		  AxNone, sound->_gain_envstream);
  ax_dprintf ("Got %ld values from gain_envelope %p onto stream %p\n",
	  n, sound->gain_envelope, sound->_gain_envstream);

  ax_seek (axel, (AxBase)data, data_offset, SEEK_SET);
  ax_seek (axel, (AxBase)sound->_gain_envstream, 0, SEEK_SET);
  n = ax_stream_mult (axel, sound->_gain_envstream, data, n);
  ax_dprintf ("Multiplied %ld values of gain\n", n);

  return n;
}

static AxCount
_ax_sound_blend (AXEL axel, AxSound sound, AxCount count, AxStream input,
		 AxStream output)
{
  AxCount n;

  /* XXX: this wanted to use 'block' not 'count': have we bounds
   * checked the sound ?? */

  ax_seek (axel, (AxBase)sound->_blend_envstream, 0, SEEK_SET);
  n = ax_process (axel, sound->blend_envelope, count, AxNone,
		  sound->_blend_envstream);
  ax_seek (axel, (AxBase)sound->_blend_envstream, 0, SEEK_SET);
  n = ax_stream_blend (axel, input, sound->_blend_envstream, output,
		       count);

  return n;
}

static AxCount
ax_sound_process (AXEL axel, AxBase base, AxCount count, AxStream input,
                  AxStream output)
{
  AxSound sound = (AxSound)base;
  AxCount remaining = count, processed = 0, block, m, n;
  AxCount offset, input_offset, output_offset;
  AxCount mixlength = _ax_base_get_mixlength (axel, sound);

  offset = ax_tell (axel, (AxBase)sound);

  ax_dprintf ("PROCESS SOUND (%p [%p], +%ld, %p -> %p) @ %ld\n",
	  sound, sound ? sound->source : NULL, count, input, output, offset);

  /* if we're beyond the source's actual length, then
   * just fade the input to the output using the sound's blend
   * envelope */
  if (offset > ax_length (axel, sound->source)) {
    ax_dprintf ("## offset %ld > length %ld\n", offset,
	    ax_length (axel, sound->source));

    /* Fade input and copy to output */
    if (sound->blend_envelope == AxNone) {
      ax_set_error (axel, AX_ERROR_NOOP);
      return -1;
    } else {
      return _ax_sound_fade (axel, sound, count, input, output);
    }
  }

  while (remaining > 0) {
    block = MIN (remaining, mixlength);

    input_offset = ax_tell (axel, (AxBase)input);
    output_offset = ax_tell (axel, (AxBase)output);

    /* Get raw sound data; handle rate conversion etc. */
    m = _ax_sound_get_raw (axel, sound, offset, block, input, output);
    if (m == -1) {
      ax_dprintf ("error getting raw sound data\n");
      break;
    } else {
      n = m;
    }

    /* Apply gain envelope */
    if (sound->gain_envelope != AxNone) {
      m = _ax_sound_apply_gain (axel, sound, offset, n, output, output_offset);
      if (m == -1) {
	ax_dprintf ("error applying gain!\n");
      } else {
	n = m;
      }
    }
    
    /* Blend input back in */
    if (sound->blend_envelope != AxNone) {
      ax_seek (axel, (AxBase)input, input_offset, SEEK_SET);
      ax_seek (axel, (AxBase)output, output_offset, SEEK_SET);
      n = _ax_sound_blend (axel, sound, n, input, output);
    }
    
    offset += n;
    processed += n;
    remaining -= n;
  }

  ax_dprintf ("[ax_sound_process] processed %ld from sound %p\n",
	      processed, sound);

  return processed;
}

static AxCount
ax_sound_length (AXEL axel, AxBase base)
{
  AxSound sound = (AxSound)base;
  AxTime duration = ax_sound_get_duration (axel, sound);
  AxTime t = ax_time_convert (axel, duration, sound->layer->timetype,
			      AX_TIME_SAMPLES);
  return t.samples;
}

static AxCount
ax_sound_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxSound sound = (AxSound)base;
  if (sound->cutlength > 0 && offset > sound->cutlength) {
    offset = sound->cutlength;
  }
  ax_seek (axel, sound->source, sound->cutin + offset, SEEK_SET);
  return offset;
}

static int
ax_sound_flush (AXEL axel, AxBase base)
{
  AxSound sound = (AxSound)base;
  return ax_flush (axel, sound->source);
}

static struct _AxMethods _ax_sound_methods = {
  ax_sound_clone_invalid, /* clone */
  ax_sound_destroy,       /* destroy */
  ax_sound_ready,         /* ready */
  ax_sound_prepare,       /* prepare */
  ax_sound_process,       /* process */
  ax_sound_length,        /* length */
  ax_sound_seek,          /* seek */
  ax_sound_flush,         /* flush */
};

static AxSound
ax_sound_optimise (AXEL axel, AxSound sound)
{
  _ax_set_methods (axel, sound, &_ax_sound_methods);
  return sound;
}

