/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxSquareTone: a square wave tone generator
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#define __AXEL__
#include "axel.h"

/* Optimisation dependencies: none */
static AxSquareTone ax_squaretone_optimise (AXEL axel,
					    AxSquareTone squaretone);

static AxSquareTone
ax_squaretone_replace_channels (AXEL axel, AxSquareTone squaretone)
{
  AxSquareToneChannel sqch;
  CDSet s, channels = ax_get_channels (axel);
  AxCount offset = ax_tell (axel, (AxBase)squaretone);

  cd_set_free_all (axel, squaretone->channels);

  for (s = channels; s; s = s->next) {
    ax_dprintf ("[ax_squaretone_replace_channels] %p replacing channel %d\n",
		squaretone, s->key);
    sqch = (AxSquareToneChannel)
      ax_malloc (sizeof (struct _AxSquareToneChannel));
    sqch->_cycle_offset = 0;
    squaretone->channels =
      cd_set_insert (axel, squaretone->channels, s->key, CD_POINTER(sqch));
  }

  if (offset > 0) ax_seek (axel, (AxBase)squaretone, offset, SEEK_SET);

  return squaretone;
}

static AxBase
ax_squaretone_init (AXEL axel, AxBase base)
{
  AxSquareTone squaretone = (AxSquareTone)base;
  squaretone->channels = NULL;
  ax_squaretone_replace_channels (axel, squaretone);
  ax_squaretone_optimise (axel, squaretone);
  return (AxBase)squaretone;
}

AxSquareTone
ax_squaretone_new (AXEL axel, float frequency)
{
  AxSquareTone squaretone =
    (AxSquareTone)ax_base_new_subclass (axel, sizeof (struct _AxSquareTone));
  ax_squaretone_init (axel, (AxBase)squaretone);
  squaretone->frequency = frequency;
  return squaretone;
}

static AxBase
ax_squaretone_clone (AXEL axel, AxBase base)
{
  AxSquareTone squaretone = (AxSquareTone)base;
  AxSquareTone new_squaretone =
    ax_squaretone_new (axel, squaretone->frequency);
  ax_squaretone_optimise (axel, new_squaretone);
  return (AxBase)new_squaretone;
}

static int
ax_squaretone_destroy (AXEL axel, AxBase base)
{
  AxSquareTone squaretone = (AxSquareTone)base;
  ax_free (squaretone);
  return 0;
}

static int
ax_squaretone_ready (AXEL axel, AxBase base)
{
  return (ax_base_has_samplerate (axel, base) &&
	  ax_base_encompasses_channels (axel, base));
}

static AxBase
ax_squaretone_prepare (AXEL axel, AxBase base)
{
  AxSquareTone squaretone = (AxSquareTone)base;
  ax_squaretone_replace_channels (axel, squaretone);
  return base;
}

float
ax_squaretone_set_frequency (AXEL axel, AxSquareTone squaretone,
			     float frequency)
{
  float old = squaretone->frequency;
  squaretone->frequency = frequency;
  return old;
}

float
ax_squaretone_get_frequency (AXEL axel, AxSquareTone squaretone)
{
  return squaretone->frequency;
}

/* An AxChunkFunc for creating squaretone data */
static AxCount
ax_squaretone_write_chunk (AXEL axel, AxChunk chunk, AxCount offset,
			   AxCount count, int channelname, void * data)
{
  AxSquareTone squaretone = (AxSquareTone)data;
  AxSquareToneChannel sqch;
  AxCount remaining = count, written = 0, n;
  AxCount wavelength;
  AxPCM * d, value;
  CDScalar k;

  ax_dprintf ("[ax_squaretone_write_chunk] (+%ld) @ %ld\n", count, offset);

  k = cd_set_find (axel, squaretone->channels, channelname);
  sqch = k.s_pointer;

  if (sqch == AxNone) {
    ax_dprintf ("[ax_squaretone_write_chunk] channel %d not found\n",
		channelname);
    ax_set_error (axel, AX_ERROR_SILENCE);
    return -1;
  }

  wavelength = (AxCount)(ax_get_samplerate (axel) / squaretone->frequency);

  ax_dprintf ("[ax_squaretone_write_chunk] wavelength %ld, cycle_offset %ld\n",
	      wavelength, sqch->_cycle_offset);
  
  if (sqch->_cycle_offset < wavelength/2) {
    n = MIN (remaining, wavelength/2 - sqch->_cycle_offset);
    value = 1.0;
  } else {
    n = MIN (remaining, wavelength - sqch->_cycle_offset);
    value = -1.0;
  }

  d = &chunk->data[offset];
  _ax_pcm_set (d, value, n);
  remaining -= n; written += n; offset += n;

  while (remaining > 0) {
    n = MIN (remaining, wavelength/2);
    value = -(value);
    d = &chunk->data[offset];
    _ax_pcm_set (d, value, n);
    remaining -= n; written += n; offset += n;
  }

  sqch->_cycle_offset += written;
  sqch->_cycle_offset %= wavelength;

  ax_dprintf ("[ax_squaretone_write_chunk] written %ld\n", written);

  return written;
}

static AxCount
ax_squaretone_process (AXEL axel, AxBase base, AxCount count,
		       AxStream input, AxStream output)
{
  AxSquareTone squaretone = (AxSquareTone)base;
  return ax_stream_chunkfuncify (axel, output, count,
				 ax_squaretone_write_chunk, squaretone);
}

static AxCount
ax_squaretone_length (AXEL axel, AxBase base)
{
  return AX_COUNT_INFINITE;
}

static AxCount
ax_squaretone_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxSquareTone squaretone = (AxSquareTone)base;
  AxCount wavelength;
  CDSet s;
  AxCount cycle_offset;
  AxSquareToneChannel sqch;

  wavelength = (AxCount)(ax_get_samplerate (axel) / squaretone->frequency);
  cycle_offset = offset % wavelength;

  for (s = squaretone->channels; s; s = s->next) {
    sqch = (AxSquareToneChannel)s->data.s_pointer;
    sqch->_cycle_offset = cycle_offset;
  }

  return offset;
}

static struct _AxMethods _ax_squaretone_methods = {
  ax_squaretone_clone,
  ax_squaretone_destroy,
  ax_squaretone_ready,
  ax_squaretone_prepare,
  ax_squaretone_process,
  ax_squaretone_length,
  ax_squaretone_seek,
  NULL, /* flush */
};

static AxSquareTone
ax_squaretone_optimise (AXEL axel, AxSquareTone squaretone)
{
  _ax_set_methods (axel, squaretone, &_ax_squaretone_methods);
  return squaretone;
}
