/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxStream: An indexed, sparse, polyphonic PCM data container.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A stream consists of multiple channels of PCM data.
 *
 * Invariants
 * ----------
 *
 * A stream is an independent entity.
 */

#define __AXEL__
#include "axel.h"

/* AxStream */

/* Optimisation dependencies: none */
static AxStream ax_stream_optimise (AXEL axel, AxStream stream);

AxBase
ax_stream_init (AXEL axel, AxBase base)
{
  AxStream stream = (AxStream)base;
  CDSet s, channels = ax_get_channels (axel);

  stream->channels = cd_set_new (axel);

  for (s = channels; s; s = s->next) {
    ax_stream_add_channel (axel, stream, s->key);
  }

  ax_stream_optimise (axel, stream);
  return (AxBase)stream;
}

static AxStream
ax_stream_add_channel_unchecked (AXEL axel, AxStream stream, int name,
				 AxChannel channel)
{
  stream->channels = cd_set_insert (axel, stream->channels, name,
				    CD_POINTER(channel));
  return stream;
}

AxChannel
ax_stream_add_channel (AXEL axel, AxStream stream, int name)
{
  AxChannel channel = ax_stream_find_channel (axel, stream, name);

  if (channel == AxNone) {
    channel = ax_channel_new (axel);
    ax_stream_add_channel_unchecked (axel, stream, name, channel);
  }

  return channel;
}

AxStream
ax_stream_new (AXEL axel)
{
  AxStream stream =
    (AxStream)ax_base_new_subclass (axel, sizeof (struct _AxStream));

  ax_stream_init (axel, (AxBase)stream);

  return stream;
}

AxStream
ax_stream_new_contiguous (AXEL axel, AxCount length)
{
  AxStream stream = ax_stream_new (axel);
  ax_stream_add_chunks (axel, stream, 0, length);
  return stream;
}

static AxBase
ax_stream_clone (AXEL axel, AxBase base)
{
  AxStream stream = (AxStream)base;
  AxStream new_stream;
  CDSet s;
  AxChannel channel, new_channel;

  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  new_stream = (AxStream)ax_stream_new (axel);

  /* clone channels */
  for (s = stream->channels; s; s = s->next) {
    channel = (AxChannel)s->data.s_pointer;
    new_channel = ax_channel_clone (axel, channel);
    ax_stream_add_channel_unchecked (axel, new_stream, s->key, new_channel);
  }
  return (AxBase)new_stream;
}

static int
ax_stream_destroy (AXEL axel, AxBase base)
{
  AxStream stream = (AxStream)base;

  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  cd_set_destroy_with (axel, stream->channels, (CDDestroyFunc)ax_destroy);

  ax_free (stream);
  return 0;
}

AxCount
ax_stream_nr_channels (AXEL axel, AxStream stream)
{
  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  return (AxCount)cd_set_size (axel, stream->channels);
}

AxChannel
ax_stream_find_channel (AXEL axel, AxStream stream, int name)
{
  CDScalar k;

  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  k = cd_set_find (axel, stream->channels, name);
  return (AxChannel)k.s_pointer;

}


AxStream
ax_stream_remove_channel (AXEL axel, AxStream stream, int name)
{
  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }
   
  stream->channels = cd_set_remove (axel, stream->channels, name);

  return stream;
}

AxStream
ax_stream_add_chunks (AXEL axel, AxStream stream, AxCount offset,
		      AxCount length)
{
  CDSet s;
  AxChannel channel;

  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return AxNone;
  }

  for (s = stream->channels; s; s = s->next) {
    channel = (AxChannel)s->data.s_pointer;
    ax_channel_add_new_chunk (axel, channel, offset, length);
  }

  return stream;
}

AxCount
ax_stream_write0 (AXEL axel, AxStream stream, AxCount count)
{
  CDSet s;
  AxChannel channel;
  AxCount offset;

  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  offset = ax_tell (axel, (AxBase)stream);

  for (s = stream->channels; s; s = s->next) {
    channel = (AxChannel)s->data.s_pointer;
    ax_channel_write0 (axel, channel, count);
  }
  
  ax_seek (axel, (AxBase)stream, offset + count, SEEK_SET);

  ax_dprintf ("[ax_stream_write0] (%p) written %ld\n", stream, count);

  return count;
}

/*
 * ax_stream_process (axel, stream, count, input, output)
 *
 * AxProcessFunc for AxStream.
 * Ignores input. Copies data from 'stream' to 'output'.
 */
static AxCount
ax_stream_process (AXEL axel, AxBase base, AxCount count, AxStream input,
                   AxStream output)
{
  AxStream stream = (AxStream)base;
  return ax_stream_write (axel, output, count, stream);
}

static AxCount
ax_stream_length (AXEL axel, AxBase base)
{
  AxStream stream = (AxStream)base;
  AxCount length, maxlength = 0;
  CDSet s;
  AxChannel channel;

  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  for (s = stream->channels; s; s = s->next) {
    channel = (AxChannel)s->data.s_pointer;
    length = _ax_channel_length (axel, channel);
    maxlength = MAX (maxlength, length);
  }

  return maxlength;
}

static AxCount
ax_stream_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxStream stream = (AxStream)base;
  CDSet s;
  AxChannel channel;

  for (s = stream->channels; s; s = s->next) {
    channel = (AxChannel)s->data.s_pointer;
    _ax_channel_seek (axel, channel, offset);
  }

  return offset;
}

static struct _AxMethods _ax_stream_methods = {
  ax_stream_clone,
  ax_stream_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  ax_stream_process,
  ax_stream_length,
  ax_stream_seek,
};

static AxStream
ax_stream_optimise (AXEL axel, AxStream stream)
{
  _ax_set_methods (axel, stream, &_ax_stream_methods);
  return stream;
}


/*
 * ax_stream_chunkfuncify (stream, count, func, data)
 *
 */
AxCount
ax_stream_chunkfuncify (AXEL axel, AxStream stream, AxCount count,
			AxChunkFunc func, void * data)
{
  AxCount n, minn = count, offset = ax_tell (axel, (AxBase)stream);
  CDSet s, channels = ax_get_channels (axel);
  AxChannel channel;

  for (s = channels; s; s = s->next) {
    ax_dprintf ("[ax_stream_chunkfuncify] thinking of channel %d\n", s->key);
  }

  if (stream == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  ax_dprintf ("[ax_stream_chunkfuncify] (%p, +%ld) @ %ld\n",
	      stream, count, offset);

  for (s = stream->channels; s; s = s->next) {
    if (cd_set_contains (axel, channels, s->key)) {
      channel = (AxChannel)s->data.s_pointer;
      n = ax_channel_chunkfuncify (axel, channel, minn, func, s->key, data);
      minn = MIN (n, minn);
    } else {
      ax_dprintf ("[ax_stream_chunkfuncify] channel %d not funced\n", s->key);
    }
  }

  ax_seek (axel, (AxBase)stream, offset + minn, SEEK_SET);

  return minn;
}

AxCount
ax_stream_chunkchunkfuncify (AXEL axel, AxStream src, AxStream dest,
			     AxCount count, AxChunkChunkFunc func,
			     void * data)
{
  AxCount n, minn = count;
  AxCount src_offset = ax_tell (axel, (AxBase)src);
  AxCount dest_offset = ax_tell (axel, (AxBase)dest);
  CDSet s, channels = ax_get_channels (axel);
  AxChannel sch, dch;

  if (dest == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  ax_dprintf ("[ax_stream_ccf...] (%p -> %p, +%ld), src @ %ld, dest @ %ld\n",
	      src, dest, count, src_offset, dest_offset);

  for (s = dest->channels; s; s = s->next) {
    if (cd_set_contains (axel, channels, s->key)) {
      dch = (AxChannel)s->data.s_pointer;
      sch = ax_stream_find_channel (axel, src, s->key);
      if (sch != AxNone) {
	n = ax_channel_chunkchunkfuncify (axel, sch, dch, count, func, s->key,
					  data);
	minn = MIN (n, minn);
      }
    }
  }

  ax_seek (axel, (AxBase)src, src_offset + minn, SEEK_SET);
  ax_seek (axel, (AxBase)dest, dest_offset + minn, SEEK_SET);

  return minn;
}

AxCount
ax_stream_chunkchunkchunkfuncify (AXEL axel, AxStream src1, AxStream src2,
				  AxStream dest, AxCount count,
				  AxChunkChunkChunkFunc func, void * data)
{
  AxCount n, minn = count;
  AxCount src1_offset = ax_tell (axel, (AxBase)src1);
  AxCount src2_offset = ax_tell (axel, (AxBase)src2);
  AxCount dest_offset = ax_tell (axel, (AxBase)dest);
  CDSet s, channels = ax_get_channels (axel);
  AxChannel s1ch, s2ch, dch;

  if (dest == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  for (s = dest->channels; s; s = s->next) {
    if (cd_set_contains (axel, channels, s->key)) {
      dch = (AxChannel)s->data.s_pointer;
      s1ch = ax_stream_find_channel (axel, src1, s->key);
      s2ch = ax_stream_find_channel (axel, src2, s->key);
      if (s1ch != AxNone && s2ch != AxNone) {
	n = ax_channel_chunkchunkchunkfuncify (axel, s1ch, s2ch, dch,
					       count, func, s->key, data);
	minn = MIN (n, minn);
      }
    }
  }

  ax_seek (axel, (AxBase)src1, src1_offset + minn, SEEK_SET);
  ax_seek (axel, (AxBase)src2, src2_offset + minn, SEEK_SET);
  ax_seek (axel, (AxBase)dest, dest_offset + minn, SEEK_SET);

  return minn;
}

/*
 * ax_stream_gain (axel, src, dest, count, gain)
 */
AxCount
ax_stream_gain (AXEL axel, AxStream stream, AxCount count, AxPCM gain)
{
  return ax_stream_chunkfuncify (axel, stream, count, _ax_chunk_gain, &gain);
}

/*
 * ax_stream_copy (axel, src, dest, count)
 *
 * Copy 'count' samples from 'src' to 'dest'
 */
AxCount
ax_stream_copy (AXEL axel, AxStream src, AxStream dest, AxCount count)
{
  ax_dprintf ("[ax_stream_copy] (%p -> %p, +%ld)\n", src, dest, count);

  return ax_stream_chunkchunkfuncify (axel, src, dest, count,
				      _ax_chunk_copy, NULL);
}

/*
 * ax_stream_write (axel, stream, count, data)
 *
 * Write 'count' samples of 'data' to the stream 'stream'.
 */
/* XXX: this is a wrapper to obsolete ax_stream_write above */
AxCount
ax_stream_write (AXEL axel, AxStream stream, AxCount count, AxStream data)
{
  ax_dprintf ("[ax_stream_write] (%p -> %p, +%ld)\n", data, stream, count);

  if (data == AxNone)
    return ax_stream_write0 (axel, stream, count);

  return ax_stream_copy (axel, data, stream, count);
}

/*
 * ax_stream_mix (src, dest, count)
 *
 * Mix 'count' samples from 'src' into 'dest'.
 */
AxCount
ax_stream_mix (AXEL axel, AxStream src, AxStream dest, AxCount count)
{
  ax_dprintf ("[ax_stream_mix] (%p -> %p, +%ld)\n", src, dest, count);

  return ax_stream_chunkchunkfuncify (axel, src, dest, count,
				      _ax_chunk_add_inplace, NULL);
}

/*
 * ax_stream_mult (src, dest, count)
 *
 * Multiply 'count' samples of 'src' into 'dest'.
 */
AxCount
ax_stream_mult (AXEL axel, AxStream src, AxStream dest, AxCount count)
{
  ax_dprintf ("[ax_stream_mult] (%p -> %p, +%ld)\n", src, dest, count);

  return ax_stream_chunkchunkfuncify (axel, src, dest, count,
				      _ax_chunk_mult_inplace, NULL);
}

/*
 * ax_stream_fade (src, dest, count)
 *
 * Fade 'count' samples of 'dest' by values in 'src'.
 */
AxCount
ax_stream_fade (AXEL axel, AxStream src, AxStream dest, AxCount count)
{
  ax_dprintf ("[ax_stream_fade] (%p -> %p, +%ld)\n", src, dest, count);

  return ax_stream_chunkchunkfuncify (axel, src, dest, count,
				      _ax_chunk_fade_inplace, NULL);
}

/*
 * ax_stream_blend (axel, src, blend, dest, count)
 *
 * Blend 'count' samples of 'src' into 'dest' by amounts in 'blend'.
 */
AxCount
ax_stream_blend (AXEL axel, AxStream src, AxStream blend, AxStream dest,
		 AxCount count)
{
  ax_dprintf ("[ax_stream_blend] (%p -> (%p) -> %p, +%ld)\n", src, blend,
	      dest, count);

  return ax_stream_chunkchunkchunkfuncify (axel, src, blend, dest, count,
					   _ax_chunk_blend_inplace, NULL);
}


/*
 * ax_streams_mix (streams, dest, count)
 *
 * Mix 'count' samples from all streams in list 'streams' together into
 * 'dest'.
 */
AxCount
ax_streams_mix (AXEL axel, CDList streams, AxStream dest, AxCount count)
{
  CDList sl;
  CDSet s;
  AxChannel sch, dch;
  AxStream stream;
  AxCount dest_start = ax_tell (axel, (AxBase)dest);
  AxCount stream_start;

  if (dest == AxNone) {
    ax_set_error (axel, AX_ERROR_NOENTITY);
    return -1;
  }

  for (s = dest->channels; s; s = s->next) {
    dch = (AxChannel)s->data.s_pointer;
    for (sl = streams; sl; sl = sl->next) {
      stream = (AxStream)sl->data.s_pointer;
      stream_start = ax_tell (axel, (AxBase)stream);
      sch = ax_stream_find_channel (axel, stream, s->key);
      if (sch != AxNone) {
        _ax_channel_seek (axel, dch, dest_start);
        ax_channel_mix (axel, sch, dch, count);
        ax_seek (axel, (AxBase)stream, stream_start, SEEK_SET);
      }
    }
  }

  for (sl = streams; sl; sl = sl->next) {
    stream = (AxStream)sl->data.s_pointer;
    stream_start = ax_tell (axel, (AxBase)stream);
    ax_seek (axel, (AxBase)stream, stream_start + count, SEEK_SET);
  }

  ax_seek (axel, (AxBase)dest, dest_start+count, SEEK_SET);

  return count;
}

/*
 * ax_stream_interleave_2 (axel, stream, name1, name2, dest, count)
 *
 * Interleave 'count' frames of the channels named 'name1' and 'name2' in
 * 'stream', placing the resulting PCM data in the memory region pointed
 * to by 'dest'.
 */
AxCount
ax_stream_interleave_2 (AXEL axel, AxStream stream,
			int name1, int name2,
			AxPCM * dest, AxCount count)
{
  AxChannel channel1, channel2;
  AxCount n = 0;

  channel1 = ax_stream_find_channel (axel, stream, name1);
  channel2 = ax_stream_find_channel (axel, stream, name2);

  if (channel1 != AxNone && channel2 != AxNone) {
      n = ax_channel_chunkchunkfuncify (axel, channel1, channel2,
					count, _ax_chunk_interleave_2,
					0, dest);
  }

  if (n > 0) ax_seek (axel, (AxBase)stream, n, SEEK_CUR);

  return n;
}

/*
 * ax_stream_deinterleave_stereo (axel, stream, name1, name2, src, count)
 *
 * Deinterleave 'count' frames from the memory region pointed to by 'src',
 * placing the resulting channels into those named 'name1' and 'name2' in
 * 'stream'.
 */
AxCount
ax_stream_deinterleave_2 (AXEL axel, AxStream stream,
			  int name1, int name2,
			  AxPCM * src, AxCount count)
{
  AxChannel channel1, channel2;
  AxCount n = 0;

  channel1 = ax_stream_find_channel (axel, stream, name1);
  channel2 = ax_stream_find_channel (axel, stream, name2);

  if (channel1 != AxNone && channel2 != AxNone) {
      n = ax_channel_chunkchunkfuncify (axel, channel1, channel2,
					count, _ax_chunk_deinterleave_2,
					0, src);
  }

  if (n > 0) ax_seek (axel, (AxBase)stream, n, SEEK_CUR);

  return n;
}
