/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxTime: A generic time abstraction for sequencing information.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#define __AXEL__
#include "axel.h"

AxTime
ax_time_zero (AxTimeType type)
{
  return _ax_time_zero (type);
}

AxTime
ax_time_invalid (AxTimeType type)
{
  return _ax_time_invalid (type);
}

int
ax_time_is_invalid (AxTimeType type, AxTime time)
{
  return _ax_time_is_invalid (type, time);
}

AxTime
ax_time_add (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_add (type, t1, t2);
}

AxTime
ax_time_sub (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_sub (type, t1, t2);
}

AxTime
ax_time_min (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_min (type, t1, t2);
}

AxTime
ax_time_max (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_max (type, t1, t2);
}

int
ax_time_eq (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_eq (type, t1, t2);
}

int
ax_time_gt (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_gt (type, t1, t2);
}

int
ax_time_lt (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_lt (type, t1, t2);
}

int
ax_time_ge (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_ge (type, t1, t2);
}

int
ax_time_le (AxTimeType type, AxTime t1, AxTime t2)
{
  return _ax_time_le (type, t1, t2);
}


static float
ax_samples_to_seconds (AxCount samples, AxSamplerate samplerate)
{
  return ((float)samples / samplerate);
}

static int
ax_samples_to_beat24s (AxCount samples, AxSamplerate samplerate,
		       AxTempo tempo)
{
  return ((int)((float)samples * tempo * 24.0 / (samplerate * 60.0)));
}

static AxCount
ax_seconds_to_samples (float seconds, AxSamplerate samplerate)
{
  return (AxCount)(seconds * samplerate);
}

static int
ax_seconds_to_beat24s (float seconds, AxTempo tempo)
{
  return (int)(seconds * tempo * 24.0 / 60.0);
}

static AxCount
ax_beat24s_to_samples (int beat24s, AxSamplerate samplerate, AxTempo tempo)
{
  return (AxCount)(beat24s * samplerate * 60.0 / (tempo * 24.0));
}

static float
ax_beat24s_to_seconds (int beat24s, AxTempo tempo)
{
  return ((float)beat24s * 60.0 / (tempo * 24.0));
}


AxTime
ax_time_convert (AXEL axel, AxTime time, AxTimeType old_type,
		 AxTimeType new_type)
{
  AxSamplerate samplerate;
  AxTempo tempo;

  if (old_type == new_type) return time;

  samplerate = ax_get_samplerate (axel);
  tempo = ax_get_tempo (axel);

  switch (old_type) {
  case AX_TIME_SAMPLES:
    switch (new_type) {
    case AX_TIME_SECONDS:
      return (AxTime) ax_samples_to_seconds (time.samples, samplerate);
      break;
    case AX_TIME_BEAT24S:
      return (AxTime) ax_samples_to_beat24s (time.samples, samplerate, tempo);
      break;
    default: break;
    }
    break;
  case AX_TIME_SECONDS:
    switch (new_type) {
    case AX_TIME_SAMPLES:
      return (AxTime) ax_seconds_to_samples (time.seconds, samplerate);
      break;
    case AX_TIME_BEAT24S:
      return (AxTime) ax_seconds_to_beat24s (time.seconds, tempo);
      break;
    default: break;
    }
    break;
  case AX_TIME_BEAT24S:
    switch (new_type) {
    case AX_TIME_SAMPLES:
      return (AxTime) ax_beat24s_to_samples (time.beat24s, samplerate, tempo);
      break;
    case AX_TIME_SECONDS:
      return (AxTime) ax_beat24s_to_seconds (time.beat24s, tempo);
      break;
    default: break;
    }
    break;
  default:
    /* Cannot convert invalid time type to anything */
    break;
  }

  return ax_time_invalid (new_type);
}
