/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * AxTrack: An AxLayer mixing abstraction.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A track is contained within a deck. A track contains a number of
 * layers which are mixed in series.
 *
 * Invariants
 * ----------
 *
 * A track must be contained within a deck.
 */

#define __AXEL__
#include "axel.h"

/* Optimisation dependencies: optimise on change of nr. layers */
static AxTrack ax_track_optimise (AXEL axel, AxTrack track);

static void
ax_track_replace_mixstreams (AXEL axel, AxTrack track)
{
  AxCount mixlength = _ax_base_get_mixlength (axel, track);

  if (track->_mixstream_a != AxNone)
    ax_destroy (axel, (AxBase)track->_mixstream_a);
  if (track->_mixstream_b != AxNone)
    ax_destroy (axel, (AxBase)track->_mixstream_b);

  track->_mixstream_a = ax_stream_new_contiguous (axel, mixlength);
  track->_mixstream_b = ax_stream_new_contiguous (axel, mixlength);
}

static AxBase
ax_track_init (AXEL axel, AxBase base)
{
  AxTrack track = (AxTrack)base;
  track->gain = 1.0;
  track->layers = cd_list_new (axel);
  ax_track_replace_mixstreams (axel, track);
  ax_track_optimise (axel, track);
  return (AxBase)track;
}

static AxTrack
ax_track_new (AXEL axel)
{
  return (AxTrack) ax_base_new_subclass (axel, sizeof (struct _AxTrack));
}

AxBase
ax_track_clone (AXEL axel, AxBase base)
{
  AxTrack track = (AxTrack)base;
  AxTrack new_track = ax_track_new (axel);

  new_track->gain = track->gain;
  new_track->layers = cd_list_clone (axel, track->layers,
				     (CDCloneFunc)ax_layer_clone);
  ax_track_optimise (axel, track);

  return (AxBase)new_track;
}

static int
ax_track_destroy (AXEL axel, AxBase base)
{
  AxTrack track = (AxTrack)base;
  ax_destroy_list (axel, track->layers);
  ax_free (track);
  return 0;
}

static int
ax_track_ready (AXEL axel, AxBase base)
{
  return (ax_base_encompasses_mixlength (axel, base) &&
	  ax_base_encompasses_channels (axel, base));
}

static AxBase
ax_track_prepare (AXEL axel, AxBase base)
{
  AxTrack track = (AxTrack)base;
  ax_track_replace_mixstreams (axel, track);
  return base;
}

AxTrack
ax_add_track (AXEL axel, AxDeck deck)
{
  AxTrack track;

  track = ax_track_new (axel);
  track->deck = deck;
  ax_track_init (axel, (AxBase)track);
  _ax_deck_add_track (axel, deck, track);

  return track;
}

static AxCount
ax_track_length (AXEL axel, AxBase base)
{
  AxTrack track = (AxTrack)base;
  AxCount length, maxlength = 0;
  CDList l;
  AxLayer layer;

  for (l = track->layers; l; l = l->next) {
    layer = (AxLayer)l->data.s_pointer;
    length = ax_length (axel, (AxBase)layer);
    ax_dprintf ("[ax_track_length] found layer %p length %ld\n",
		layer, length);
    maxlength = MAX (maxlength, length);
  }

  return maxlength;
}

AxPCM
ax_track_set_gain (AXEL axel, AxTrack track, AxPCM gain)
{
  AxPCM old = track->gain;
  track->gain = gain;
  return old;
}

AxPCM
ax_track_get_gain (AXEL axel, AxTrack track)
{
  return track->gain;
}

void
ax_remove_track (AXEL axel, AxTrack track)
{
  AxDeck deck = track->deck;

  if (deck)
    _ax_deck_remove_track (axel, deck, track);
}

AxDeck
ax_track_get_deck (AXEL axel, AxTrack track)
{
  return track->deck;
}

/*
 * _ax_track_add_layer_above (axel, track, layer, above)
 *
 * Adds 'layer' above 'above'. If above is AxNone, adds 'layer' on top
 * of 'track'.
 */
AxLayer
_ax_track_add_layer_above (AXEL axel, AxTrack track, AxLayer layer,
			   AxLayer above)
{
  layer->track = track;
  if (above == AxNone)
    above =
      (AxLayer)(cd_list_last (axel, track->layers, CD_TYPE_POINTER)).s_pointer;
  track->layers = cd_list_add_after (axel, track->layers, CD_TYPE_POINTER,
				     CD_POINTER(layer), CD_POINTER(above));
  ax_track_optimise (axel, track);
  return layer;
}

AxLayer
_ax_track_remove_layer (AXEL axel, AxTrack track, AxLayer layer)
{
  track->layers = cd_list_remove (axel, track->layers, CD_TYPE_POINTER,
				  CD_POINTER(layer));
  ax_track_optimise (axel, track);
  return layer;
}

/*
 * gets layer above 'above'. If above is AxNone, returns topmost layer
 */
AxLayer
_ax_track_get_layer_above (AXEL axel, AxTrack track, AxLayer above)
{
  CDList above_item;
  AxLayer layer;

  if (track->layers == NULL) return AxNone;

  if (above == AxNone) {
    layer =
      (AxLayer)(cd_list_last (axel, track->layers, CD_TYPE_POINTER)).s_pointer;
  } else {
    above_item = cd_list_find (axel, track->layers, CD_TYPE_POINTER,
			       CD_POINTER(above));
    if (above_item == NULL) return AxNone;
    above_item = above_item->next;
    if (above_item == NULL) return AxNone;
    layer = (AxLayer) above_item->data.s_pointer;
  }

  return layer;
}

/*
 * gets layer below 'below'. If below is AxNone, returns lowest layer
 */
AxLayer
_ax_track_get_layer_below (AXEL axel, AxTrack track, AxLayer below)
{
  CDList below_item;
  AxLayer layer;

  if (track->layers == NULL) return AxNone;

  if (below == AxNone) {
    layer = (AxLayer) track->layers->data.s_pointer;
  } else {
    below_item = cd_list_find (axel, track->layers, CD_TYPE_POINTER,
			       CD_POINTER(below));
    if (below_item == NULL) return AxNone;
    below_item = below_item->prev;
    if (below_item == NULL) return AxNone;
    layer = (AxLayer) below_item->data.s_pointer;
  }

  return layer;
}


static AxCount
ax_track_process (AXEL axel, AxBase base, AxCount count, AxStream input,
                  AxStream output)
{
  AxTrack track = (AxTrack)base;
  CDList l;
  AxBase layer;
  AxStream si, so, swap_stream;
  AxCount remaining = count, processed = 0, n = 0;
  AxCount mixlength = _ax_base_get_mixlength (axel, track);

  ax_dprintf ("PROCESS TRACK (%p, +%ld, %p -> %p) @ %ld\n",
	      track, count, input, output, ax_tell (axel, base));

  if (track->layers == AxNone) {
    ax_set_error (axel, AX_ERROR_NOOP);
    return 0;
  }

  while (remaining > 0) {
    l = track->layers;
    si = input;
    so = track->_mixstream_a;
    n = MIN (remaining, mixlength);

    while (l) {
      layer = (AxBase)l->data.s_pointer;

      if (si == input) {
	swap_stream = track->_mixstream_b;
      } else {
	swap_stream = si;
	ax_seek (axel, (AxBase)si, 0, SEEK_SET);
      }
      
      if (l->next == AxNone) {
	so = output;
      } else {
	ax_seek (axel, (AxBase)so, 0, SEEK_SET);
      }
      
      n = ax_process (axel, layer, n, si, so);
      
      l = l->next;

      /* swap si, st using swap_stream as evaluated above */
      si = so; so = swap_stream;
    }
    remaining -= n;
    processed += n;
  }

  ax_dprintf ("[ax_track_process] processed %ld\n", processed);

  return processed;
}

static AxCount
ax_track_twolayer_process (AXEL axel, AxBase base, AxCount count,
			   AxStream input, AxStream output)
{
  AxTrack track = (AxTrack)base;
  CDList l;
  AxBase layer1, layer2;
  AxCount remaining = count, processed = 0, n = 0;
  AxStream mix = track->_mixstream_a;
  AxCount current_offset = ax_tell (axel, base);
  AxCount mixlength = _ax_base_get_mixlength (axel, track);

  ax_dprintf ("PROCESS TRACK [twolayer] (%p, +%ld, %p -> %p) @ %ld\n",
	      track, count, input, output, current_offset);

  l = track->layers;
  layer1 = (AxBase)l->data.s_pointer;
  ax_seek (axel, (AxBase)layer1, current_offset, SEEK_SET);
  
  l = l->next;
  layer2 = (AxBase)l->data.s_pointer;
  ax_seek (axel, (AxBase)layer2, current_offset, SEEK_SET);

  while (remaining > 0) {
    n = MIN (remaining, mixlength);

    ax_seek (axel, (AxBase)mix, 0, SEEK_SET);
    n = ax_process (axel, layer1, n, input, mix);

    ax_seek (axel, (AxBase)mix, 0, SEEK_SET);
    n = ax_process (axel, layer2, n, mix, output);

    remaining -= n;
    processed += n;
  }

  ax_dprintf ("*** PRE-SEEK: track @ %ld\n", ax_tell (axel, base));

  /*ax_seek (axel, base, current_offset + processed, SEEK_SET);*/

  ax_dprintf ("*** POST-SEEK: track @ %ld\n", ax_tell (axel, base));

  ax_dprintf ("[ax_track_twolayer_process] processed %ld\n", processed);

  return processed;
}

static AxCount
ax_track_onelayer_process (AXEL axel, AxBase base, AxCount count,
			   AxStream input, AxStream output)
{
  AxTrack track = (AxTrack)base;
  CDList l = track->layers;
  AxBase layer = (AxBase)l->data.s_pointer;
  AxCount n;

  ax_dprintf ("PROCESS TRACK [onelayer] (%p, +%ld, %p -> %p) @ %ld\n",
	      track, count, input, output, ax_tell (axel, base));

  n = ax_process (axel, layer, count, input, output);

  ax_dprintf ("[ax_track_onelayer_process] processed %ld\n", n);

  return n;
}

static AxCount
ax_track_seek (AXEL axel, AxBase base, AxCount offset)
{
  AxTrack track = (AxTrack)base;
  CDList l;
  AxLayer layer;

  for (l = track->layers; l; l = l->next) {
    layer = (AxLayer)l->data.s_pointer;
    ax_seek (axel, (AxBase)layer, offset, SEEK_SET);
  }

  return offset;
}

static int
ax_track_flush (AXEL axel, AxBase base)
{
  AxTrack track = (AxTrack)base;
  CDList l;
  AxLayer layer;

  for (l = track->layers; l; l = l->next) {
    layer = (AxLayer)l->data.s_pointer;
    ax_flush (axel, (AxBase)layer);
  }

  return 0;
}

static struct _AxMethods _ax_track_empty_methods = {
  ax_track_clone,   /* clone */
  ax_track_destroy, /* destroy */
  ax_track_ready,   /* ready */
  ax_track_prepare, /* prepare */
  ax_null_process,  /* process */
  ax_null_length,   /* length */
  ax_null_seek,     /* seek */
  NULL,             /* flush */
};

static struct _AxMethods _ax_track_methods = {
  ax_track_clone,   /* clone */
  ax_track_destroy, /* destroy */
  ax_track_ready,   /* ready */
  ax_track_prepare, /* prepare */
  ax_track_process, /* process */
  ax_track_length,  /* length */
  ax_track_seek,    /* seek */
  ax_track_flush,            /* flush */
};

static struct _AxMethods _ax_track_onelayer_methods = {
  ax_track_clone,            /* clone */
  ax_track_destroy,          /* destroy */
  ax_track_ready,            /* ready */
  ax_track_prepare,          /* prepare */
  ax_track_onelayer_process, /* process */
  ax_track_length,           /* length */
  ax_track_seek,             /* seek */
  ax_track_flush,            /* flush */
};

static struct _AxMethods _ax_track_twolayer_methods = {
  ax_track_clone,            /* clone */
  ax_track_destroy,          /* destroy */
  ax_track_ready,            /* ready */
  ax_track_prepare,          /* prepare */
  ax_track_twolayer_process, /* process */
  ax_track_length,           /* length */
  ax_track_seek,             /* seek */
  ax_track_flush,            /* flush */
};

static AxTrack
ax_track_optimise (AXEL axel, AxTrack track)
{
  AxCount nr_layers = cd_list_length (axel, track->layers);

  switch (nr_layers) {
  case 0:
    _ax_set_methods (axel, track, &_ax_track_empty_methods); break;
  case 1:
    _ax_set_methods (axel, track, &_ax_track_onelayer_methods); break;
  case 2:
    _ax_set_methods (axel, track, &_ax_track_twolayer_methods); break;
  default:
      _ax_set_methods (axel, track, &_ax_track_methods); break;
  }

  return track;
}
