/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * axel_plugin.h -- AXEL internal data types and functions.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#ifndef __AXEL_PLUGIN_H__
#define __AXEL_PLUGIN_H__

#define DEBUG

#if defined(__AXEL_PLUGIN__) || defined(__AXEL__)

#include <stdio.h>

#include "syssimple.h"
#include "ctxdata.h"

#define AXEL_PLUGIN_API_MAJOR 1
#define AXEL_PLUGIN_API_MINOR 0
#define AXEL_PLUGIN_API_REVISION 0

typedef struct _AxMetaAuthor * AxMetaAuthor;
typedef struct _AxMetaText * AxMetaText;

typedef struct _AxParameterScheme * AxParameterScheme;
typedef struct _AxParameterRange * AxParameterRange;

typedef struct _AxNamedParameter * AxNamedParameter;

typedef union _AxConstraint AxConstraint;

typedef struct _AxMethods * AxMethods;;

#define AX_PLUGIN_WRITEABLE 1<<0
#define AX_PLUGIN_SEEKABLE  1<<1
#define AX_PLUGIN_CACHEABLE 1<<2
#define AX_PLUGIN_CAUSAL    1<<3

/* An base of a plugin */
typedef struct _AxPlugin * AxPlugin;
/*typedef struct _AxBase * AxBase;*/

typedef struct _AxChunk * AxChunk;

#if defined (__AXEL__)
#include "axel_private.h"
#else
typedef AxOpaque AXEL;
typedef AxOpaque AxPoint;
typedef AxOpaque AxEnvelope;
typedef AxOpaque AxStream;
typedef AxOpaque AxChannel;
typedef AxOpaque AxDeck;
typedef AxOpaque AxTrack;
typedef AxOpaque AxLayer;
typedef AxOpaque AxSound;
typedef AxOpaque AxSquareTone;
typedef AxOpaque AxMonitor;
#endif


typedef CDList (*AxPluginInitFunc) (AXEL axel);

typedef AxBase (*AxInitFunc) (AXEL axel, AxBase base, CDSet parameters);

typedef CDSet (*AxSuggestFunc) (AXEL axel, AxPlugin plugin,
				CDSet parameters,
				void * plugin_data);

typedef AxBase (*AxCloneFunc) (AXEL axel, AxBase base);
typedef int (*AxDestroyFunc) (AXEL axel, AxBase base);
typedef int (*AxReadyFunc) (AXEL axel, AxBase base);
typedef AxBase (*AxPrepareFunc) (AXEL axel, AxBase base);
typedef AxCount (*AxSeekFunc) (AXEL axel, AxBase base, AxCount count);
typedef AxCount (*AxLengthFunc) (AXEL axel, AxBase base);
typedef AxCount (*AxProcessFunc) (AXEL axel, AxBase base, AxCount count,
                                  AxStream input, AxStream output);
typedef int (*AxFlushFunc) (AXEL axel, AxBase base);

#define AX_FLAGS_NONE (0)

#define AX_AUTHOR(n,e) ((struct _AxMetaAuthor){(n),(e)})
#define AX_ONE_AUTHOR(n,e) CD_SINGLETON_LIST(CD_POINTER(&(AX_AUTHOR((n),(e)))))

/* ChunkChunkFuncs and all the rest */

/* AxChunkFunc: a function to apply to one chunk */
typedef AxCount (*AxChunkFunc) (AXEL axel, AxChunk chunk, AxCount offset,
				AxCount count, int channelname, void * data);

/* AxChunkChunkFunc: a function to apply between two chunks :) */
typedef AxCount (*AxChunkChunkFunc) (AXEL axel,
				     AxChunk src, AxCount src_offset,
				     AxChunk dest, AxCount dest_offset,
				     AxCount count, int channelname,
				     void * data);

/* AxChunkChunkChunkFunc: a function to apply between THREE chunks !! */
typedef AxCount (*AxChunkChunkChunkFunc) (AXEL axel,
					  AxChunk src1, AxCount src1_offset,
					  AxChunk src2, AxCount src2_offset,
					  AxChunk dest, AxCount dest_offset,
					  AxCount count, int channelname,
					  void * data);

struct _AxMetaAuthor {
  char * name;
  char * email;
};

struct _AxMetaText {
  char * identifier;
  char * category;
  char * description;
  char * copyright;
  char * url;
  CDList authors;
};

struct _AxParameterRange {
  AxFlags valid_mask;
  AxParameter lower;
  AxParameter upper;
  AxParameter step;
};

struct _AxNamedParameter {
  char * name;
  AxParameter parameter;
};

#define AX_NAMED_PARAMETER(n,p) (&((struct _AxNamedParameter){(n),(p)}))

union _AxConstraint {
  CDList list; /* list of AxNamedParameter */
  AxParameterRange range;
};

struct _AxParameterScheme {
  char * name;
  char * description;
  AxParameterType type;
  AxConstraintType constraint_type;
  AxConstraint constraint;
  AxFlags hints;
};

struct _AxPlugin {
  AxMetaText metatext;
  AxFlags flags;
  CDSet init_scheme;
  AxInitFunc init;
  CDSet process_scheme;
  AxSuggestFunc suggest;
  void * plugin_data;
};

struct _AxMethods {
  AxCloneFunc clone;
  AxDestroyFunc destroy;
  AxReadyFunc ready;
  AxPrepareFunc prepare;
  AxProcessFunc process;
  AxLengthFunc length;
  AxSeekFunc seek;
  AxFlushFunc flush;
};


struct _AxChunk {
  AxCount start_index;
  AxCount length;
  AxPCM * data;
};


/* debug */
void ax_dprintf (const char * fmt, ...);

/* AXEL, ax_context */

AxError ax_set_error (AXEL axel, AxError error);


/* ax_base */
AxBase ax_base_new (AXEL axel);

AxPlugin ax_base_set_plugin (AXEL axel, AxBase base, AxPlugin plugin);
AxPlugin ax_base_get_plugin (AXEL axel, AxBase base);
AxMethods ax_base_set_methods (AXEL axel, AxBase base, AxMethods methods);
void * ax_base_set_instance_data (AXEL axel, AxBase base, void * data);
void * ax_base_get_instance_data (AXEL axel, AxBase base);

AxCount ax_base_get_mixlength (AXEL axel, AxBase base);
AxSamplerate ax_base_get_samplerate (AXEL axel, AxBase base);
AxTempo ax_base_get_tempo (AXEL axel, AxBase base);
CDSet ax_base_get_channels (AXEL axel, AxBase base);

int ax_base_has_samplerate (AXEL axel, AxBase base);
int ax_base_has_tempo (AXEL axel, AxBase base);
int ax_base_encompasses_mixlength (AXEL axel, AxBase base);
int ax_base_encompasses_channels (AXEL axel, AxBase base);


/* ax_meta */
AxMetaText ax_meta_text_new (AXEL axel);
void ax_meta_text_free (AXEL axel, AxMetaText mt);
AxMetaText ax_get_meta_text (AXEL axel, AxBase base);
AxMetaText ax_set_meta_text (AXEL axel, AxBase base, AxMetaText mt);


/* ax_null */
AxCount ax_null_length (AXEL axel, AxBase base);
AxCount ax_null_process (AXEL axel, AxBase base, AxCount count,
			 AxStream input, AxStream output);
AxCount ax_null_seek (AXEL axel, AxBase base, AxCount offset);


/* ax_stream */
AxCount ax_stream_chunkfuncify (AXEL axel, AxStream stream, AxCount count,
				AxChunkFunc func, void * data);
AxCount ax_stream_chunkchunkfuncify (AXEL axel, AxStream src, AxStream dest,
				     AxCount count, AxChunkChunkFunc func,
				     void * data);
AxCount ax_stream_chunkchunkchunkfuncify (AXEL axel,
					  AxStream src1, AxStream src2,
					  AxStream dest, AxCount count,
					  AxChunkChunkChunkFunc func,
					  void * data);

/* AxChannel */
AxCount ax_channel_chunkfuncify (AXEL axel, AxChannel channel, AxCount count,
				 AxChunkFunc func, int channelname,
				 void * data);
AxCount ax_channel_chunkchunkfuncify (AXEL axel, AxChannel src, AxChannel dest,
				      AxCount count, AxChunkChunkFunc func,
				      int channelname, void * data);
AxCount ax_channel_chunkchunkchunkfuncify (AXEL axel,
					   AxChannel src1, AxChannel src2,
					   AxChannel dest, AxCount count,
					   AxChunkChunkChunkFunc func,
					   int channelname, void * data);
/* AxPCM */

AxCount _ax_pcm_clear_region (AxPCM * data, AxCount count, void * unused);
AxCount _ax_pcm_set (AxPCM * data, AxPCM value, AxCount count);
AxCount _ax_pcm_gain (AxPCM * data, AxCount count,
		      /* (AxPCM *) */ void * gain);
AxCount _ax_pcm_copy (AxPCM * src, AxPCM * dest, AxCount count, void * unused);
AxCount _ax_pcm_add (AxPCM * src, AxPCM * dest, AxCount count, void * unused);
AxCount _ax_pcm_mult (AxPCM * src, AxPCM * dest, AxCount count, void * unused);
AxCount _ax_pcm_fade (AxPCM * src, AxPCM * dest, AxCount count, void * unused);
AxCount _ax_pcm_interleave_2 (AxPCM * src1, AxPCM * src2, AxCount count,
			      void * data);
AxCount _ax_pcm_deinterleave_2 (AxPCM * dest1, AxPCM * dest2, AxCount count,
				void * data);
AxCount _ax_pcm_blend (AxPCM * src, AxPCM * blend, AxPCM * dest,
		       AxCount count, void * unused);
AxCount _ax_pcm_write_linear (AxPCM * data, AxCount x1, AxPCM y1,
			      AxCount x2, AxPCM y2,
			      AxCount offset, AxCount count);

#endif /* defined(__AXEL__) */

#endif /* __AXEL_PLUGIN_H__ */
