/*
 * AXEL -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * axel_private.h -- AXEL internal data types and functions.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#ifndef __AXEL_PRIVATE_H__
#define __AXEL_PRIVATE_H__

/*#define DEBUG*/

#if defined(__AXEL__)

#include <stdio.h>

#include "syssimple.h"
#include "ctxdata.h"

#include "axel_plugin.h"

/* Max line length for error messages etc. */
#define AXEL_MAXLINE 4096

#define AXEL_DEFAULT_MIXLENGTH 1024
#define AXEL_DEFAULT_SAMPLERATE 44100
#define AXEL_DEFAULT_TEMPO 120

typedef struct _AxThreadContext * AxThreadContext;
typedef struct _AxEnvironment * AxEnvironment;
typedef struct _AxContext * AxContext;

typedef AxThreadContext AXEL;

typedef struct _AxPoint * AxPoint;
typedef struct _AxEnvelope * AxEnvelope;
/*typedef struct _AxChunk * AxChunk;*/
typedef struct _AxChannel * AxChannel;
typedef struct _AxStream * AxStream;
typedef struct _AxDeck * AxDeck;
typedef struct _AxTrack * AxTrack;
typedef struct _AxLayer * AxLayer;
typedef struct _AxSound * AxSound;


struct _AxThreadContext {
  AxError last_error;
  AxContext context;
  AxEnvironment environment;
};

struct _AxEnvironment {
  AxCount refcount;
  CDList plugins;
  CDList bases;
  int purging;
};

struct _AxContext {
  AxSamplerate samplerate;
  AxTempo tempo;
  CDSet channels;
  AxCount mixlength;
};

struct _AxBase {
  AxPlugin plugin;
  AxMethods methods;
  CDSet parameters;
  AxCount offset; /* current position */
  struct _AxContext context_limit;
  void * instance_data;
};

struct _AxPoint {
  AxTime time;
  AxPCM value;
};

struct _AxEnvelope {
  struct _AxBase base;
  AxEnvelopeType type;
  AxTimeType timetype;
  CDList points;
  CDList _current_point_item;
  AxCount _current_offset;
};

/* XXX: multichannel envelopes ? */


struct _AxStream {
  struct _AxBase base;
  CDSet channels;
};

struct _AxChannel {
  CDList chunks;
  AxCount _current_offset;
  CDList _current_chunk;
};

struct _AxDeck {
  struct _AxBase base;
  CDList tracks;
  AxStream _mixstream;
};

struct _AxTrack {
  struct _AxBase base;
  AxDeck deck;
  AxPCM gain;
  CDList layers;
  AxStream _mixstream_a;
  AxStream _mixstream_b;
};

struct _AxLayer {
  struct _AxBase base;
  AxTrack track;
  AxTimeType timetype;
  CDList sounds;
  AxTime _current_time;
  CDList _current_sound_item;
};

struct _AxSound {
  struct _AxBase base;
  AxBase source;
  AxBase rate_envelope;
  AxBase gain_envelope;
  AxBase blend_envelope;
  AxLayer layer;
  AxTime start_time; /* position in layer */
  AxTime duration; /* maximum time length */
  AxCount cutin;  /* start offset into sound source */
  AxCount cutlength;
  AxCount _current_source_offset;
  AxStream _rate_envstream;
  AxStream _gain_envstream;
  AxStream _blend_envstream;
};

typedef struct _AxSquareToneChannel * AxSquareToneChannel;
typedef struct _AxSquareTone * AxSquareTone;

struct _AxSquareToneChannel {
  AxCount _cycle_offset;
};

struct _AxSquareTone {
  struct _AxBase base;
  float frequency;
  CDSet channels;
};

typedef struct _AxMonitor * AxMonitor;

#define AX_MONITOR_BUFFERLEN 2048

struct _AxMonitor {
  struct _AxBase base;
  AxPCM databuffer[AX_MONITOR_BUFFERLEN];
  short playbuffer[AX_MONITOR_BUFFERLEN];
  int file;
  int mode;
  int mask;
  int format;
  int stereo;
  int frequency;
  int numfrags;
  int fragsize;
};

#define _ax_time_zero(t) (AxTime)\
  (((t)==AX_TIME_SAMPLES) ? ((AxCount)0) : \
  (((t)==AX_TIME_SECONDS) ? ((float)0.0) : \
                            ((int)0)))

#define _ax_time_invalid(t) (AxTime)\
  (((t)==AX_TIME_SAMPLES) ? ((AxCount)-1) : \
  (((t)==AX_TIME_SECONDS) ? ((float)-1.0) : \
                            ((int)-1)))

#define _ax_time_is_invalid(t,ti) \
  (((t)==AX_TIME_SAMPLES) ? ((ti).samples < (AxCount)0) : \
  (((t)==AX_TIME_SECONDS) ? ((ti).seconds < (float)0.0) : \
                            ((ti).beat24s < (int)0)))

#define _ax_time_add(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? AX_SAMPLES((t1).samples + (t2).samples) : \
  (((t)==AX_TIME_SECONDS) ? AX_SECONDS((t1).seconds + (t2).seconds) : \
                            AX_BEAT24S((t1).beat24s + (t2).beat24s)))

#define _ax_time_sub(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? AX_SAMPLES((t1).samples - (t2).samples) : \
  (((t)==AX_TIME_SECONDS) ? AX_SECONDS((t1).seconds - (t2).seconds) : \
                            AX_BEAT24S((t1).beat24s - (t2).beat24s)))

#define _ax_time_eq(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? ((t1).samples == (t2).samples) : \
  (((t)==AX_TIME_SECONDS) ? ((t1).seconds == (t2).seconds) : \
                            ((t1).beat24s == (t2).beat24s)))

#define _ax_time_gt(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? ((t1).samples > (t2).samples) : \
  (((t)==AX_TIME_SECONDS) ? ((t1).seconds > (t2).seconds) : \
                            ((t1).beat24s > (t2).beat24s)))

#define _ax_time_lt(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? ((t1).samples < (t2).samples) : \
  (((t)==AX_TIME_SECONDS) ? ((t1).seconds < (t2).seconds) : \
                            ((t1).beat24s < (t2).beat24s)))

#define _ax_time_ge(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? ((t1).samples >= (t2).samples) : \
  (((t)==AX_TIME_SECONDS) ? ((t1).seconds >= (t2).seconds) : \
                            ((t1).beat24s >= (t2).beat24s)))

#define _ax_time_le(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? ((t1).samples <= (t2).samples) : \
  (((t)==AX_TIME_SECONDS) ? ((t1).seconds <= (t2).seconds) : \
                            ((t1).beat24s <= (t2).beat24s)))

#define _ax_time_min(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? ((t1).samples < (t2).samples ? (t1) : (t2)) : \
  (((t)==AX_TIME_SECONDS) ? ((t1).seconds < (t2).seconds ? (t1) : (t2)) : \
                            ((t1).beat24s < (t2).beat24s ? (t1) : (t2))))

#define _ax_time_max(t,t1,t2) \
  (((t)==AX_TIME_SAMPLES) ? ((t1).samples > (t2).samples ? (t1) : (t2)) : \
  (((t)==AX_TIME_SECONDS) ? ((t1).seconds > (t2).seconds ? (t1) : (t2)) : \
                            ((t1).beat24s > (t2).beat24s ? (t1) : (t2))))

#define _ax_base_get_samplerate(a,b) (((AxBase)b)->context_limit.samplerate)
#define _ax_base_get_tempo(a,b) (((AxBase)b)->context_limit.tempo)
#define _ax_base_get_mixlength(a,b) (((AxBase)b)->context_limit.mixlength)
#define _ax_base_get_channels(a,b) (((AxBase)b)->context_limit.channels)

#define _ax_set_plugin(a,b,p) (((AxBase)b)->plugin = (p))
#define _ax_get_plugin(a,b) (((AxBase)b)->plugin)
#define _ax_set_methods(a,b,m) (((AxBase)b)->methods = (m))
#define _ax_get_methods(a,b) (((AxBase)b)->methods)
#define _ax_set_instance_data(a,b,d) (((AxBase)b)->instance_data = (d))
#define _ax_get_instance_data(a,b) (((AxBase)b)->instance_data)
#define _ax_set_name(a,b,n) (((AxBase)b)->name = (n))
#define _ax_get_name(a,b) (((AxBase)b)->name)
#define _ax_clone(a,b) (((AxBase)b)->methods->clone ((a), ((AxBase)b)))
#define _ax_destroy(a,b) (((AxBase)b)->methods->destroy ((a), ((AxBase)b)))
#define _ax_prepare(a,b) (((AxBase)b)->methods->prepare ((a), ((AxBase)b)))
#define _ax_process(a,b,c,i,o) \
        (((AxBase)b)->methods->process ((a),((AxBase)b),(c),(i),(o)))
#define _ax_length(a,b) (((AxBase)b)->methods->length ((a), ((AxBase)b)))
#define _ax_flush(a,b) (((AxBase)b)->methods->flush ((a), ((AxBase)b)))


/* util */
#define ax_malloc ss_malloc
#define ax_free ss_free

/* debug */
void ax_debug_down (void);
void ax_debug_up (void);

/* AXEL, ax_context */

AxBase ax_base_new_subclass (AXEL axel, size_t size);

AxContext _ax_context_copy (AXEL axel, AxContext dest);
AxContext _ax_context_merge (AXEL axel, AxContext dest);
AXEL _ax_register_plugin (AXEL axel, AxPlugin plugin);
AXEL _ax_unregister_plugin (AXEL axel, AxPlugin plugin);
AXEL _ax_register_base (AXEL axel, AxBase base);
AXEL _ax_unregister_base (AXEL axel, AxBase base);

/* ax_plugin */
void ax_plugin_initialise_defaults (AXEL axel);

/* ax_deck */
AxTrack _ax_deck_add_track (AXEL axel, AxDeck deck, AxTrack track);
AxTrack _ax_deck_remove_track (AXEL axel, AxDeck deck, AxTrack track);

/* ax_track */
AxBase ax_track_clone (AXEL axel, AxBase base);
AxLayer _ax_track_add_layer_above (AXEL axel, AxTrack track, AxLayer layer,
                                   AxLayer above);
AxLayer _ax_track_remove_layer (AXEL axel, AxTrack track, AxLayer layer);
AxLayer _ax_track_get_layer_above (AXEL axel, AxTrack track, AxLayer above);
AxLayer _ax_track_get_layer_below (AXEL axel, AxTrack track, AxLayer below);

/* ax_layer */
AxLayer _ax_remove_layer (AXEL axel, AxLayer layer);
AxBase ax_layer_clone (AXEL axel, AxBase base);
AxSound _ax_layer_add_sound (AXEL axel, AxLayer layer, AxSound sound,
			     AxTime position);
AxSound _ax_layer_remove_sound (AXEL axel, AxLayer layer, AxSound sound);
AxSound _ax_layer_get_sound_prev (AXEL axel, AxLayer layer, AxSound sound);
AxSound _ax_layer_get_sound_next (AXEL axel, AxLayer layer, AxSound sound);

/* ax_sound */
AxBase ax_sound_clone_with_layer (AXEL axel, AxBase base,
				    AxLayer new_layer);
int ax_sound_later (AXEL axel, AxSound s1, AxSound s2);

/* ax_envelope */
AxBase ax_envelope_clone (AXEL axel, AxBase base);


/* ax_channel */
AxChannel ax_channel_new (AXEL axel);
AxChannel ax_channel_clone (AXEL axel, AxChannel channel);
int ax_channel_destroy (AXEL axel, AxBase base);

AxChunk ax_channel_add_new_chunk (AXEL axel, AxChannel channel,
				  AxCount offset, AxCount length);

AxCount ax_channel_write0 (AXEL axel, AxChannel channel, AxCount length);

AxCount _ax_channel_write (AXEL axel, AxChannel channel, AxCount count,
			   AxChannel data);
AxCount _ax_channel_length (AXEL axel, AxChannel channel);
AxCount _ax_channel_seek (AXEL axel, AxChannel channel, AxCount offset);

AxCount ax_channel_interleave_2 (AXEL axel, AxChannel src1, AxChannel src2,
				 AxPCM * dest, AxCount count);
AxCount ax_channel_deinterleave_2 (AXEL axel, AxChannel dest1, AxChannel dest2,
				   AxPCM * src, AxCount count);
AxCount ax_channel_mix (AXEL axel, AxChannel src, AxChannel dest,
			AxCount count);


/* ax_channelset */
void ax_channelset_initialise_defaults (AXEL axel);

/* ax_chunk */
AxChunk ax_chunk_new (AXEL axel, AxCount start_index, AxCount length);
AxChunk ax_chunk_clone (AXEL axel, AxChunk chunk);
void ax_chunk_free (AXEL axel, AxChunk chunk);
AxCount _ax_chunk_clear_region (AXEL axel, AxChunk chunk, AxCount start,
				AxCount length, int channelname,
				void * unused);
AxCount _ax_chunk_gain (AXEL axel, AxChunk chunk, AxCount start, AxCount count,
			int channelname, /* (AxPCM *) */ void * gain);
AxCount _ax_chunk_copy (AXEL axel, AxChunk src, AxCount src_offset,
			AxChunk dest, AxCount dest_offset, AxCount count,
			int channelname, void * unused);
AxCount _ax_chunk_add_inplace (AXEL axel, AxChunk src, AxCount src_offset,
			       AxChunk dest, AxCount dest_offset,
			       AxCount count, int channelname, void * unused);
AxCount _ax_chunk_mult_inplace (AXEL axel, AxChunk src, AxCount src_offset,
				AxChunk dest, AxCount dest_offset,
				AxCount count, int channelname, void * unused);
AxCount _ax_chunk_fade_inplace (AXEL axel, AxChunk src, AxCount src_offset,
				AxChunk dest, AxCount dest_offset,
				AxCount count, int channelname, void * unused);
AxCount _ax_chunk_interleave_2 (AXEL axel, AxChunk src1, AxCount src1_offset,
				AxChunk src2, AxCount src2_offset,
				AxCount count, int unused, void * dest);
AxCount _ax_chunk_deinterleave_2 (AXEL axel,
				  AxChunk dest1, AxCount dest1_offset,
				  AxChunk dest2, AxCount dest2_offset,
				  AxCount count, int unused, void * src);
AxCount _ax_chunk_blend_inplace (AXEL axel, AxChunk src, AxCount src_offset,
				 AxChunk blend, AxCount blend_offset,
				 AxChunk dest, AxCount dest_offset,
				 AxCount count, int channelname,
				 void * unused);

/* XXX: remove these when dynamic! */
CDList __gain_init (AXEL axel);
CDList __sndfile_init (AXEL axel);

#endif /* defined(__AXEL__) */

#endif /* __AXEL_PRIVATE_H__ */
